# MIT License
#
# Copyright 2017 Broad Institute
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# This code was used when Evan and Arpy had to convert previous curation results after subclustering output had
# changed, mostly due to switch to OpenBLAS.  It should not be used in general.



#' @param old_dir Analysis directory containing curation_sheets directory with CSVs to be converted
#' @param new_dir Analysis directory in which curation_sheets directory will be created (if not already present)
#' and curation sheets written to it.
#' @return TRUE if successful.
ConvertCurationSheets=function(old_dir, new_dir, old_name, new_name,
            stop_if_failure=FALSE,
            new_curation_sheets_dirname=curation_sheets_dir) {
    old_dir = ensure_trailing_slash(old_dir)
    new_dir = ensure_trailing_slash(new_dir)
    old_ica_files = getNewIcaRdsFiles(old_dir, old_name)
    new_ica_files = getNewIcaRdsFiles(new_dir, new_name)
    if (length(old_ica_files) != length(new_ica_files)) {
        warning(sprintf("Different number of clusters in %s(%d) and %s(%d)",
        old_dir, length(old_ica_files), new_dir, length(new_ica_files)), immediate.=TRUE)
    }
    succeeded = TRUE
    old_curation_sheets_dir = ensure_directory_exists(paste0(old_dir, "/", curation_sheets_dir))
    new_curation_sheets_dir = ensure_directory_exists(paste0(new_dir, "/", new_curation_sheets_dirname))
    old_curation_sheets = get_curation_sheet_files(old_curation_sheets_dir)
    pdf(paste0(new_dir, "/", new_curation_sheets_dirname, "/", new_name, ".curation_corr.pdf"))
    top_level_mapping = plotRound1ClusterConfusionMatrix(old_dir, new_dir, old_name, new_name, threshold=0.6)
    for (new_cluster_name in names(top_level_mapping)) {
        if (is.na(new_ica_files[new_cluster_name])) {
           warning(paste0("Cluster ", new_cluster_name, " did not subcluster, so no curation sheet will be generated"), immediate.=TRUE)
        } else {
            new_ica = readRDS(new_ica_files[new_cluster_name])
            num_cells = nrow(new_ica$cell_rotations)
            if (is.na(top_level_mapping[new_cluster_name])) {
                warning(paste0("Could not find strong enough mapping for ", new_cluster_name, "; creating empty curation sheet"), immediate.=TRUE)
                make_new_curation_sheet(make_curation_sheet_path(new_curation_sheets_dir, new_cluster_name),
                    new_ica$gene_loadings, num_cells)
            } else {
                old_cluster_name = toString(top_level_mapping[new_cluster_name])
                curation_sheet = old_curation_sheets[old_cluster_name]
                if (is.na(curation_sheet)) {
                    warning(paste0("No curation sheet to convert for cluster ", new_cluster_name, "; creating empty curation sheet"), immediate.=TRUE)
                    make_new_curation_sheet(make_curation_sheet_path(new_curation_sheets_dir, new_cluster_name),
                    new_ica$gene_loadings, num_cells)
                } else {
                    cat_with_time(sprintf("Correlating %s$cell_rotations with %s$cell_rotations", old_ica_files[old_cluster_name], new_ica_files[new_cluster_name]))
                    old_cell_rotations = readRDS(old_ica_files[old_cluster_name])$cell_rotations
                    new_ica = readRDS(new_ica_files[new_cluster_name])
                    new_cell_rotations = new_ica$cell_rotations
                    new_gene_loadings = new_ica$gene_loadings
                    new_old_map = mapIcs(old_cell_rotations, new_cell_rotations, old_cluster_name, new_cluster_name)
                    new_curation_sheet = replace_cluster_number_in_curation_sheet_filename(curation_sheet, new_cluster_name)
                    map_curation_sheet(new_old_map, new_gene_loadings, num_cells,
                        file.path(old_curation_sheets_dir, curation_sheet),
                        file.path(new_curation_sheets_dir, new_curation_sheet))
                }
            }
        }
    }
    dev.off()
    if (stop_if_failure && !succeeded) {
        stop(paste0("Errors converting curation sheets from ", old_dir, " to ", new_dir))
    }
    return(succeeded)
}

#' subset of ConvertCurationSheets functionality that just creates the PDF
CorrelateOldAndNewIcs=function(old_dir, new_dir, old_name, new_name) {
    old_dir = ensure_trailing_slash(old_dir)
    new_dir = ensure_trailing_slash(new_dir)
    old_ica_files = getNewIcaRdsFiles(old_dir, old_name)
    new_ica_files = getNewIcaRdsFiles(new_dir, new_name)
    if (length(old_ica_files) != length(new_ica_files)) {
        warning(sprintf("Different number of clusters in %s(%d) and %s(%d)",
        old_dir, length(old_ica_files), new_dir, length(new_ica_files)), immediate.=TRUE)
    }
    succeeded = TRUE
    pdf(paste0(new_dir, "/", curation_sheets_dir, "/", new_name, ".cluster_corr.pdf"))
    top_level_cluster_mapping = plotRound1ClusterConfusionMatrix(old_dir, new_dir, old_name, new_name)
    for (cluster_name in names(old_ica_files)) {
        if (is.na(new_ica_files[cluster_name])) {
            succeeded = FALSE
            warning(paste0(cluster_name, " exists in ", old_dir, " but not in ", new_dir), immediate.=TRUE)
        } else {
            cat_with_time(sprintf("Correlating %s$cell_rotations to %s$cell_rotations", old_ica_files[cluster_name], new_ica_files[cluster_name]))
            old_cell_rotations = readRDS(old_ica_files[cluster_name])$cell_rotations
            new_cell_rotations = readRDS(new_ica_files[cluster_name])$cell_rotations
            try(mapIcs(old_cell_rotations, new_cell_rotations, cluster_name))
        }
    }
    dev.off()
    return(succeeded)
}

#' In addition to plotting the confusion matrix, returns a vector in which index/name is the new cluster ID,
#' and value is the old cluster index/name, so long as it is above threshold, else NA.
plotRound1ClusterConfusionMatrix=function(old_dir, new_dir, old_name, new_name, threshold=0.0) {
    new_top_level_clusters = readRDS(paste0(new_dir, "assign/", new_name, ".cluster.assign.RDS"))
    old_top_level_clusters = readRDS(paste0(old_dir, "assign/", old_name, ".cluster.assign.RDS"))
    confusion_matrix = make_class_intersection_matrix(old_top_level_clusters, new_top_level_clusters)
    cm_title = paste0("\nTop-level cluster overlap as a fraction of old cluster size.\nRows: old; Columns: new; Only in old: ",
    length(confusion_matrix$a_only), "; Only in new: ", length(confusion_matrix$b_only))
    corrplot(confusion_matrix$confusion_matrix, method="number", title=cm_title)
    ret=sapply(1:ncol(confusion_matrix$confusion_matrix), function(col_num) which.max(confusion_matrix$confusion_matrix[, col_num]))
    ret = sapply(1:length(ret), function(i) {
        best_correlation = confusion_matrix$confusion_matrix[ret[i], i]
        if (best_correlation < threshold) {
            warning(sprintf("Best top-level cluster mapping old(%d) => new(%d) is too low %f", ret[i], i, best_correlation), immediate.=TRUE)
            return(NA)
        } else {
            return(ret[i])
        }
    })
    names(ret) = levels(new_top_level_clusters)
    return(ret)
}

#' @return vector of file names, in which name of element is cluster number
getOldIcaRdsFiles=function(thedir, name) {
    components_dir = paste0(thedir, "/Components/")
    prefix = paste0(name, "Cluster")
    suffix = ".ICA.RDS"
    ica_files = list.files(components_dir, pattern=paste0(prefix, "[[:digit:]]+", suffix))
    cluster_names = sapply(ica_files, function(f) substr(f, nchar(prefix) + 1, nchar(f) - nchar(suffix)))
    ica_files = sapply(ica_files, function(f) paste0(components_dir, f))
    names(ica_files) = cluster_names
    return(ica_files)
}

getNewIcaRdsFiles=function(thedir, name) {
    cluster_dirs = list.files(thedir, pattern="cluster[[:digit:]]+", include.dirs=TRUE)
    cluster_names = sapply(cluster_dirs, function(d) substring(d, nchar("cluster") + 1))
    ica_files = sapply(cluster_names, function(cluster_name) getLatestIcaRds(cluster_name, name, thedir))
    names(ica_files) = cluster_names
    ica_files = ica_files[!sapply(ica_files, is.null)]
    # If there were nulls in ica_files, it was a list instead of a vector, so make it a vector
    ica_files = unlist(ica_files)
    return(ica_files[file.exists(ica_files)])
}

#' @return a map in which the indices are the old ICs and the values are the new ICs
mapIcs=function(old_cell_rotations, new_cell_rotations, old_cluster_name, new_cluster_name, threshold=0.75) {
    if (!identical(rownames(old_cell_rotations), rownames(new_cell_rotations))) {
        warning(sprintf("For new cluster %s (old cluster %s), cell barcode membership is not identical, so doing correlation on intersection.",
            new_cluster_name, old_cluster_name), immediate.=TRUE)
        shared_cells = intersect(rownames(old_cell_rotations), rownames(new_cell_rotations))
        old_cell_rotations = old_cell_rotations[shared_cells, ]
        new_cell_rotations = new_cell_rotations[shared_cells, ]
    }
    ic_corr = cor.r2(old_cell_rotations, new_cell_rotations)
    corrplot(matrix.sort(ic_corr, require_square=FALSE), title=paste0("\nCluster ", old_cluster_name, "=>", new_cluster_name), method="number")
    the_map = sapply(1:ncol(ic_corr), function(i) best_correlation(ic_corr[,i], threshold, new_cluster_name, i))

    # If an old IC maps to more than one new IC, use it for only the strongest correlation and set the other to NA
    for (old_ic in unique(the_map)) {
        if (!is.na(old_ic)) {
            new_ics = which(the_map == old_ic)
            if (length(new_ics) > 1) {
                corrs = sapply(new_ics, function(new_ic) {max(ic_corr[, new_ic])})
                good_ic = new_ics[which(corrs == max(corrs))]
                warning(sprintf("For new cluster %s (old cluster %s) old IC %d maps to multiple new ICs (%s), mapping to %d",
                            new_cluster_name, old_cluster_name, old_ic, paste(new_ics, collapse=", "), good_ic),
                        immediate.=TRUE)
                for (new_ic in new_ics) {
                    if (new_ic != good_ic) {
                        the_map[new_ic] = NA
                    }
                }
            }
        }
    }

    return(the_map)
}

cor.r2=function(x, y) {
    return(cor(x, y)^2)
}

best_correlation=function(corrs, threshold, cluster_name, old_ic) {
    best = max(corrs)
    ret = which(corrs == best)
    if (best < threshold) {
        warning(sprintf("In new cluster %s, best correlation for old IC %d is new IC %d with low correlation %f",
                    cluster_name, old_ic, ret, best), immediate.=TRUE)
        return(NA)
    }
    return(ret)
}

map_curation_sheet=function(new_old_map, new_gene_loadings, num_cells, old_sheet_file, new_sheet_file) {
    # load old curation sheet
    cat_with_time(sprintf("Mapping %s to %s", old_sheet_file, new_sheet_file))
    cs=read.table(old_sheet_file, stringsAsFactors=F, sep=",", header=T, quote="\"")
    # Old curation sheets are inconsistent on columns
    if (!('significant_negative_loading' %in% colnames(cs))) {
       cs$significant_negative_loading = rep(as.character(NA), nrow(cs))
    }

    expected_columns = c('ic.number', 'replicate_size_effects', 'cell_class', 'status', 'hypothesized_common_name',
        'molecular_label', 'n_cells', 'anatomical_region', 'positive_loading', 'negative_loading', 'confidence',
        'use_for_clustering', 'ic_notes', 'allen_stainings_examined', 'references', 'significant_negative_loading')

    if (!identical(expected_columns, colnames(cs))) {
        extra_columns = setdiff(colnames(cs), expected_columns)
        if (length(extra_columns) > 0) {
            warning(sprintf("Dropping unexpected columns (%s) from %s", paste(extra_columns, collapse=", "), old_sheet_file),
                immediate.=TRUE)
        }
        missing_columns = setdiff(expected_columns, colnames(cs))
        if (length(missing_columns) > 0) {
            stop(sprintf("%s is missing columns %s", old_sheet_file, paste(missing_columns, collapse=", ")), immediate.=TRUE)
        }
        cs = cs[, expected_columns]
    }

    if (length(which(is.na(cs$ic.number))) > 0) {
       warning(paste0("Discarding rows from ", old_sheet_file, " that have NA for ic.number"), immediate.=TRUE)
       cs=cs[!is.na(cs$ic.number), ]
    }

    # discard old ICs that don't map to a new IC
    cs = cs[cs$ic.number %in% new_old_map, ]
    # renumber based on new_old_map
    cs$ic.number=sapply(cs$ic.number, function(old_ic_number) which(new_old_map==old_ic_number))

    for (new_ic_number in setdiff(1:length(new_old_map), cs$ic.number)) {
        cs = add_new_curation_sheet_row(cs, num_cells, new_gene_loadings[, new_ic_number], new_ic_number)
    }

    # reorder based on new numbers
    cs = cs[order(cs$ic.number),]

    # updated positive_loading and negative_loading based on new ICs
    loadings = lapply(1:ncol(new_gene_loadings), function(ic_num) get_pos_and_neg_loadings_for_ic(new_gene_loadings[, ic_num]))
    cs$positive_loading = sapply(loadings, function(l) l$positive_loading)
    cs$negative_loading = sapply(loadings, function(l) l$negative_loading)

    write.csv(cs, file=new_sheet_file, quote=TRUE, row.names=FALSE, na="")
    # Read new file in order to confirm that quoting worked properly
    written = read.table(new_sheet_file, stringsAsFactors=F, sep=",", header=T, quote="\"")
    # Only pay attention to the columns we care about, because of differences in NA vs. empty string, etc.
    columns_to_compare = c("ic.number", "use_for_clustering", "status", "positive_loading", "negative_loading")
    cs.sub = cs[, columns_to_compare]
    written.sub = written[, columns_to_compare]
    # read.table capriciously puts in NA or "" depending on other values in the column.
    written.sub$use_for_clustering[is.na(written.sub$use_for_clustering)] = ""
    written.sub$status[is.na(written.sub$status)] = ""
    # Need to fix rownames after ordering, writing, reading
    attr(cs.sub, "row.names") = attributes(written.sub)$row.names
    if (all.equal(cs.sub, written.sub) != TRUE) {
        message = paste0("Problem converting ", old_sheet_file, " to ", new_sheet_file, "\n",
            paste(all.equal(cs.sub, written.sub), sep="\n"))
        stop(message)
    }
}

#' @return list(positive_loading, negative_loading), both represented as strings
get_pos_and_neg_loadings_for_ic=function(gene_loadings) {
    loading_list_length = 15
    loading_sep = "   "
    l = get_pos_and_neg_loading_lists_for_ic(gene_loadings, loading_list_length)
    return(list(positive_loading=paste(l$positive_loading, collapse=loading_sep),
                negative_loading=paste(l$negative_loading, collapse=loading_sep)))
}

make_new_curation_sheet=function(sheet_file, gene_loadings, num_cells) {
    cs = data.frame(ic.number=integer(0),
                    status=character(0),
                    use_for_clustering=character(0),
                    replicate_size_effects=logical(0),
                    cell_class=character(0),
                    hypothesized_common_name=character(0),
                    molecular_label=logical(0),
                    n_cells=integer(0),
                    anatomical_region=character(0),
                    positive_loading=character(0),
                    negative_loading=character(0),
                    confidence=integer(0),
                    ic_notes=character(0),
                    allen_stainings_examined=character(0),
                    references=logical(0),
                    significant_negative_loading=character(0), stringsAsFactors=FALSE)
    for (ic.number in 1:ncol(gene_loadings)) {
        cs = add_new_curation_sheet_row(cs, num_cells, gene_loadings[, ic.number], ic.number)
    }
    write.csv(cs, file=sheet_file, quote=TRUE, row.names=FALSE, na="")
}

add_new_curation_sheet_row=function(cs, num_cells, gene_loadings, ic.number) {
    loadings = get_pos_and_neg_loadings_for_ic(gene_loadings)
    cs = rbind(cs, 
           data.frame(ic.number=ic.number,
                  status="",
                  use_for_clustering="",
                  replicate_size_effects=as.logical(NA),
                  cell_class=NA,
                  hypothesized_common_name=NA,
                  molecular_label=as.logical(NA),
                  n_cells=num_cells, 
                  anatomical_region=NA,
                  positive_loading=loadings$positive_loading, 
                  negative_loading=loadings$negative_loading,
                  confidence=as.integer(NA),
                  ic_notes=NA,
                  allen_stainings_examined=NA,
                  references=as.logical(NA),
                  significant_negative_loading=NA, stringsAsFactors=FALSE), stringsAsFactors=FALSE)
    return(cs)
}
