# MIT License
#
# Copyright 2017 Broad Institute
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

IcaCluster = function(yamlpath=NULL) {

    pipeline_mgr = PipelineManager$new(
            subdirectory_names=workflow_directories,
            yamlpath=yamlpath,
            stages=all_stages,
            defaults=default_settings)

    stage_mgr = pipeline_mgr$stage_mgr
    pipeline.parameters = pipeline_mgr$pipeline.parameters
    load_dge_parameters = pipeline.parameters$load_dge

    dge_file_prefix = get_dge_file_prefix(pipeline_mgr)
    raw_dge_path = get_raw_dge_path(pipeline_mgr)
    filtered_raw_dge_path = get_filtered_raw_dge_path(pipeline_mgr)
    filtered_scaled_dge_path = get_filtered_scaled_dge_path(pipeline_mgr)
    scaled_dge_path = get_scaled_dge_path(pipeline_mgr)
    random.seed=pipeline.parameters$workflow$random.seed
    seurat_compatibility = pipeline.parameters$load_dge$seurat_compatibility

    # Is raw dge going to be necessary?  Does it need to be linked from a previous up-stream analysis

    if (pipeline_mgr$stage_mgr$beforeStage("load_dge")) {
        # TODO: estimate heap size based on size of input DGEs?
        merge_dge_heap_size = load_dge_parameters$MergeDgeHeapSize
        if (is.null(load_dge_parameters$filtered_gene_regexp) || seurat_compatibility) {
            filtered_gene_re_args = list()
        } else {
            filtered_gene_re_args = lapply(load_dge_parameters$filtered_gene_regexp, function(v) paste0("FILTERED_GENE_RE=", v))
        }
        if (is.null(load_dge_parameters$select_cells_file)) {
            select_cells_file_args = list()
        } else {
            select_cells_file_args = lapply(load_dge_parameters$select_cells_file, function(v) paste0("SELECTED_CELL_BARCODES=", v))
        }
        if (seurat_compatibility) {
            # defer filtering until after scaling
            min_genes = 0
            min_cells = 0
            min_transcripts = 0
        } else {
            min_genes = load_dge_parameters$min_genes
            min_cells = load_dge_parameters$min_cells
            min_transcripts = load_dge_parameters$min_transcripts
        }

        args = c(paste0("-Xmx", merge_dge_heap_size),
                "-jar", get_jar_path("dropseq"), "MergeDgeSparse",
                "YAML=", pipeline_mgr$yamlpath,
                "CELL_SIZE_OUTPUT_FILE=",  paste0(dge_file_prefix, ".cell_size.txt.gz"),
                "RAW_DGE_OUTPUT_FILE=", raw_dge_path,
                "SCALED_DGE_OUTPUT_FILE=", scaled_dge_path,
                "MIN_GENES=", min_genes,
                "MIN_CELLS=", min_cells,
                "MIN_TRANSCRIPTS=", min_transcripts,
                filtered_gene_re_args,
                select_cells_file_args, recursive=TRUE)
        ret = system2("java", args)
        if (ret != 0) {
            stop(paste0(ret, " exit status returned by: java ", paste(args)))
        }
        stage_mgr$afterStage()
    }
    if (pipeline_mgr$stage_mgr$beforeStage("select_genes")) {
        raw_dge = loadSparseDge(raw_dge_path)
        scaled_dge = loadSparseDge(scaled_dge_path)

        if (seurat_compatibility) {
            num.genes<-colSums(raw_dge>0)
            cellidx.use=which(num.genes>=load_dge_parameters$min_genes)
            num.cells<-rowSums(raw_dge>0)
            geneidx.use = which(num.cells>=load_dge_parameters$min_cells)
            if (!is.null(load_dge_parameters$filtered_gene_regexp)) {
                genes = rownames(scaled_dge)
                for (filtered_gene_regexp in load_dge_parameters$filtered_gene_regexp) {
                    mito.idx=grep(rownames(raw_dge), pattern=filtered_gene_regexp)
                    geneidx.use = setdiff(geneidx.use, mito.idx)
                }
            }
            scaled_dge = scaled_dge[geneidx.use, cellidx.use]
            filtered_raw_dge = raw_dge[geneidx.use, cellidx.use]
        } else {
            filtered_raw_dge = raw_dge
        }
        saveRDS(filtered_raw_dge, file=filtered_raw_dge_path)
        saveRDS(scaled_dge, file=filtered_scaled_dge_path)
        pdf(paste0(pipeline.parameters$name, '.MeanVarPlot.pdf'), width = 10, height = 10)
        genes.use = select_genes(raw_dge=filtered_raw_dge, scaled_dge=scaled_dge,
            numGenes=pipeline.parameters$selectgenes$num_most_expressed_genes,
            alphathresh=pipeline.parameters$selectgenes$alphathresh,
            varthresh=pipeline.parameters$selectgenes$varthresh)
        dev.off()
        rm(filtered_raw_dge)
        cat_with_time(paste0('There are ', length(genes.use), ' variable genes in this dataset.'))
        write.table(genes.use,
                  file=paste0(pipeline.parameters$name, '.selected_genes.txt'),
                  quote=FALSE, row.names=FALSE, col.names=FALSE)
        scaled_centered_gene_selected_dge=scale(t(scaled_dge[genes.use,]),center=TRUE,scale=TRUE)
        rm(scaled_dge)
        stage_mgr$afterStage("scaled_centered_gene_selected_dge")
    }
    if (stage_mgr$beforeStage("ica")) {
        n.components = if (pipeline.parameters$ica$n.components <= 0) {
            cat_with_time("Using jackstraw to select number of ICs")
            length(jackStrawFast(scaled.data=t(scaled_centered_gene_selected_dge), num.replicate=100, random.seed=random.seed))
        } else {
            cat_with_time(paste0("Number of ICs set manually: ", pipeline.parameters$ica$n.components))
            pipeline.parameters$ica$n.components
        }
        ica.obj<-run_ica(data.use=scaled_centered_gene_selected_dge,
                    n.comp=n.components,
                    random.seed=random.seed,
                    method=pipeline.parameters$ica$ica.method,
                    verbose=pipeline.parameters$ica$ica.verbose)
        cat_with_time("Starting Rtsne")
        tsne=do_tsne(ica.obj$S,pca=F, random.seed=random.seed)
        cat_with_time("Rtsne done")
        rownames(tsne)<-rownames(ica.obj$S)
        saveRDS(list(gene_loadings=ica.obj$A, cell_rotations=ica.obj$S, tsne=tsne),
            file=file.path(pipeline_mgr$directories$components, paste0(pipeline.parameters$name, '.raw_ica.RDS')))
        saveRDS(tsne, file=file.path(pipeline_mgr$directories$tSNE, paste0(pipeline.parameters$name, auto_tsne_xy_extension)))

        if (!exists("raw_dge")) {raw_dge = loadSparseDge(raw_dge_path)}
        pdffile = paste0(pipeline_mgr$directories$components, pipeline.parameters$name, "_FirstRoundICAreport.pdf")
        ica.obj<-CurateICAOne(ica.obj, dge=raw_dge,file=pdffile, skewness_threshold=0,kurtosis_threshold=0,
                                featureplot=T,tsrots=tsne)
        saveRDS(list(gene_loadings=ica.obj$A, cell_rotations=ica.obj$S, tsne=tsne), file=file.path(pipeline_mgr$directories$components, paste0(pipeline.parameters$name, '.flipped_sorted_raw_ica.RDS')))

        # create top-level blank curation sheet
        make_new_curation_sheet(file.path(pipeline_mgr$directories$blank_curation_sheets, paste0(pipeline.parameters$name, first_round_curation_sheet_extension)),
            gene_loadings=ica.obj$A, num_cells=nrow(ica.obj$S))


        stage_mgr$afterStage("tsne", "ica.obj", "scaled_centered_gene_selected_dge", "n.components")
    }

    if (stage_mgr$beforeStage("apply_round1_ic_curation")) {

        if (!exists("filtered_scaled_dge")) {filtered_scaled_dge = readRDS(filtered_scaled_dge_path)}

        first_round_curation_sheet_path = file.path(pipeline_mgr$directories$curation_sheets, paste0(pipeline.parameters$name, first_round_curation_sheet_extension))
        if (file.exists(first_round_curation_sheet_path)) {
            cat_with_time(paste0("Applying curation sheet ", first_round_curation_sheet_path))
            curation.data = readAndValidateCurationSheet(first_round_curation_sheet_path, maxIcs=ncol(ica.obj$S), pipeline_mgr=pipeline_mgr)
            if (is.null(curation.data)) {
                stop(paste0("Problem with curation sheet ", first_round_curation_sheet_path))
            }

            curated.ica.obj = CurateICAManual(list(gene_loadings=ica.obj$A, cell_rotations=ica.obj$S), curation.data, cluster.doublets=TRUE,
                    doubletalpha=0.1, file=file.path(pipeline_mgr$directories$components, paste0(pipeline.parameters$name, ".ManualFirstRoundICAreport.pdf")),
                    suggest.centers=FALSE, unordered.tsne=tsne)
            # Not sure if this will be needed, but saving just in case.
            saveRDS(curated.ica.obj, file=file.path(pipeline_mgr$directories$components, paste0(pipeline.parameters$name, firstround_manual_ica_extension)))

            # replace existing ica.obj slots with the curated stuff
            ica.obj$S = curated.ica.obj$S
            ica.obj$A = curated.ica.obj$A
            tsne = curated.ica.obj$tSNE
            saveRDS(tsne, file=file.path(pipeline_mgr$directories$tSNE, paste0(pipeline.parameters$name, manual_tsne_xy_extension)))
            saveRDS(tsne, file=file.path(pipeline_mgr$directories$tSNE, paste0(pipeline.parameters$name, tsne_xy_extension)))
        } else {
            cat_with_time(sprintf("Ccuration sheet %s not found, so not curating first-round ICs", first_round_curation_sheet_path))
            file.copy(from=file.path(pipeline_mgr$directories$tSNE, paste0(pipeline.parameters$name, auto_tsne_xy_extension)),
                    to=file.path(pipeline_mgr$directories$tSNE, paste0(pipeline.parameters$name, tsne_xy_extension)),
                    overwrite=TRUE)
        }

        # assign column names
        ica.obj = lapply(ica.obj, function(x) {colnames(x)=paste0("IC", seq_len(ncol(x)));return(x)})
        if (!exists("n.components")) {
            # This should only happen for older stage files, because n.components was added to prev stage 25-04-2017
            n.components = pipeline.parameters$ica$n.components
        }
        dr_rds_label = sprintf("%s, ICA k=%d", pipeline.parameters$name, n.components)

        saveRDS(list(gene_loadings=ica.obj$A, cell_rotations=ica.obj$S, name=dr_rds_label, tsne=tsne),
            file=paste0(pipeline_mgr$directories$components, pipeline.parameters$name, ".", ica_file_extension))

        if (!is.null(pipeline.parameters$ica$canonical_gene_sets_file)) {
            PlotCanonicalGenes(tsne=tsne, norm.dge=filtered_scaled_dge,
                output_directory=file.path(pipeline_mgr$directories$featureplots, canonical_plots_dir),
                canonical_genes_yaml_file=pipeline.parameters$ica$canonical_gene_sets_file)
        }
        stage_mgr$afterStage("tsne", "ica.obj", "scaled_centered_gene_selected_dge")
    }

    if (stage_mgr$beforeStage("cluster")) {
        small.cluster.size = pipeline.parameters$cluster$small.cluster.size
        if (pipeline.parameters$cluster$skip.first.round) {
            idents = makeDummyLouvainCluster(rownames(ica.obj$S))
        } else {
            nn=25
            idents = louvainCluster(data=ica.obj$S,min.clust=small.cluster.size, nn=nn, nstart=100, iter.max=20,
                    jaccard=T,R=pipeline.parameters$cluster$res.clust, orig.scale = scaled_centered_gene_selected_dge,
                    SNN=NULL,generate.tree=T,reorder.tree=T,reorder.numeric=T, random.seed=random.seed,
                    out_edge_file=
                        if (pipeline.parameters$cluster$save.edges)
                            file.path(pipeline_mgr$directories$assign,sprintf("%s.nn%d.resolution%s%s",
                                pipeline.parameters$name, nn, pipeline.parameters$cluster$res.clust, louvain_edges_extension))
                        else NA)
        }
        rm(scaled_centered_gene_selected_dge)

        saveRDS(idents$louvain, file=paste0(pipeline_mgr$directories$assign, pipeline.parameters$name, cluster_assign_extension))
        if (!exists("raw_dge")) {raw_dge = loadSparseDge(raw_dge_path)}
        # Not all cells will have been put into clusters, so only create metacells of those that are in a cluster.
        metacells = sum_dge_by_cluster(idents=idents$louvain, raw_dge=raw_dge[, names(idents$louvain)])
        saveRDS(metacells, file=paste0(pipeline_mgr$directories$metacells, pipeline.parameters$name,".metacells.RDS"))

        # Louvain will assign a small set of cells to singleton clusters, because they could not be clustered with the other cells.  Flag and report these cells in the "DroppedSingletons.Robj" file.
        c=intersect(names(idents$louvain),rownames(ica.obj$S))
        cells.dropped=setdiff(rownames(ica.obj$S),c)
        saveRDS(cells.dropped,file=paste0(pipeline_mgr$directories$assign, pipeline.parameters$name,".DroppedSingletons.RDS"))
        cat_with_time(paste0("Dropped ",length(cells.dropped)," singletons in round 1 of clustering."))

        #plot and save the round 1 Louvain Clustering results, along with the k-means prior that is used to remove the small "goofy" cells
        idx.small.clusters <- levels(idents$kmeans)[which(table(idents$kmeans) < small.cluster.size)]
        smallcluster.cells <- which(idents$kmeans %in% idx.small.clusters)

        pdf(paste0(pipeline_mgr$directories$tSNE, pipeline.parameters$name,"_tSNE.pdf"))
        ident.plot(tsne[c,],factor(idents$louvain[c]), legend=TRUE)
        title(main="First round, Louvain Clustering")
        if (length(idents$kmeans) > 0) {
            ident.plot(tsne,factor(idents$kmeans))
            title(main="K means clustering (only for detecting small clusters)")
            highlight.cells(tsne,smallcluster.cells,col='blue')
            title(main="Small cluster cells identified by Kmeans")
        }
        dev.off()

        if (!exists("filtered_scaled_dge")) {filtered_scaled_dge = readRDS(filtered_scaled_dge_path)}
        if (!exists("filtered_raw_dge")) {filtered_raw_dge = readRDS(filtered_raw_dge_path) }

        pairwise_and_cluster_diffexp = pairwiseAndClusterDiffExpression(idents=idents$louvain,
                raw_dge=filtered_raw_dge, scaled_dge=filtered_scaled_dge, only.pos=FALSE,
                FOLDch.thresh=pipeline.parameters$differential_expression$FOLDch.threshold,
                pval.thresh=pipeline.parameters$differential_expression$pval.threshold)

        saveRDS(pairwise_and_cluster_diffexp$pairwise_markers,file=paste0(pipeline.parameters$name, firstround_label, pairwise_markers_extension))
        saveRDS(pairwise_and_cluster_diffexp$cluster_markers, file=paste0(pipeline.parameters$name,firstround_label, cluster_markers_extension))
        rm(pairwise_and_cluster_diffexp)


        # A new report pdf I am developing to provide helpful metrics about the resulting clustering:  differentially expressed genes, most enriched ICs, etc.
        pdf(paste0(pipeline.parameters$name,"clusterreport_ROUND1.pdf"))
        clusterreport(idents=idents$louvain,tree=idents$tree,tsne=tsne,cell.matrix=ica.obj$S,gene.matrix=t(ica.obj$A),norm=filtered_scaled_dge,dge=filtered_raw_dge)
        dev.off()

        saveRDS(idents$tree, file=file.path(pipeline_mgr$directories$phylo, paste0(pipeline.parameters$name, ".phylo.RDS")))
        if (is.null(idents$tree)) {
            cat_with_time("Not generating first-round phylo.pdf because not enough nodes in tree")
        } else {
            pdf(file.path(pipeline_mgr$directories$phylo, paste0(pipeline.parameters$name, ".phylo.pdf")))
            plot.phylo(idents$tree,direction="downwards", show.node.label=TRUE)
            nodelabels() # For some reason show.node.label==TRUE doesn't work, but this does
            dev.off()
        }

        stage_mgr$afterStage("idents")
    }
    if (stage_mgr$beforeStage("subica")) {

        min.subclust=pipeline.parameters$subica$min.subclust
        # If number of ICs has been specified manually, run sub-ICA regardless of whether the cluster is big enough.
        to.subcluster=levels(idents$louvain)[which(table(idents$louvain)>min.subclust | levels(idents$louvain) %in%  names(pipeline.parameters$subica$n_components))]

        if (!exists("raw_dge")) {raw_dge = loadSparseDge(raw_dge_path)}
        if (!exists("filtered_scaled_dge")) {filtered_scaled_dge = readRDS(filtered_scaled_dge_path)}
        cs<-colSums(raw_dge)

        #This process can easily be parallelized with "mclapply"
        did_subcluster = sapply(to.subcluster,function(cluster.num){

            cluster.dir=get_cluster_directory(cluster.num)
            cells.all=names(idents$louvain)[which(idents$louvain ==cluster.num)]
            if (pipeline.parameters$subica$drop.small.cells > 0) {
                # Retain cells larger than given percentile of cell sizes.
                cat_with_time(sprintf("Only keeping cells > than %f percentile cell size", pipeline.parameters$subica$drop.small.cells * 100))
                cells.use=cells.all[which(cs[cells.all]>quantile(cs[cells.all], pipeline.parameters$subica$drop.small.cells, names=FALSE))]
            } else {
                cells.use = cells.all
            }

            norm.use=filtered_scaled_dge[,cells.use]
            norm.use=norm.use[which(rowSums(norm.use)>0),]
            dge.use=raw_dge[, cells.use]

            dge.use=dge.use[which(rowSums(dge.use)>0),]
            saveRDS(dge.use, file=paste0(cluster.dir, pipeline.parameters$name,"_cluster",cluster.num,".DGE.RDS"))


            cat_with_time(paste0(ncol(dge.use)," cells are going into subanalysis of cluster ",cluster.num,"."))
            pdf(paste0(cluster.dir,"MeanVarPlot_",pipeline.parameters$name,"_",cluster.num,"cluster.pdf"))
            var.genes=mean.var.auto(dge.use, norm.use, alphathresh=pipeline.parameters$selectgenes$alphathresh,
                                    varthresh=pipeline.parameters$selectgenes$varthresh)
            dev.off()
            scaled=t(scale(t(norm.use[var.genes,]),center=T,scale=T))
                                                #scaled=apply(norm.use[var.genes,],1,scalar1)
                                                #rownames(scaled)<-colnames(norm.use)
                                                #colnames(scaled)<-var.genes
            rownames(scaled)<-var.genes
            colnames(scaled)<-colnames(norm.use)
            cat_with_time(dim(scaled))

            # Save this for multiple subclustering runs with different parameters
            saveRDS(list(scaled_gene_selected_dge=scaled, raw_dge=dge.use),
                file=paste0(cluster.dir, pipeline.parameters$name, ".subcluster_inputs.RDS"))

            n.components = pipeline.parameters$subica$n_components[[toString(cluster.num)]]
            use_jackstraw = is.null(n.components)
            if (use_jackstraw) {
                cat_with_time(paste0("Using jackstraw to select number of components for cluster ", cluster.num))
                cat_with_time(system.time(n.components <-
                    length(jackStrawFast(scaled.data=scaled,num.replicate=100, random.seed=random.seed))))
            } else {
                cat_with_time(paste0("Number of components specified manually for cluster ", cluster.num))
            }
            cat_with_time(paste0("Chose ",n.components," components."))
            if(n.components > 1) {
                ica_and_curate(scaled_gene_selected_dge=scaled,
                                raw_dge=dge.use,
                                n.components=n.components,
                                cluster.dir=cluster.dir,
                                components.dir=pipeline_mgr$directories$components,
                                blank.curation.sheets.dir=pipeline_mgr$directories$blank_curation_sheets,
                                cluster.num=cluster.num,
                                file_prefix=make_subcluster_file_prefix(project.name=pipeline.parameters$name,
                                                            cluster.num=cluster.num,
                                                            n.components=n.components,
                                                            used_jackstraw=use_jackstraw),
                                random.seed=random.seed,
                                ica.method=pipeline.parameters$subica$ica.method,
                                ica.verbose=pipeline.parameters$subica$ica.verbose)
                return(TRUE)
            } else {
                cat_with_time(paste0("One or fewer significant components found.  No subclustering performed."))
                return(FALSE)
            }
        })
        # Remove clusters that did not have enough significant ICs to subcluster
        to.subcluster=to.subcluster[which(did_subcluster)]
        stage_mgr$afterStage("to.subcluster")
    }

    if (stage_mgr$beforeStage("agg_subica")) {
        subcluster_ic_selections = lapply(to.subcluster, function(cluster.num) {
            return(c(cluster.num=cluster.num, get_subcluster_ic_selection(project.name=pipeline.parameters$name, cluster.num=cluster.num)))
        })
        ICs = lapply(subcluster_ic_selections, function(l) {
            subcluster_path = getLatestIcaRds(cluster.num=l$cluster.num, name=pipeline.parameters$name)
            cat_with_time(paste0("Loading ", subcluster_path))
            readRDS(subcluster_path)
        })
        names(ICs) = to.subcluster
        saveRDS(ICs, file=paste0(pipeline.parameters$name, all_ics_extension))

        manual_n_component_overrides = c(lapply(subcluster_ic_selections, function(subcluster_ic_selection) {
                if (!subcluster_ic_selection$used_jackstraw) {
                    ret = list(subcluster_ic_selection$n.components)
                    names(ret) = c(subcluster_ic_selection$cluster.num)
                    return(ret)
                } else {
                    return(list())
                }
        }), recursive=TRUE)
        comment = "# This file may be pasted into a YAML for IcaLouvain in order to force the manual overrides for number of ICs that were used in this clustering run.
#Note that if there already is a 'subcluster' section of the YAML, the content of this file should be added to that."
        write(paste(comment, as.yaml(list(subica=list(n_components=as.list(manual_n_component_overrides)))), sep="\n"),
                file=paste0(pipeline.parameters$name, ".manual_num_component_overrides.yaml"))


        stage_mgr$afterStage("ICs")
    }

    # Do this unconditionally so that path to results.curated can be established for all downstream steps
    subcluster_parameters = pipeline.parameters$subcluster
    cluster.method = subcluster_parameters$cluster.method
    cluster_method_label = get_cluster_method_label(pipeline.parameters)
    center.KDE = subcluster_parameters$center.KDE
    results.curated_path = get_curated_results_path(pipeline.parameters)

    if (pipeline_mgr$stage_mgr$beforeStage("apply_ic_curation")) {
        cat_with_time(paste0("Curating with center.KDE=", center.KDE))
        curation.files=get_curation_sheet_files(pipeline_mgr$directories$curation_sheets)

        if (!exists("filtered_scaled_dge")) {filtered_scaled_dge = readRDS(filtered_scaled_dge_path)}

        curated_ic_objs_and_curation_data = lapply(names(ICs), function(cluster.name) {
            ic.data=ICs[[cluster.name]]
            curation.file = curation.files[cluster.name]
            if (is.na(curation.file)) {
                cat_with_time(paste0("No curation file found for cluster ", cluster.name, "; no subclustering will be done"))
                return(NULL)
            }
            cat_with_time(paste0("Applying file ", curation.file))
            curation.data = readAndValidateCurationSheet(file.path(pipeline_mgr$directories$curation_sheets, curation.file),
                        maxIcs=ncol(ic.data$cell_rotations), pipeline_mgr=pipeline_mgr)
            if (is.null(curation.data)) {
                return(NULL)
            }

            output_dir = ensure_directory_exists(file.path(paste0("cluster", cluster.name), cluster_method_label))
            samplename=paste0(pipeline.parameters$name, ".cluster", cluster.name)
            ic=CurateICAManual(ic.data, curation.data, filtered_scaled_dge, genes.project=0, cluster.doublets=TRUE,
                    doubletalpha=0.1,file=paste0(output_dir, samplename,".ManualICAReports.pdf"),suggest.centers=center.KDE)
            saveRDS(object=ic$tSNE,file=paste0(pipeline_mgr$directories$tSNE,samplename,curated_tsne_extension))
            curation.data$cluster = cluster.name
            if(ncol(as.matrix(ic$S)) == 0) {
                cat_with_time(paste0("No ICs resulting from curation for cluster ", cluster.name, " so all cells will go into single subcluster"))
                return(list(curated.ic.obj=NULL, curation.data=curation.data, cells=rownames(ic.data$cell_rotations)))
            }
            return(list(curated.ic.obj=ic, curation.data=curation.data, cells=rownames(ic.data$cell_rotations)))
        })
        # Save the curation sheets in an easy-to-process form
        saveRDS(merge_curation_sheets(lapply(curated_ic_objs_and_curation_data, function(l) return(l$curation.data))),
                file=file.path(pipeline_mgr$directories$components, paste0(pipeline.parameters$name, ic_curation_summary_extension)))

        curated.ic.objs = lapply(curated_ic_objs_and_curation_data, function(l) return(l$curated.ic.obj))
        names(curated.ic.objs) = names(ICs)
        curation_cell_holdouts = lapply(curated.ic.objs, function(ic) {
            return(list(doublets=ic$doublets, outliers=ic$outliers, doublet_outlier_thresholds=ic$doublet_outlier_thresholds))
        })
        saveRDS(curation_cell_holdouts,
            file=file.path(pipeline_mgr$directories$assign, paste0(pipeline.parameters$name, curation_cell_holdouts_extension)))

        non.null.ic.objs = curated.ic.objs[sapply(curated.ic.objs, function(v) !is.null(v))]

        if (length(non.null.ic.objs) == 0) {
            stop("ERROR: No clusters survived IC curation.")
        }

        if (cluster.method == louvain_subcluster_method) {
            # generate blank curation sheet for subcluster Louvain resolutions
            blank_louvain_curation_sheet_file = file.path(pipeline_mgr$directories$blank_curation_sheets, paste0(pipeline.parameters$name, louvain_parameters_sheet_extension))
            cat_with_time(paste0("Writing ", blank_louvain_curation_sheet_file))
            write.csv(data.frame(cluster=names(non.null.ic.objs), resolution="", nn="", comments=""), quote=TRUE, row.names=FALSE, na="",
                    file=blank_louvain_curation_sheet_file)
        }

        for(cluster.name in names(non.null.ic.objs)) {
            ic = curated.ic.objs[[cluster.name]]
            if (is.null(ic$tSNE)) {
                # If ic$tSNE is null, which can happen if there were 2 or fewer ICs in a cluster,
                # then tSNE is missing, so create it here
                cat_with_time(paste0("Caching tSNE for ", cluster.name))
                if (!exists("filtered_scaled_dge")) {filtered_scaled_dge = readRDS(filtered_scaled_dge_path)}
                os=scale(t(filtered_scaled_dge[rownames(ic$A), rownames(ic$S)]), center=TRUE, scale=TRUE)
                idx=which(is.finite(colSums(os)))
                os=os[,idx]
                ic$tSNE = do_tsne(os[rownames(ic$S),], random.seed=random.seed)
                rownames(ic$tSNE)<-rownames(ic$S)
                # Save tSNE for final subclustering in next stage
                curated.ic.objs[[cluster.name]]$tSNE = ic$tSNE
            }
        }

        # Capture clusters that did not have any real ICs after curation, so that each cluster can be given
        # a single subcuster.
        no_ics_clusters = sapply(curated_ic_objs_and_curation_data,
            function(l) if (!is.null(l$cells) && is.null(l$curated.ic.obj)) {return(l$cells)} else {return(NULL)})
        names(no_ics_clusters) = names(ICs)
        no_ics_clusters = no_ics_clusters[sapply(no_ics_clusters, function(v) !is.null(v))]

        apply_ic_curation_version = current_apply_ic_curation_version
        stage_mgr$afterStage("curated.ic.objs", "no_ics_clusters", "apply_ic_curation_version")
    }


    if (pipeline_mgr$stage_mgr$beforeStage("subcluster")) {
        if (!exists("apply_ic_curation_version") || apply_ic_curation_version != current_apply_ic_curation_version) {
            stop("Old version of apply_ic_curation stage file.  Run stage apply_ic_curation again.")
        }
        gmm.model.threshold = subcluster_parameters$gmm.model.threshold
        cat_with_time(sprintf("Subclustering with cluster.method %s; cluster_method_label %s, gmm.model.threshold %s",
                    cluster.method, cluster_method_label, gmm.model.threshold))

        if (!exists("filtered_scaled_dge")) {filtered_scaled_dge = readRDS(filtered_scaled_dge_path)}
        if (!exists("raw_dge")) {raw_dge = loadSparseDge(raw_dge_path)}

        if (cluster.method == louvain_subcluster_method) {
            louvain_resolutions_file = get_single_file_matching_regexp(louvain_parameters_sheet_extension,
                path=pipeline_mgr$directories$curation_sheets,
                pattern=glob2rx(paste0("*", louvain_parameters_sheet_extension)), full.names=TRUE)
            conn=pipeline_mgr$open_curation_file_with_history(louvain_resolutions_file)
            louvain_parameters = read.table(conn, stringsAsFactors=F, sep=",", header=T, quote="\"")
            close(conn)
            rownames(louvain_parameters) = louvain_parameters$cluster
        } else {
            louvain_resolutions = data.frame()
        }


        results.curated = lapply(names(curated.ic.objs), function(cluster.name) {
            ic=curated.ic.objs[[cluster.name]]
            if (is.null(ic)) return(NULL)

            output_dir = ensure_directory_exists(file.path(paste0("cluster", cluster.name), cluster_method_label))
            samplename=paste0(pipeline.parameters$name, ".cluster", cluster.name)
            cluster_report_pdf = file.path(output_dir, paste0(samplename,"_", cluster_method_label,".clusterreport.pdf"))
            subcluster_titrate_paths = if (cluster.method == louvain_subcluster_method) {
                make_subcluster_titrate_paths(output_dir, samplename=samplename, cluster_method_label=cluster_method_label,
                    resolution=louvain_parameters[cluster.name,]$resolution, nn=louvain_parameters[cluster.name,]$nn)
            } else {
                NA
            }
            if (!pipeline.parameters$subcluster$force.resubcluster && !is.na(subcluster_titrate_paths) &&
                all(file.exists(unlist(subcluster_titrate_paths)))) {
                cat_with_time(sprintf("Recycling %s for subcluster stage", subcluster_titrate_paths$rds))
                prev_stage_file = pipeline_mgr$stage_mgr$getStageFile(pipeline_mgr$stage_mgr$currentStage - 1)
                if (file.mtime(subcluster_titrate_paths$rds) < file.mtime(prev_stage_file)) {
                    cat_with_time(sprintf("WARNING: %s is older than %s.  Is the subclustering output stale?", subcluster_titrate_paths$rds, prev_stage_file))
                }
                idents = readRDS(subcluster_titrate_paths$rds)
                cat_with_time(sprintf("Copying cached subcluster pdf %s to final %s", subcluster_titrate_paths$pdf, cluster_report_pdf))
                file.copy(from=subcluster_titrate_paths$pdf, to=cluster_report_pdf, overwrite=TRUE)
            } else {
                os=scale(t(filtered_scaled_dge[rownames(ic$A), rownames(ic$S)]), center=TRUE, scale=TRUE)
                idx=which(is.finite(colSums(os)))
                os=os[,idx]
                if (cluster.method == louvain_subcluster_method) {
                    cat_with_time(sprintf("Louvain parameters for cluster %s: %s",
                        cluster.name, paste0(louvain_parameters[cluster.name,], collapse=", ")))
                }
                idents=clusterRound2(data=ic$S, method=cluster.method, orig.scale=os,
                        ica.obj=ic, norm=filtered_scaled_dge[,rownames(ic$S)],
                        dge=raw_dge[rownames(filtered_scaled_dge),rownames(ic$S)],
                        resolution=louvain_parameters[cluster.name,]$resolution, nn=louvain_parameters[cluster.name,]$nn,
                        file=cluster_report_pdf,
                        tolFun=pipeline.parameters$subcluster$tolFun, gmm.model.threshold=gmm.model.threshold,
                        out_edge_file=
                            if (pipeline.parameters$subcluster$save.edges)
                                make_subcluster_edge_path(pipeline.parameters, cluster.name,
                                                        resolution=louvain_parameters[cluster.name,]$resolution,
                                                        nn=louvain_parameters[cluster.name,]$nn)
                            else NA
                )
                if (!is.na(subcluster_titrate_paths)) {
                    cat_with_time(sprintf("Caching subcluster results %s and %s", subcluster_titrate_paths$rds, subcluster_titrate_paths$pdf))
                    file.copy(from=cluster_report_pdf, to=subcluster_titrate_paths$pdf, overwrite=TRUE)
                    saveRDS(idents, file=subcluster_titrate_paths$rds)
                }
            }
            return(list(clustering=idents,ic=ic))
        })
        names(results.curated) = names(curated.ic.objs)
        saveRDS(results.curated, file=results.curated_path)



        subcluster_idents = get_subcluster_idents(results.curated)

        if (length(no_ics_clusters) > 0) {
            # for each cluster that did not subcluster due to lack of ICs after curation, create a single subcluster
            # If USE.NAMES is not FALSE, then the cell barcodes get mangled when the factors are merged.
            no_subcluster_idents = unlist(lapply(names(no_ics_clusters),
                function(cluster_name) {
                    cells = no_ics_clusters[[cluster_name]]
                    f = as.factor(rep(paste0(cluster_name, "-1"), length(cells)))
                    names(f) = cells
                    return(f)
                }))
            idents = unlist(list(subcluster_idents, no_subcluster_idents))
        } else {
            idents = subcluster_idents
        }

        saveRDS(idents, file=file.path(pipeline_mgr$directories$assign, paste0(pipeline.parameters$name,".subcluster.assign.RDS")))

        metacells = sum_dge_by_cluster(idents=idents, raw_dge=raw_dge[, names(idents)])
        saveRDS(metacells, file=paste0(pipeline_mgr$directories$metacells, pipeline.parameters$name,".subcluster.metacells.RDS"))

        d = make.dendrogram(data=t(filtered_scaled_dge), partition=idents, min.clust.size=1)
        saveRDS(d$tree, file=file.path(pipeline_mgr$directories$phylo, paste0(pipeline.parameters$name, ".subcluster.phylo.RDS")))
        pdf(file.path(pipeline_mgr$directories$phylo, paste0(pipeline.parameters$name, ".subcluster.phylo.pdf")))
        plot.phylo(d$tree,direction="downwards", show.node.label=TRUE)
        nodelabels() # For some reason show.node.label==TRUE doesn't work, but this does
        dev.off()

        if (!exists("tsne")) { tsne = readRDS(file.path(pipeline_mgr$directories$tSNE, paste0(pipeline.parameters$name, tsne_xy_extension)))}

        pdf(file.path(pipeline_mgr$directories$tSNE, paste0(pipeline.parameters$name, ".subcluster.tSNE.pdf")))
        ident.plot(tsne[names(idents),], idents, legend=TRUE)
        title(main=paste0(pipeline.parameters$name, " Final Clustering"))
        dev.off()

        # Determine what happened to every cell, and write to file.
        cell_cluster_outcomes = DetermineCellClusterOutcome(pipeline_mgr)
        saveRDS(cell_cluster_outcomes,
            file=file.path(pipeline_mgr$directories$assign, paste0(pipeline_mgr$pipeline.parameters$name, cell_cluster_outcomes_extension)))

        if (!exists("filtered_raw_dge")) filtered_raw_dge = readRDS(filtered_raw_dge_path)
        if (!exists("filtered_scaled_dge")) {filtered_scaled_dge = readRDS(filtered_scaled_dge_path)}
        clean_markers_and_metacells=MakeCleanMarkersAndMetacells(cell_cluster_outcomes=cell_cluster_outcomes,
            idents=readRDS(file.path(pipeline_mgr$directories$assign, paste0(pipeline.parameters$name, cluster_assign_extension))),
            raw_dge=raw_dge,
            filtered_raw_dge=filtered_raw_dge,
            filtered_scaled_dge=filtered_scaled_dge)
        SaveCleanMarkersAndMetacells(pipeline_mgr, clean_markers_and_metacells)
        rm(clean_markers_and_metacells)
        rm(cell_cluster_outcomes)
        stage_mgr$afterStage("idents")
    }

    if (pipeline_mgr$stage_mgr$beforeStage("cluster_annotation")) {

        cluster_class_file = get_single_file_matching_regexp("cluster_class.csv",
                    path=pipeline_mgr$directories$curation_sheets, pattern=glob2rx("*.cluster_class.csv"),
                    full.names=TRUE)
        conn=pipeline_mgr$open_curation_file_with_history(cluster_class_file)
        class.list = read.table(conn, stringsAsFactors=F, sep=",", header=T, quote="\"")
        close(conn)
        # In Evan's code this is 'norm'
        if (!exists("filtered_scaled_dge")) {filtered_scaled_dge = readRDS(filtered_scaled_dge_path)}
        # In Evan's code this is 'dge.master'
        filtered_raw_dge = readRDS(filtered_raw_dge_path)

        idents.round1 = readRDS(paste0(pipeline_mgr$directories$assign, pipeline.parameters$name, cluster_assign_extension))


        idents.all = idents

        metacells = apply(filtered_scaled_dge[,names(idents.all)],1,function(q){tapply(q,idents.all,mean)})
        norm.use = filtered_scaled_dge[,names(idents.all)]


        #Get the most differentially expressed genes for each cluster, and use that full gene list to generate correlations.
        dge.use = filtered_raw_dge[rownames(norm.use),names(idents.all)]
        DEgenes = unique(unlist(lapply(levels(idents.all),function(q){
            cat_with_time(paste0("Computing differential expression for ", q))
            cells.use = names(idents.all)[which(idents.all ==q)]
            cells.comp.all = setdiff(colnames(norm.use),cells.use)
            clust = strsplit(q,split="-")[[1]][1]
            clust.sub=levels(idents.all)[grep(levels(idents.all),pattern=paste0("^", clust,"-"))] # ^ => anchor match to start of string

            cells.comp.sub = setdiff(colnames(norm.use)[which(idents.all %in% clust.sub)],cells.use)
            DEgenes.all = diffexp(norm = norm.use,dge = dge.use,cells = cells.use,cells.comp = cells.comp.all, FOLDch.thresh = 2)
            DEgenes.sub = diffexp(norm = norm.use,dge = dge.use,cells = cells.use,cells.comp = cells.comp.sub, FOLDch.thresh = 2)
            toret= unique(c(as.character(DEgenes.all$GENE),as.character(DEgenes.sub$GENE)))
            idx = which (!is.na(toret))
            return(toret[idx])

        })))

        cors = cor(t(metacells[,DEgenes]))

        d = lapply(1:nlevels(idents.round1),function(q){

            cluster.name = levels(idents.round1)[q]
            subcluster.names = levels(idents.all)[which(strStartsWith(levels(idents.all), paste0(cluster.name, "-")))]
            if (length(subcluster.names) == 0) {
                cat_with_time(paste0(cluster.name, " was not subclustered so no cluster sheet will be created"))
            } else {
                cluster_sheet_file = file.path(pipeline_mgr$directories$blank_cluster_sheets, paste0("Cluster_",cluster.name,".csv"))
                cat_with_time(paste0("Generating cluster sheet ", cluster_sheet_file))
                cluster.number = as.numeric(cluster.name)
                cluster.class = class.list[class.list$cluster_number==cluster.number,]$cluster_class
                cells.subclustered = names(idents.all[idents.all %in% subcluster.names])
                dge.use = filtered_raw_dge[,cells.subclustered]
                dge.use = dge.use[which(rowSums(dge.use)>0),]
                norm.use= filtered_scaled_dge[rownames(dge.use),colnames(dge.use)]

                rows = lapply(subcluster.names,function(sub.name){

                    cells.sub = names(idents.all[idents.all == sub.name])
                    cells.comp = setdiff(colnames(norm.use),cells.sub)
                    DEgenes = data.frame(diffexp(norm.use,dge.use,cells.sub,cells.comp, FOLDch.thresh=2))
                    DEgenes = DEgenes[rowSums(is.na(DEgenes))==0,]
                    if (nrow(DEgenes) > 1) {
                        DEgenes.use = DEgenes$GENE[order(as.vector(DEgenes$FOLDch),decreasing=T)[1:min(c(40,nrow(DEgenes)))] ]
                        idx = which(DEgenes$pct.1 > 0.4 & DEgenes$pct.2 < 0.1)
                        pctgenes.use = DEgenes$GENE[order(DEgenes$pct.2[idx])[1:min(c(40,length(idx)))] ]
                    } else {
                        DEgenes.use = DEgenes$GENE
                        pctgenes.use = DEgenes$GENE
                    }

                    cor.ordered = colnames(cors)[order(cors[sub.name,],decreasing=T)]
                    cor.name = cor.ordered[min(which(!grepl(pattern=paste0(cluster.number,"-"),x = cor.ordered)))]
                    if (is.na(cor.name)) {
                        cor.max = NA
                    } else {
                        cor.max = round(cors[sub.name,cor.name],4)
                    }
                    return(c(tissue=pipeline.parameters$name,
                                cluster_number=sub.name,
                                class=cluster.class,
                                class_marker="",
                                type_marker="",
                                full_name="",
                                common_name="",
                                DE_genes=paste(DEgenes.use,collapse=" "),
                                unique_genes=paste(pctgenes.use,collapse= " "),
                                max_cor_clust=cor.name,
                                max_cor=cor.max,
                                comments=""))
                })
                cluster_sheet = do.call(rbind, rows)
                write.csv(cluster_sheet, file=cluster_sheet_file, quote=TRUE, row.names=FALSE, na="")
            }
        })

        stage_mgr$afterStage()
    }

    if (pipeline_mgr$stage_mgr$beforeStage("differential_expression")) {
        if (!exists("idents")) {
            idents = readRDS(file.path(pipeline_mgr$directories$assign, paste0(pipeline.parameters$name,".subcluster.assign.RDS")))
        }
        if (!exists("filtered_scaled_dge")) {filtered_scaled_dge = readRDS(filtered_scaled_dge_path)}
        if (!exists("filtered_raw_dge")) {filtered_raw_dge = readRDS(filtered_raw_dge_path) }

        pairwise_and_cluster_diffexp = pairwiseAndClusterDiffExpression(idents=idents,
                raw_dge=filtered_raw_dge, scaled_dge=filtered_scaled_dge, only.pos=FALSE,
                FOLDch.thresh=pipeline.parameters$differential_expression$FOLDch.threshold,
                pval.thresh=pipeline.parameters$differential_expression$pval.threshold)

        saveRDS(pairwise_and_cluster_diffexp$pairwise_markers,file=paste0(pipeline.parameters$name, pairwise_markers_extension))
        saveRDS(pairwise_and_cluster_diffexp$cluster_markers, file=paste0(pipeline.parameters$name, cluster_markers_extension))
        rm(pairwise_and_cluster_diffexp)
        stage_mgr$afterStage()
    }
}
