# MIT License
#
# Copyright 2017 Broad Institute
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

### GENE SELECTION #####

# rowMeans is implemented for sparse matrices, but there is no existing implementation row variances.
compute_row_mean_and_variance=function(x) {
    if (is.matrix(x) || is.data.frame(x)) {
        return(list(mean=rowMeans(x), var=apply(x, 1, var  )))
    } else {
        # Assume triplet sparse matrix.  Is there a way to confirm this?
        # From https://en.wikipedia.org/wiki/Standard_deviation#Rapid_calculation_methods
        # Note that the below is not ideal because it may be subject to overflow and rounding error.
        # However, the methods that address that problem require iterating over every value, not just
        # non-zero values.
        n = ncol(x)
        sums = rowSums(x)
        # Awesome idea courtesy of Giulio.  Since rowSums(sparseMatrix) is implemented in C, just sum_squares
        # all the values in the sparse matrix implementation, and then use rowSums to get sum of squares.
        x@x = x@x^2
        sum_squares = rowSums(x)
        means = sums/n
        variances = (n * sum_squares - (sums^2))/(n * (n-1))
        names(means) = rownames(x)
        names(variances)=rownames(x)
        return(list(mean=means, var=variances))
    }
}

mean.var.auto<-function(raw_dge, scaled_dge, alphathresh=NULL,varthresh=NULL,cex.use=0.3) {
    if (is.null(alphathresh)) {
        alphathresh= 0.99
    }
    if (is.null(varthresh)) {
        varthresh = 0.1
    }

    trx_per_cell <- colSums(raw_dge)

    m_v = compute_row_mean_and_variance(scaled_dge)
    gene_expr_mean <- m_v$mean  # Each gene's mean expression level (across all cells)
    gene_expr_var  <- m_v$var  # Each gene's expression variance (across all cells)
    nolan_constant <- mean((1 / trx_per_cell ))
    alphathresh.corrected=alphathresh/dim(raw_dge)[1]
    genemeanupper <- gene_expr_mean+qnorm(1-alphathresh.corrected/2)*sqrt(gene_expr_mean*nolan_constant/dim(raw_dge)[2])
    genes.use=names(gene_expr_var)[which(gene_expr_var/nolan_constant> genemeanupper & log10(gene_expr_var) > log10(gene_expr_mean)+(log10(nolan_constant)+varthresh))]
    plot( log10(gene_expr_mean), log10(gene_expr_var), cex=cex.use
    )

    points(log10(gene_expr_mean[genes.use]),log10(gene_expr_var[genes.use]),cex=cex.use,col='green')
    abline(log10(nolan_constant),1,col='purple')

    legend("bottomright",paste0("Selected genes: ",length(genes.use)),pch=20,col='green')
    return(genes.use)
}

#' pick the numGenes most expressed genes
select_most_expressed_genes<-function(raw_dge, scaled_dge, numGenes) {
    # Filter raw_dge in case some genes have been removed by setup() because they don't have min.cells
    rawDgeSelectedGenes = raw_dge[rownames(scaled_dge), ]
    rs=rowSums(raw_dgeSelectedGenes)
    names(rs) = rownames(rawDgeSelectedGenes)
    sortedRs = sort(rs, decreasing=TRUE)
    return(names(sortedRs)[1:numGenes])
}

#' If is.null(numGenes), invoke mean.var.auto, otherwise pick the numGenes most expressed gened
select_genes<-function(raw_dge, scaled_dge, numGenes=NULL, ...) {
    if (is.null(numGenes)) {
        mean.var.auto(raw_dge=raw_dge, scaled_dge=scaled_dge, ...)
    } else {
        select_most_expressed_genes(raw_dge=raw_dge, scaled_dge=scaled_dge, numGenes=numGenes)
    }
}

##### ICA FUNCTIONS #######

flip_and_reorder_ics<-function(ica.obj) {
    # Calculate skews of cell scores
    cell_skews_initial = apply( ica.obj$S, 2, skewness )

    # Flip certain ICs so that all ICs have positive skew for cell scores
    wh_ICs_flip = which( cell_skews_initial < 0 )
    ica.obj$A[,wh_ICs_flip] = -1 * ica.obj$A[,wh_ICs_flip]
    ica.obj$S[,wh_ICs_flip] = -1 * ica.obj$S[,wh_ICs_flip]

    # Reorder the ICs by their skew (from best to worst)
    IC_reorder = order( abs(cell_skews_initial), decreasing=T )
    ica.obj$A = ica.obj$A[,IC_reorder]
    ica.obj$S = ica.obj$S[,IC_reorder]
    return(ica.obj)
}

run_ica<-function(data.use, n.comp, random.seed=1, data.project=NULL, method="C", verbose=FALSE, flip_and_reorder=TRUE) {
    cat_with_time("Starting fastICA")
    set.seed(random.seed)
    ica.raw=fastICA(X=data.use,n.comp=n.comp,maxit=10000,method=method, verbose=verbose)
    ica.obj<-list(S= ica.raw $S,A=t(ica.raw $A),K=ica.raw$K,W=ica.raw$W)
    rownames(ica.obj$S)<-rownames(data.use)
    rownames(ica.obj$A)<-colnames(data.use)

    cat_with_time("fastICA complete.")
    if (!is.null(data.project)) {
        projected.IC <- mrdivide(t(data.project), t(ica.obj$S))
        projected.IC[!is.finite(projected.IC)]<-0
        ica.obj$A <- rbind(ica.obj$A,projected.IC)
        rownames(ica.obj$A)<-c(colnames(data.use),colnames(data.project))
    }
    if (flip_and_reorder) {
        ica.obj = flip_and_reorder_ics(ica.obj)
    }
    return(ica.obj)
}


#### LOUVAIN CLUSTERING FUNCTIONS ######

louvainCluster <- function (data, min.clust=10,nn = 25, nstart= 100,iter.max=10,jaccard=T,SNN, R=0.1,
            orig.scale,generate.tree=T,reorder.tree=T,reorder.numeric=T, random.seed=1, out_edge_file=NA)

{
    #Kmeans-cluster small cell populations with factors
    set.seed(random.seed)
    id.kmeans=kmeans(data,centers=ncol(data),nstart=50,iter.max=200)$cluster
    names(id.kmeans)<-rownames(data)
    idx.hold<-which(table(id.kmeans) < min.clust)
    cells.hold=rownames(data)[which(id.kmeans %in% idx.hold)]
    id.hold<-id.kmeans[cells.hold]
    cells.cluster=setdiff(rownames(data),cells.hold)
    cat_with_time(paste0("Withheld ", length(cells.hold)," cells from louvain clustering composing ",length(unique(id.hold))," clusters."))
    data.orig=data
    data=data[cells.cluster,]
    cellnames <- rownames(data)
    if(is.data.frame(data)) data <- as.matrix(data)

    n_cell <- nrow(data)

    if (is.null(SNN)) {
        my.knn <- get.knn(as.matrix(data), k=min(c(10*nn,n_cell-1)))

        nn.ranked <- cbind(1:n_cell,my.knn$nn.index[,1:(nn-1)])

        nn.large <- my.knn$nn.index

        #Seurat weights the similarity graph by jaccard distance agreement--that is, the number of NNs that are shared between two cells.  I compared this graph construction to that of euclidean distance-based weighting in the pairwise half analysis and found jaccard performed more reliably.

        if (jaccard) {

            cat_with_time("making jaccard matrix.")

            w <- sparsejaccard(n_cell=n_cell, k_param=nn,nn.large=nn.large, nn.ranked=nn.ranked, update=0.25)

        }
    } else {

        w <- SNN

    }

    #This is code taken from Seurat for preparing data for modularity-based clustering.
    diag(w) <- 0
    w <- as(w,"dgTMatrix")
    edge <- cbind(i = w@i, j = w@j, x = w@x)

    ident=modclust(edge=edge,resolution=R,n.iter= iter.max,n.start=nstart,
                cellnames=rownames(data), random.seed=random.seed, out_edge_file=out_edge_file)
    n=names(ident)
    ident=as.numeric(ident)
    id.hold=max(unique(ident))+id.hold
    ident.all=factor(c(ident,id.hold))
    levels(ident.all)<-1:length(levels(ident.all))
    names(ident.all)<-c(n,cells.hold)
    tree.round1=NULL
    if (generate.tree) {
        tree.round1<-make.dendrogram(data= orig.scale, partition=ident.all,reorder.partition=reorder.tree,reorder.numeric=reorder.numeric,min.clust.size=min.clust)
        ident.all=tree.round1$partition
        tree.round1=tree.round1$tree
    }
    return(list(louvain=ident.all,kmeans=factor(id.kmeans),tree=tree.round1	))
}

#' @return a list like returned by louvainCluster, but with all cells in one cluster.
makeDummyLouvainCluster=function(cell_barcodes) {
    ident.all = factor(rep(1, length(cell_barcodes)))
    names(ident.all) = cell_barcodes
    return(list(louvain=ident.all,kmeans=factor(),tree=NULL))
}


validate_subcluster_method=function(method) {
    if (!is.element(method, subcluster_methods)) {
        stop(sprintf("Invalid cluster.method %s.  Choose %s", method, paste(subcluster_methods, collapse=" or ")))
    }
}

#' @param resolution must be specified if method == LOUVAIN
#' @param edge make be pre-computed with makeLouvainEdgeGraph() in order to speed up multiple invocations with different resolutions
clusterRound2<-function(data,method,ica.obj,norm,dge, orig.scale ,resolution=NULL,random.seed=1,
                        file=paste0(samplename,method,"clusterreport.pdf"), tolFun=NULL, gmm.model.threshold=0,
                        edge=NULL, nn=default_subcluster_louvain_nn, out_edge_file=NA) {

    validate_subcluster_method(method)
    if (ncol(as.matrix(ica.obj$S)) > 0) {
        if (!is.null(ica.obj$celllist)) {

            # If clusters were suggested, find multivariate centers and rbind them together into a matrix for kmeans initiation.

            centers = as.matrix(do.call(rbind,lapply(ica.obj$celllist,function(q){
                do.call(rbind,lapply(q,function(r){
                    x=colMeans(as.matrix(ica.obj$S[r,]))

                }))

            })))

            num_cols = ncol(centers)
            centers = centers[!rowSums(!is.finite(centers)),]
            if (!is.matrix(centers)) {
                # this can happen if the row selection above results in only one row
                if (!is.vector(centers)) {stop(paste0("Strange class(centers) ", class(centers)))}
                centers = matrix(centers, ncol=num_cols)
            }
            cat_with_time("CENTERS:")
            cat_with_time(centers)

        } else {
            centers = NULL
        }
        # Generate a tSNE if it wasn't done already (usually not done because too few ICs.. in this case just use the scaled data)
        if(is.null(ica.obj$tSNE)) {
            ica.obj$tSNE = do_tsne(orig.scale[rownames(data),], random.seed=random.seed)
            rownames(ica.obj$tSNE)<-rownames(data)
        }


        pdf(file)

         # There is an option to do Louvain or GMM-based clustering.

        if (method == louvain_subcluster_method) {
            if (ncol(data) == 1 && !is.null(ica.obj$celllist)) {
                idents=list(louvain=subcluster_from_single_ic(data, ica.obj$celllist), tree=NULL)

            } else {
                if (is.null(resolution) || is.na(resolution)) stop("Resolution must be specified for Louvain subclustering")
                idents = louvainClusterRound2(data, orig.scale= orig.scale,edge=edge, resolution=resolution, nn=nn,
                        out_edge_file=out_edge_file)
            }

        } else if (method == gmm_subcluster_method) {
            if (nrow(centers) > 1) {
                idents = GMMClusterRound2(data,centers=centers,orig.scale=orig.scale,tsne=ica.obj$tSNE, tolFun=tolFun, gmm.model.threshold=gmm.model.threshold)
            } else {
                return(list(idents=make_null_subclustering_idents(rownames(ica.obj$S), name=gmm_idents_label)))
            }
        }

        #Check how far a cluster center has deviated from the prior.  Maybe this can be a way of removing cluster centers that are poor (as in they migrate so far from an IC center that they are no longer describing that IC)
        centers.clustered = apply(as.matrix(ica.obj$S[names(idents$louvain),]),2,function(q){tapply(q,idents$louvain,mean)})

        # Handle degenerate case in which there is a single cluster found, which causes centers.clustered not to be a matrix.
        if (!is.matrix(centers.clustered)) {
            centers.clustered = matrix(centers.clustered, ncol=length(centers.clustered))
        }

        cors.centers = cor(t(centers.clustered),t(centers))
        idents[[length(idents)+1]] <-cors.centers
        names(idents)[length(idents)] = "cors.centers"

        # Generate an identity plot for this cluster.
        par(mfrow = c(1,1))
        ident.plot(ica.obj$tSNE[names(idents$louvain),],idents$louvain)
        title(main = paste0("Final ",method," clustering"))

        x = clusterreport(idents = idents$louvain,tree = idents$tree,tsne = ica.obj$tSNE,cell.matrix = ica.obj$S,gene.matrix = t(ica.obj$A),norm = norm,dge = dge,ic.labels=ica.obj$oldlabels)
        dev.off()

        return(idents)
    } else {
        cat_with_time("No ICs selected for this cluster.  No subclustering performed.")

        return(list(idents=make_null_subclustering_idents(rownames(ica.obj$S), name=if (method==gmm_subcluster_method) gmm_idents_label else louvain_idents_label)))
    }

}

subcluster_from_single_ic=function(data, cell.list) {
    positive_cells = cell.list[[1]][[1]]
    negative_cells = setdiff(rownames(data), positive_cells)
    idents = factor(c(rep(1, length(positive_cells)), rep(2, length(negative_cells))))
    names(idents) = c(positive_cells, negative_cells)
    return(idents)
}

#' @return factors with all cells in the same cluster
make_null_subclustering_idents<-function(cell_names, name) {
    id.none = factor(rep(1,times = length(cell_names)))
    names(id.none) = cell_names
    ret=list(id.none)
    names(ret) = name
    return(ret)
}

# Assumes current directory is where matlab output is.
process_GMMoutput<-function(soft, BIC, loglik, sigma, lambda, mu, thresh,cells.use) {
    rownames(soft) = cells.use

    max=apply(soft,1,max)
    qthresh=quantile(max,thresh)
    idx=which(max >= qthresh)

    hard<-factor(apply(soft,1,which.max))

    names(hard)<-cells.use
    #AIC=read.table("AIC.txt",sep=",",header=F,stringsAsFactors=F)
    model=list(mu=as.matrix(mu),sigma=as.matrix(sigma),lambda=as.numeric(lambda[,1]))
    idents=list(hard=hard,soft=soft)
    IC=list(BIC=BIC,loglik=loglik)
    return(list(idents=idents,IC=IC,model=model))

}



GMMClusterRound2<-function(data,centers,generate.tree=T, reorder.tree=T,reorder.numeric=T,orig.scale,random.seed=1,
                            min.clust=10,tsne, tolFun=NULL, gmm.model.threshold=0) {
    cat_with_time("Generating k-means prior...")
    set.seed(random.seed)

    #	id.prior=factor(kmeans(data,centers=centers,nstart=50,iter.max=200)$cluster)
    #id.prior=factor(kmeans(data,centers=centers,nstart = )$cluster)
    dists = as.matrix(pdist(data,centers))
                  id.prior = factor(apply(dists,1,function(q){order(q)[1]}))

    names(id.prior)<-rownames(data)
    levels(id.prior)<-1:length(levels(id.prior))
    # Plot the prior generated by kmeans.
    ident.plot(tsne[names(id.prior),],id.prior)
    n=length(levels(id.prior))
    title(main=paste0("KDE prior", n," centers"))

    cat_with_time("Starting Gaussian Mixture Model Clustering.")
    data = as.matrix(data[names(id.prior),])
    gmm_outputs = matlab_gmm_cluster(data, id.prior, tolFun=tolFun)

    output<-process_GMMoutput(soft=gmm_outputs$soft, BIC=gmm_outputs$BIC, loglik=gmm_outputs$loglik,
                        sigma=gmm_outputs$sigma, lambda=gmm_outputs$lambda, mu=gmm_outputs$mu,
                        thresh=gmm.model.threshold,cells.use= rownames(data))

    model=output$model
    idents=output$idents$hard
    tree.round1=NULL
    levels.use= names(table(idents))[which(table(idents) > min.clust)]
    if (length(levels.use) < 3) {
        generate.tree=F
    }
    if (generate.tree) {
	tree.round1<-make.dendrogram(data = orig.scale, partition = idents,reorder.partition=reorder.tree,reorder.numeric=reorder.numeric,min.clust.size=min.clust)
	idents=tree.round1$partition
	tree.round1=tree.round1$tree
    }
    return(list(idents=idents,tree=tree.round1,model=model, soft.model=gmm_outputs$soft))

}



makeLouvainEdgeGraph=function(data, nn) {
    if(is.data.frame(data)) data <- as.matrix(data)

    n_cell <- nrow(data)
    my.knn <- get.knn(as.matrix(data), k=min(c(10*nn,n_cell-1)))
    # Can't get more nearest neighbors than their are cells.
    nn = min(nn, nrow(data))
    nn.ranked <- cbind(1:n_cell,my.knn$nn.index[,1:(nn-1)])
    nn.large <- my.knn$nn.index

        #Seurat weights the similarity graph by jaccard distance agreement--that is, the number of NNs that are shared between two cells.  I compared this graph construction to that of euclidean distance-based weighting in the pairwise half analysis and found jaccard performed more reliably.
    cat_with_time("making jaccard matrix.")
    w <- sparsejaccard(n_cell=n_cell, k_param=nn,nn.large=nn.large, nn.ranked=nn.ranked, update=0.25)

    #This is code taken from Seurat for preparing data for modularity-based clustering.
    diag(w) <- 0
    w <- as(w,"dgTMatrix")
    edge <- cbind(i = w@i, j = w@j, x = w@x)
    return(edge)
}

louvainClusterRound2 <- function (data, min.clust=10,nn = default_subcluster_louvain_nn,
    edge=NULL,orig.scale,generate.tree=T,reorder.tree=T,reorder.numeric=T,random.seed=1,resolution = 0.1,
        out_edge_file=NA)

{
    if (is.null(edge)) {
        edge = makeLouvainEdgeGraph(data, nn)
    }
    #if (!is.null(nrow(centers))) {
    #    centers = nrow(centers)
    #}
    #ident <- getkclusters(edge,modularity=1,K=centers,maxiter=100,tolerance=0,r.range=c(0.01,15),cellnames=rownames(data),algorithm = 1)
    ident = modclust(edge, modularity=1,resolution=resolution,n.start=100,n.iter=25,random.seed=random.seed,algorithm=1,
                rownames(data), out_edge_file=out_edge_file)
    n=names(ident)
    ident.all=factor(as.numeric(ident))
    names(ident.all)<-n
    tree.round1=NULL
    if (generate.tree) {
        tree.round1<-make.dendrogram(data= orig.scale, partition=ident.all,reorder.partition=reorder.tree,reorder.numeric=reorder.numeric,min.clust.size=min.clust)
        ident.all=tree.round1$partition
        tree.round1=tree.round1$tree
    }
    droppedSingletons = setdiff(rownames(data), names(ident.all))
    return(list(louvain=ident.all,tree=tree.round1, droppedSingletons=droppedSingletons))
}



sparsejaccard = function(n_cell, k_param, nn.large, nn.ranked, prune.SNN=0, update){
    options(scipen=999)
    counter = 1
    idx1 = vector(mode="integer", length = n_cell^2/k_param)
    idx2 = vector(mode="integer", length = n_cell^2/k_param)
    edge_weight = vector(mode="double", length = n_cell^2/k_param)
    id = 1

    #fill out the adjacency matrix w with edge weights only between your target cell and its 10*k_param-nearest neighbors
    #speed things up (don't have to calculate all pairwise distances)
    #define the edge weights with Jaccard distance
    for (i in 1:n_cell){
        for (j in 1:ncol(nn.large)){
            s=intersect(nn.ranked[i,],nn.ranked[nn.large[i,j],])
            u=union(nn.ranked[i,],nn.ranked[nn.large[i,j],])
            e = length(s)/length(u)
            if(e>prune.SNN){
                idx1[id]=i
                idx2[id]= nn.large[i,j]
                edge_weight[id] = e
                id = id + 1
            }
        }
        if (i==round(counter*n_cell*update)) {
            cat_with_time(paste("SNN : processed ", i, " cells", sep=""))
            counter= counter+1;
        }
    }
    idx1 = idx1[!is.na(idx1) & idx1!=0]; idx2 = idx2[!is.na(idx2) & idx2!=0]; edge_weight = edge_weight[!is.na(edge_weight)& edge_weight!=0]
    w = sparseMatrix(i=idx1, j=idx2, x=edge_weight, dims=c(n_cell, n_cell))
    diag(w) = 1
    return(w)
}

getkclusters <- function(edge,modularity=1,K=centers,maxiter=100,tolerance=0,r.range=c(0.01,15),mod.restart = 100,mod.iter=10,cellnames,algorithm = 1) {

    tf <- TRUE
    r.orig=r.range
    tol <- tolerance*K

    newr <- r.range[2]

    #This is the wrapper function for modularity clustering with the java implementation.
    idents <- modclust(edge=edge, modularity=modularity,resolution=newr,n.iter=mod.iter,n.start=mod.restart,cellnames=cellnames,algorithm = algorithm)

    N <- length(levels(idents))

    i <- 1

    while(tf & i < maxiter) {
        if (i == round(maxiter/2,0)) {
            r.range=r.orig
            r.range[1] = 0.7 * r.range[1]
            r.range[2] = 1.2 * r.range[2]
            newr=r.range[2]
            idents <- modclust(edge=edge, modularity=modularity,resolution=newr,n.iter=mod.iter,n.start=mod.restart,cellnames=cellnames,algorithm = algorithm)
            N <- length(levels(idents))
        }
        if ((K-N) > 0) {

            r.range[1] <- newr
            newr <- (newr + r.range[2])/2

            cat_with_time(paste0("Retrying clustering with resolution of ", newr))

            idents <- modclust(edge=edge, modularity=modularity,resolution=newr,n.iter=mod.iter,n.start=mod.restart,cellnames=cellnames,algorithm = algorithm)

            N <- length(levels(idents))

            cat_with_time(paste0("Number of clusters detected: ", N))

        }	else {
            r.range[2] <- newr

            newr <- (newr + r.range[1]) / 2

            cat_with_time(paste0("Retrying clustering with resolution of ", newr))

            idents <- modclust(edge=edge, modularity=modularity,resolution=newr,n.iter=mod.iter,n.start=10,cellnames=cellnames,algorithm = algorithm)

            N <- length(levels(idents))

            cat_with_time(paste0("Number of clusters detected: ", N))

        }

        test <- abs(K-N)

        if(test <= tol) {

            tf <- FALSE
        }

    	cat_with_time(paste0("Iteration number ", i, " complete."))
        i <- i+1

    }
    if (i == maxiter) {

        cat_with_time(paste0("Could not achieve clustering with specified K of ", K," within ",maxiter, " iterations.  Returning the last clustering output."))

        return(idents)

    }


    return(idents)
}

modclust <- function(edge, modularity=1,resolution,n.start=100,n.iter=25,random.seed=1,algorithm=1,
        cellnames, out_edge_file=NA){

    ModularityJarFile=get_jar_path("ModularityOptimizer")

    edge_file <- if (is.na(out_edge_file)) tempfile("edge", fileext=".txt") else out_edge_file
    # Make sure no scientific notation written to edge file
    saveScipen=options(scipen=10000)[[1]]
    write.table(x=edge,file=edge_file,sep="\t",row.names=F,col.names=F)
    options(scipen=saveScipen)
    output_file <- tempfile("louvain.out", fileext=".txt")
    if (is.null(random.seed)) {
        # NULL random.seed disallowed for this program.
        random.seed = 0
    }

    command <- paste("java -jar", ModularityJarFile, edge_file, output_file,modularity, resolution, algorithm, n.start,
                        n.iter, random.seed, 0, sep = " ")
    ret = system(command, wait = TRUE)
    if (ret != 0) {
        stop(paste0(ret, " exit status from ", command))
    }
    if (is.na(out_edge_file)) {
        unlink(edge_file)
    }
    ident.use <- factor(read.table(file = output_file, header = FALSE, sep = "\t")[, 1])
    unlink(output_file)
    tab <- table(ident.use)
    names(ident.use)<-cellnames

    #In Seurat, singleton cells are grouped into other clusters by similarity.  This takes a lot of computational time, and I'm not sure it's the right move, since
    levels.keep <- levels(ident.use)[which(tab > 1)]
    cells.keep <- names(ident.use)[which(ident.use %in% levels.keep)]
    ident.use <- ident.use[cells.keep]
    ident.use <- droplevels(ident.use)
    return(ident.use)
}



#' Find ICs and subcluster a cluster
#' @param scaled_gene_selected_dge scaled DGE for this cluster, with only the 50% largest cells, and mean-var selected genes
#' @param raw_dge same cells as scaled_gene_selected_dge, but raw expression counts, and not gene selected
#' @param n.components Number of ICs to extract
#' @param cluster.dir Target directory for most of the file outputs of this function
#' @param components.dir Target directory for the curated ICs output file
#' @param blank.curation.sheets.dir where to write a blank curation sheet for this cluster.
ica_and_curate=function(scaled_gene_selected_dge, raw_dge,
                n.components, cluster.dir, components.dir,
                blank.curation.sheets.dir, cluster.num, file_prefix,
                ica.method, ica.verbose, random.seed=1) {
    ica.obj=run_ica(data.use=t(scaled_gene_selected_dge), n.comp=n.components, random.seed=random.seed,
            method=ica.method, verbose=ica.verbose)

    tsne.sub=do_tsne(ica.obj$S,pca=F, random.seed=random.seed)
    rownames(tsne.sub)<-colnames(scaled_gene_selected_dge)
    saveRDS(tsne.sub,paste0(cluster.dir,file_prefix, cluster_tSNExy_extension))
    saveRDS(ica.obj,file=paste0(cluster.dir,file_prefix, "icaobjRAW.RDS"))
    gene_loadings = ica.obj$A
    cell_rotations = ica.obj$S
    saveRDS(list(gene_loadings=gene_loadings, cell_rotations=cell_rotations, name= "ICA", tsne=tsne.sub),
        file=file.path(components.dir, paste0(file_prefix, ica_file_extension)))
    make_new_curation_sheet(make_curation_sheet_path(blank.curation.sheets.dir, cluster.num),
                    gene_loadings, nrow(cell_rotations))

    #In this second round of curation I make suggestions in the pdf about which ICs to keep.
    ica.curated <-CurateICAOne(ica.obj,raw_dge,
        file=paste0(cluster.dir,file_prefix, "AUTOCURATED_SubclustICAreport.pdf"),
            featureplot=T,tsrots=tsne.sub)
}

### ANNOTATION / REPORT FUNCTIONS   #####
CurateICAOne = function(ica.data, dge, file="review_ICA.pdf", skewness_threshold = 1.1,kurtosis_threshold = 2.1, featureplot=FALSE,tsrots=NULL) {

    x = list(A=ica.data$A, S=ica.data$S)

    n_ICs = dim(x$A)[2]
    if (!is.null(file)) {
    }

    # Calculate cell-kurtosis(ck) and cell-skew(cs) for the ICs
    ck = apply( x$S, 2, kurtosis )
    cs = apply( x$S, 2, skewness )
    #  This is a "quick and dirty" way to flag technical ICs.. high skew ICs tend to be true biological signal, as do ICs with ultra low kurtosis
    keep_IC=(cs > skewness_threshold | ck < kurtosis_threshold)
    if (!any(keep_IC)) {
        # Loosen threshold if all ICs are eliminated
        keep_IC=(cs >= skewness_threshold | ck <= kurtosis_threshold)
    }
    if (!any(keep_IC)) {
        warning("Skewness and kurtosis thresholds have eliminated all ICs")
    }

    if (!is.null(file)) {
        pdf( file )
        # Calculate gene-kurtosis(gk), gene-skew(gs) for the ICs
        gk = apply( x$A, 2, kurtosis )
        gs = apply( x$A, 2, skewness )
        cs.use=colSums(dge[,rownames(x$S)])
        corbysize=apply(x$S,2,function(i){
            cor(i,cs.use)

        })
        # Skew-kurtosis plot of the ICs
        par(mfrow=c(1,1))
        plot(  log(ck,2), cs , col="white",
        xlab = "log2( Cell-score kurtosis)", ylab="Cell-score skew",
        main = "Cell-score skew and kurtosis (after component reordering)")
        rect( 0, 0, log2(kurtosis_threshold), skewness_threshold, col="lightgray", border=NA )

        # Deal with situation when there are no ICs removed.
        # TODO: Is this condition ever true?
        if (length(keep_IC) < length (gk)) {
            text( log(ck[!keep_IC], 2), cs[!keep_IC], which(!keep_IC) , col="red"   )
        }

        if (any(keep_IC)) {
            text( log(ck[ keep_IC], 2), cs[ keep_IC], which( keep_IC)   , col="black" )
        }
        if (any(!keep_IC)) {
            text( log(ck[ !keep_IC], 2), cs[ !keep_IC], which( !keep_IC)   , col="red" )
        }
        abline(v=0); abline(h=0)

        #  Review the cell-score and gene-loading distributions for each component

        for(i in 1:n_ICs) {
            par(mfrow=c(2,1))
            cat_with_time(paste0("Curating IC ", i))
            top_genes = row.names( x$A )[ order(x$A[,i], decreasing=T )[1:7] ]
            top_genes_textstring = paste(top_genes, collapse=", ")
            stats_textstring1 = paste( "k=", round(ck[i],1), ", s=", round(cs[i],1),"  Size correlation = ",round(corbysize[i],2) )
            stats_textstring2 = paste( "k=", round(gk[i],1), ", s=", round(gs[i],1))
            component_textstring = paste( "Component", i )
            if( !keep_IC[i] ) component_textstring=paste(component_textstring, "(REMOVE)")
            plot_title1 = paste( component_textstring, "\n", top_genes_textstring, "\n", stats_textstring1 )
            plot_title2 = paste( component_textstring, "\n", top_genes_textstring, "\n", stats_textstring2 )

            # Plot cell scores
            plot( x$S[,i], cex=0.2, main=plot_title1, xlab="Cell", ylab="Score"  )
            # Plot gene loadings
            plot( x$A[,i], cex=0.2, main=plot_title2, xlab="Gene", ylab="Loading"  )
            if (featureplot) {
                par(mfrow=c(1,1))
                fplot(tsrots[rownames(x$S),],x$S[,i],title=component_textstring)
            }
        }

        # Skew-skew plot of the ICs
        par(mfrow=c(1,1))
        plot(  gs, cs , col="white",
        xlab = "Gene-loading skew", ylab="Cell-score skew",
        main = "Cell-score and gene-loading skew (after component reordering)")
        if (length(keep_IC) < length (gk)) {

            text( gs[!keep_IC], cs[!keep_IC], which(!keep_IC) , col="red"   )

        }
        if (any(keep_IC)) {
            text( gs[ keep_IC], cs[ keep_IC], which(keep_IC)   , col="black" )
        }
        if (any(!keep_IC)) {
            text( gs[ !keep_IC], cs[ !keep_IC], which(!keep_IC)   , col="red" )
        }
        abline(v=0); abline(h=0)
        dev.off()
    }


    # Drop the dodgy ICs before returning
    return(list(S=x$S[ , which(keep_IC) ],A=x$A[ , which(keep_IC) ],K=ica.data$K,W=ica.data$W))

}



ICAprojectgenes<-function(ICA.cellmatrix,ICA.genematrix,data.project,genes.project=10000) {

    cells.use=rownames(ICA.cellmatrix)
    idx=order(rowSums(data.project[, cells.use]),decreasing=T)[1:genes.project]
    genes.use=rownames(data.project)[idx]
    data.use= data.project[genes.use,cells.use]

    scaled=t(scale(t(data.use),center=T,scale=T))
    projected=mrdivide(scaled,t(ICA.cellmatrix))
    toadd=setdiff(rownames(ICA.genematrix),rownames(projected))
    rbind(projected, ICA.genematrix[toadd,])


}

#' @param unordered.tsne Ideally ICs would be reordered before invoking t-SNE in all cases, but we have lots
#'        of legacy output for which that is not the case, so in order to maintain reproducibility, if all ICs are being
#'        kept and unordered.tsne is not NULL, use that rather than invoking t-SNE again.
CurateICAManual = function(ica.data, curation.data, norm=NULL, genes.project=0,file="manualICAreport.pdf",
    cluster.doublets=T,doubletalpha=0.01,featureplot=T, peakthresh = 0.15,peakpos=0.9,centeralpha=0.004,
    min.cells.ic = 3, suggest.centers=T,nullthresh=0.7,random.seed=1,mi.limit=0.15, unordered.tsne=NULL) {

    x = list(gene_loadings=ica.data$gene_loadings,cell_rotations=ica.data$cell_rotations)

    n_ICs = ncol(x$gene_loadings)


    # If desired, additional genes onto the ICA space

    if(genes.project > 0) {
        if (is.null(norm)) {
            stop("norm should not be NULL if genes.project > 0")
        }
	    x$gene_loadings<-ICAprojectgenes(x$cell_rotations,x$gene_loadings,norm,genes.project)
    }

    pdf( file )

    # Calculate cell-kurtosis(k), gene-skew(gs), and cell-skew(cs) for the reordered ICs
    gk = apply( x$gene_loadings, 2, kurtosis )
    ck = apply( x$cell_rotations, 2, kurtosis )
    gs = apply( x$gene_loadings, 2, skewness )
    cs = apply( x$cell_rotations, 2, skewness )


    #This column of the curation data flags whether we feel the IC is biologically meaningful, a doublet, an outlier, or a technical artifact.
    curation.names = curation.data$status


    # Routine for 1-D clustering to remove Doublet and Outlier cells.
    if(cluster.doublets) {


        idx.doublet=sort(grep(curation.names,pattern="Doublet",ignore.case=T))
        idx.outlier=sort(grep(curation.names,pattern="Outlier",ignore.case=T))
        idx.doublet_and_outlier = sort(c(idx.doublet, idx.outlier))

        #Obtain the mode for centering the 1-D gaussian using the density() function, and the turnpoints function from pastecs

        get_doublet_outlier_params = function(i) {
            q=x$cell_rotations[,i]
            d = density(q)
            ts_y = ts(d$y)
            tp = turnpoints(ts_y)
            idx = which.max(d$y[tp$tppos])
            return(list(ic=i, center=d$x[tp$tppos[idx]], sd=sd(q)))
        }

        find_doublets_or_outliers = function(param_list){
            i=param_list$ic
            q=x$cell_rotations[,i]
            center=param_list$center
            sd=param_list$sd
            pvalsUpper=pnorm(q,mean=center,sd=sd,lower.tail=F)
            pvalsUpper=p.adjust(pvalsUpper,"fdr")
            idx=which(pvalsUpper <= doubletalpha)
            return(idx)
        }

        # compute the parameters for thresholding in advance, so they can be captured and returned.
        doublet_params = lapply(idx.doublet, get_doublet_outlier_params)
        outlier_params = lapply(idx.outlier, get_doublet_outlier_params)
        # keep track of the doublets and outliers separately for each IC, so they can be plotted in red in the corresponding IC
        doublets_by_ic = lapply(doublet_params, find_doublets_or_outliers)
        names(doublets_by_ic) = idx.doublet
        outliers_by_ic = lapply(outlier_params, find_doublets_or_outliers)
        names(outliers_by_ic) = idx.outlier
        doublet_outlier_thresholds = do.call(rbind.data.frame, append(doublet_params, outlier_params))
        rownames(doublet_outlier_thresholds) = doublet_outlier_thresholds$ic

    } else {
    	idx.doublet=integer()
        idx.outlier = integer()
        doublets = character()
        outliers = character()
        doublets_by_ic=list()
        outliers_by_ic=list()
        doublet_outlier_thresholds = NA
    }
    doublets_and_outliers_by_ic = append(doublets_by_ic, outliers_by_ic)
    doublets = unlist(doublets_by_ic)
    outliers = unlist(outliers_by_ic)
    doublets_and_outliers = unlist(doublets_and_outliers_by_ic)
    idx.doublet_and_outlier = c(idx.doublet, idx.outlier)

    keep_IC=(curation.data$use_for_clustering =="Y" & !(grepl(curation.names,pattern="Doublet",ignore.case=T) | grepl(curation.names,pattern="Outlier",ignore.case=T)))

    cat_with_time(keep_IC)
    idx.keep=which(keep_IC)
    cells.keep=setdiff(1:nrow(x$cell_rotations),doublets_and_outliers)
    if (suggest.centers & length(idx.keep) > 0) {
	    cell.list=suggest.centers(ICAcell.matrix = as.matrix(x$cell_rotations[cells.keep,idx.keep]), peakthresh = peakthresh, peakpos = peakpos, centeralpha= centeralpha, nullthresh=nullthresh, min.cells.ic = min.cells.ic, idx.keep)
    } else {
  	    cell.list=NULL
    }


    # Option to make a new tSNE map based on only included ICs for use in visualizing IC loadings / clustering.
    if(featureplot & length(which(keep_IC)) >=  2) {
        if (!is.null(unordered.tsne) && length(which(keep_IC)) == ncol(x$cell_rotations)) {
            # Don't re-run t-SNE with ICs in different order if keeping all ICs and a t-SNE result is passed in,
            # in order to maintain compatibility with legacy outputs in which t-SNE is run before reordering ICs.
            tsrots = unordered.tsne
        } else {
            tsrots=do_tsne(x$cell_rotations[,keep_IC], random.seed=random.seed)
            rownames(tsrots)<-rownames(x$cell_rotations)
        }
    }  else {
        featureplot=FALSE
        tsrots=NULL
    }

    #  Review the cell-score and gene-loading distributions for each component

    for(i in 1:n_ICs) {
  	    par(mfrow=c(2,1))
        cat_with_time(i)
        top_genes = row.names( x$gene_loadings )[ order(x$gene_loadings[,i], decreasing=T )[1:7] ]
        top_genes_textstring = paste(top_genes, collapse=", ")
        stats_textstring1 = paste( "k=", round(ck[i],1), ", s=", round(cs[i],1))
        stats_textstring2 = paste( "k=", round(gk[i],1), ", s=", round(gs[i],1))
        component_textstring = paste( "Component", i, "  (", curation.names[i],")" )
        if( !keep_IC[i] ) component_textstring=paste(component_textstring, "(REMOVE)" )
        plot_title1 = paste( component_textstring, "\n", top_genes_textstring, "\n", stats_textstring1 )
        plot_title2 = paste( component_textstring, "\n", top_genes_textstring, "\n", stats_textstring2 )

        # Plot cell scores
        plot( x$cell_rotations[,i], cex=0.2, main=plot_title1, xlab="Cell", ylab="Score"  )

        # Highlight the doublet- and outlier-flagged cells.
        if (i %in% idx.doublet_and_outlier) {
            idx.use = as.character(i)
            points(doublets_and_outliers_by_ic[[idx.use]],x$cell_rotations[doublets_and_outliers_by_ic[[idx.use]],i],cex=0.5,col='red')
        }

        # Highlight the cells used to suggest cluster centers.
        if (i %in% idx.keep & suggest.centers) {
            idx.use = which(idx.keep==i)
            cols = rainbow(length(cell.list[[idx.use]]))
            for (j in 1:length(cell.list[[idx.use]]))
                points(which(rownames(x$cell_rotations) %in% cell.list[[idx.use]][[j]]),x$cell_rotations[cell.list[[idx.use]][[j]],i],cex=0.3,col=cols[j])

        }


        # Plot gene loadings
        plot( x$gene_loadings[,i], cex=0.2, main=plot_title2, xlab="Gene", ylab="Loading"  )

        # Feature plot ICs.
        if (featureplot) {
  	    par(mfrow=c(1,1))
            fplot(tsrots[rownames(x$cell_rotations),],x$cell_rotations[,i],title=component_textstring)
        }
    }

    if (length(idx.keep) > 2) {
        data.use=as.matrix(x$cell_rotations[cells.keep , idx.keep ])
        mut_info<-muti(data.use, verbose=FALSE)
        diag(mut_info)<-0
        rownames(mut_info)<-idx.keep
        colnames(mut_info)<-idx.keep
        par(oma=c(0,0,2,0))

        corrplot(mut_info,cl.lim=c(0,min(c(1,2*max(mut_info)))),is.corr=F,method="number",number.cex=0.6,  title="Mutual Information, IC pairs",mar=c(0,0,1,0))
        par(oma=c(0,0,0,0))

        idx=as.matrix(which(lower.tri(mut_info),arr.ind=T))

        plot(idx.keep [idx[,1] ],mut_info[idx],pch=20,cex=0.0001,ylab="mutual_info",ylim=c(0,min(c(1,2*max(mut_info)))),xlab="First IC in Pair",main="Mutual Information, IC pairs")
        text(idx.keep [idx[,1] ],mut_info[idx],labels= paste0(idx.keep[idx[,1]],",",idx.keep [idx[,2]]),cex=0.4)
        abline(h= mi.limit,lty=2,col='blue')
        idx=as.matrix(which((lower.tri(mut_info) & mut_info > mi.limit),arr.ind=T))
        if (length(idx) > 0){
            for (i in 1:nrow(idx)) {
                plot(data.use[,idx[i,1] ],data.use[,idx[i,2] ],pch=20,cex=0.5,xlab=paste0("IC",idx.keep[idx[i,1] ]),ylab=paste0("IC",idx.keep[idx[i,2] ]))
                title(main=paste0("IC",idx.keep[idx[i,1] ]," v IC",idx.keep[idx[i,2] ],"  mi = ",round(mut_info[idx[i,1],idx[i,2] ],3)))
            }
        }
    }
    dev.off()

    # Drop the dodgy ICs before returning
    toret=list(S=as.matrix(x$cell_rotations[cells.keep , idx.keep ]),A=as.matrix(x$gene_loadings[, idx.keep ]),tSNE=tsrots,
                celllist=cell.list,oldlabels=idx.keep, doublets=rownames(x$cell_rotations)[doublets],
                outliers=rownames(x$cell_rotations)[outliers], doublet_outlier_thresholds=doublet_outlier_thresholds)
    return(toret)

}

#' @return NULL if curation sheet has a problem
readAndValidateCurationSheet=function(curation.file, maxIcs, pipeline_mgr) {
    conn = pipeline_mgr$open_curation_file_with_history(curation.file)
    curation.data = read.table(conn, stringsAsFactors=F, sep=",", header=T, quote="\"")
    close(conn)
    if (max(curation.data$ic.number) != maxIcs) {
        cat_with_time(sprintf("Skipping %s because max(curation.data$ic.number)(%d) != number of ICs(%d)",
            curation.file, max(curation.data$ic.number), maxIcs))
        return(NULL)
    } else {
        return(curation.data)
    }
}

## This function generates a list of cell barcodes from which to generate centers that will be suggested to the GMM kmeans prior.
## Peakthresh: How high a mode has to be to be flagged.  # Peakpos: When more than one peak is present, the minimum displacement from zero for a peak to be considered for clustering.
# min.cells.ic : The minimum number of cells that have to be called as "positive" for that IC in order for those cells to seed a cluster center.
suggest.centers <- function(ICAcell.matrix, peakthresh = 0.15, peakpos = 0.9, centeralpha=0.1, nullthresh=0.7,min.cells.ic = 5, idx.keep) {
    cell.list=lapply(1:ncol(ICAcell.matrix),function(q){
        ic = ICAcell.matrix[,q]
        cat_with_time(paste0("IC",idx.keep[q]))
        # Obtain the modes from each IC.
        d = density(ic)
        plot(d, main = paste0("IC",idx.keep[q]))
        ts_y = ts(d$y)
        tp = turnpoints(ts_y)

        if (tp$firstispeak) {
            peak.pos = tp$tppos[((1:length(tp$tppos) %% 2) ==1)]
            #pit.pos =  tp$tppos[((1:length(tp$tppos) %% 2) ==0)]
        } else {
            peak.pos = tp$tppos[((1:length(tp$tppos) %% 2) ==0)]
            #pit.pos =  tp$tppos[((1:length(tp$tppos) %% 2) ==1)]
        }

        idx = min(which(d$y[peak.pos] > peakthresh))
        center.use =d$x[peak.pos[idx]]
        points(center.use,d$y[peak.pos[idx]],col='red')

        #For each center, use the left trough of each peak to identify a reasonable range of the distribution from which to calculate the variance.
        low.idx = max(which(d$y[tp$tppos] < (peakpos*d$y[peak.pos[idx]]) & d$x[tp$tppos] < center.use))
        cat_with_time("PEAKPOS")
        cat_with_time(peakpos)
        cat_with_time("PEAKPOSIDX:")
        cat_with_time(d$y[peak.pos[idx]])
        cat_with_time(peakpos*d$y[peak.pos[idx]])

        if(!is.finite(low.idx)) {
            low.idx = min(ic)
        } else {
            low.idx=d$x[tp$tppos[low.idx]]
        }

        abline(v = low.idx,lty = 2,col ='red')
        ic.use = ic[which(ic < center.use & ic > low.idx)]
        ic.reflected = 2*center.use - ic.use

        pvalsUpper=pnorm(ic,mean=center.use,sd= sd(c(ic.use,ic.reflected)),lower.tail=F)
        #	pvalsUpper=p.adjust(pvalsUpper,"fdr")
        idx=which(pvalsUpper <= centeralpha)
        if (length(idx) > min.cells.ic) {
            return(list(rownames(ICAcell.matrix)[idx]))
        } else {
            return(NULL)
        }
    })

    # In many instances, there will be a large population of "null" or "default" cells that do not have their own positive IC loading.   In these instances, add an extra cluster center to accommodate them.
    u = unique(unlist(cell.list))

    frac = (length(u)/nrow(ICAcell.matrix))
    cat_with_time(frac)
    if (frac < nullthresh) {

        cat_with_time("adding null.")
        nulls = setdiff(c(rownames(ICAcell.matrix)),unlist(cell.list))
        cell.list[[length(cell.list) + 1]] <- list(nulls)
    }
    return(cell.list)
}

# Combine into one data frame, adding NAs for columns that don't exist in all input DFs
merge_curation_sheets=function(data_frame_list) {
    # remove null elements
    data_frame_list = data_frame_list[!sapply(data_frame_list, is.null)]
    all_colnames = unique(unlist(lapply(data_frame_list, function(df) return(colnames(df)))))
    data_frame_list = lapply(data_frame_list, function(df) {
        for (colname in setdiff(all_colnames, colnames(df))) {
            df[, colname] = NA
        }
        return(df)
    })
    return(do.call('rbind', data_frame_list))
}

######### Differential expression #############

#' Value to return if no differential expression, either because not enough cells, or because
#' no genes that achieve desired p-value threshold
make_null_diffexp=function(genes.return) {
    ent=rep(NA,times=genes.return)
    return(data.frame(GENE=ent,FOLDch=ent,pval=ent, pct.1 = ent, pct.2 = ent))
}

#' @return list(meanexp=log-transformed row-means of normalized expression, digsums=row-sums of raw expression), both for the given cells
compute_diffexp_mean_and_sum=function(norm,dge,cells) {
    genes = intersect(rownames(norm), rownames(dge))
    if (length(cells) == 1) {
        norm.row.means = mean(norm[genes,cells])
        raw.row.sums = sum(dge[genes,cells])
    } else {
        norm.row.means = rowMeans(norm[genes,cells])
        raw.row.sums = rowSums(dge[genes,cells])
    }
    return(list(meanexp=log(10000*norm.row.means+1), digsum=raw.row.sums))
}

# digits to round diffexp FOLDch by default
diffexp_FOLDch_round_digits=3

diffexp_default_FOLDch.thresh=1.5
diffexp_default_pval.thresh=1e-4

#' Compute differential expression
#' @param norm scaled digital expression
#' @param dge raw digital expression
#' @param cells first set of cells to compare
#' @param cells.comp second set of cells to compare
#' @param meanexp_digsums value returned by compute_diffexp_mean_and_sum(), or NULL
#' @param meanexp_digsums.comp value returned by compute_diffexp_mean_and_sum(), or NULL
#' @param FOLDch.thresh only return genes with fold change > this
#' @param only.pos apply thresh to increase in expression only.
#' @param pval.thresh only return genes with p-value < this.
#' @param round.FOLDch usually FOLDch is rounded to 3 digits, but when it might be inverted, it is not rounded for better reproducibility
diffexp<-function(norm,dge,cells,cells.comp, meanexp_digsums=NULL, meanexp_digsums.comp=NULL,
        FOLDch.thresh=diffexp_default_FOLDch.thresh, only.pos=TRUE, pval.thresh=diffexp_default_pval.thresh, round.FOLDch=TRUE) {
    if(min(c(length(cells),length(cells.comp)))<2) {
        return(make_null_diffexp(1))
    } else {
        if (is.null(meanexp_digsums)) {
            meanexp_digsums = compute_diffexp_mean_and_sum(norm, dge, cells)
        }
        if (is.null(meanexp_digsums.comp)) {
            meanexp_digsums.comp = compute_diffexp_mean_and_sum(norm, dge, cells.comp)
        }
        meanexp.1<-meanexp_digsums$meanexp
        meanexp.2<-meanexp_digsums.comp$meanexp
        digsums.1<-meanexp_digsums$digsum
        digsums.2<-meanexp_digsums.comp$digsum

        diffs<-meanexp.1 - meanexp.2
        if (only.pos) {
            genes.test<-which(diffs > log(FOLDch.thresh))
        } else {
            genes.test<-which(abs(diffs) > log(FOLDch.thresh))
        }

        diffs=diffs[genes.test]
        pvals=binomTest(digsums.1[genes.test],digsums.2[genes.test],sum(digsums.1),sum(digsums.2))
        idx=which(pvals < pval.thresh)
        diffs=diffs[idx]
        if (length(idx) > 0) {
                genes.return=length(idx)
                genes.include=order(diffs,decreasing=T)[1:genes.return]

                genes = names(diffs)[genes.include]
                if (length(genes) > 1) {
                   pct.1 = round(rowSums(norm[genes,cells]>0) / length(cells) , 2)
                   pct.2 = round(rowSums(norm[genes,cells.comp]>0) / length(cells.comp), 2)
                } else {
                   pct.1 = round(sum(norm[genes,cells]>0) / length(cells) , 2)
                   pct.2 = round(sum(norm[genes,cells.comp]>0) / length(cells.comp), 2)
                }
                FOLDch = exp(diffs[genes.include])
                if (round.FOLDch) {
                    FOLDch = round(FOLDch, diffexp_FOLDch_round_digits)
                }
                return.matrix<-data.frame(GENE=genes,FOLDch=FOLDch, pval=pvals[idx][genes.include], pct.1 = pct.1, pct.2 = pct.2)
                return(return.matrix)
        } else {
            return(make_null_diffexp(1))
        }
    }
}

#' Compute differential expression of every pair of clusters.
#' @param ellipsis args passed to diffexp
#' @return list(markers.pairwise=
#'              list(for each cluster
#'                   list(for each other cluster: data frame as returned by diffexp)
#'              means_and_sums=list(for each cluster: value returned by compute_diffexp_mean_and_sum
pairwiseDiffExpression=function(idents, raw_dge, scaled_dge, ...) {
    clusters = levels(idents)

    # precompute to save time.
    diffexp_mean_and_sums = lapply(clusters, function(cluster) {
        cat_with_time(paste0("Precomputing pairwise differential expression for ", cluster))
        cells = names(idents)[which(idents == cluster)]
        return(compute_diffexp_mean_and_sum(norm=scaled_dge, dge=raw_dge, cells))
    })
    names(diffexp_mean_and_sums) = clusters

    markers.pairwise = lapply(clusters, function(c1, ...) {
        cat_with_time(paste0("Computing pairwise differential expression for ", c1))
        c1.cells = names(idents)[which(idents == c1)]
        clusters.comp = setdiff(clusters, c1)
            ret = lapply(clusters.comp, function(c2, ...) {
                c2.cells = names(idents)[which(idents == c2)]
                diffexp(norm=scaled_dge, dge=raw_dge,
                    meanexp_digsums=diffexp_mean_and_sums[[c1]], meanexp_digsums.comp=diffexp_mean_and_sums[[c2]],
                    c1.cells, c2.cells, ...)
            }, ...)
            names(ret) = clusters.comp
            return(ret)
    }, ...)
    names(markers.pairwise) = clusters
    return(list(markers.pairwise=markers.pairwise, means_and_sums=diffexp_mean_and_sums))
}

#' Compute both pairwise differential expression of each pair of clusters, and differential expression of each cluster
#' vs. all the other cells.
#' The downside of this function is that the same thresholds must be used for both types of expression.
#' The upside is that some of the computation is shared.
#' @return list(pairwise_markers, cluster_markers)
pairwiseAndClusterDiffExpression=function(idents, raw_dge, scaled_dge, ...) {
    pairwise = pairwiseDiffExpression(idents=idents,
            raw_dge=raw_dge, scaled_dge=scaled_dge, ...)

    cluster.markers=lapply(levels(idents), function(cluster_name, ...) {
        cat_with_time(paste0("Computing cluster markers for ", cluster_name))
        cells=names(idents)[which(idents==cluster_name)]
        cells.comp=setdiff(names(idents),cells)
        diffexp(norm=scaled_dge, dge=raw_dge, cells=cells, cells.comp=cells.comp,
            meanexp_digsums=pairwise$means_and_sums[[cluster_name]], ...)
    }, ...)
    names(cluster.markers) = levels(idents)
    return(list(pairwise_markers=pairwise$markers.pairwise, cluster_markers=cluster.markers))
}

clusterreport<-function(idents, tree,tsne, cell.matrix,gene.matrix,norm,dge,ic.labels=NULL) {

    cell.matrix=as.matrix(cell.matrix[names(idents),])
    if (is.null(ic.labels)) {
        ic.labels = 1:ncol(cell.matrix)
    }

    if (nlevels(idents) > 1) {
        enrichment.by.cluster<-do.call(rbind,lapply(levels(idents),function(k){
            cells.cluster=names(idents)[which(idents==k)]
            cells.comp=setdiff(names(idents),cells.cluster)
            colMeans(as.matrix(cell.matrix[cells.cluster,])-colMeans(as.matrix(cell.matrix[cells.comp,])))

        }))

        rownames(enrichment.by.cluster)<-levels(idents)
    } else {
        enrichment.by.cluster=NULL
    }
    for (i in levels(idents)) {

        cells.cluster=names(idents)[which(idents == i)]
        cells.comp=setdiff(names(idents),cells.cluster)

        DEdata<-diffexp(norm,dge,cells.cluster,cells.comp)
        par(oma=c(0,0,3,0))
        layout(matrix(c(1,2,3,3,4,4),nrow=3,ncol=2,byrow=T))
        highlight.cells(tsne,cells.cluster,col='red',labels=F)
        if(!is.null(tree) & length(tree$tip.label) > 2) {
            tree.plot=as.dendrogram(tree)
            if (i %in% labels(tree.plot)) {
		cols=rep('black',times=length(labels(tree.plot)))
		names(cols) = labels(tree.plot)
		cols[i]<-'blue'
		cex=rep(1,times=length(labels(tree.plot)))
		names(cex)<-labels(tree.plot)
		cex[i]<-2
		tree.plot<-tree.plot %>% set("labels_col",cols)
		tree.plot<-tree.plot %>% set("labels_cex",cex)
		plot(tree.plot)
            } else {
                plot(c(0, 1), c(0, 1), ann = F, bty = 'n', type = 'n', xaxt = 'n', yaxt = 'n')
		text(x = 0.5, y = 0.5, paste("Small cluster (not on tree)"),
                     cex = 0.9, col = "black")
            }
        } else {
            plot(c(0, 1), c(0, 1), ann = F, bty = 'n', type = 'n', xaxt = 'n', yaxt = 'n')
            text(x = 0.5, y = 0.5, paste("Less than 3 clusters; no tree"),
                 cex = 0.9, col = "black")

        }
        if (!is.null(enrichment.by.cluster)) {
            topenrich=which.max(enrichment.by.cluster[i,])
            genesfromICs<-colnames(gene.matrix)[order(gene.matrix[topenrich,],decreasing=T)[1:5] ]
            barplot(enrichment.by.cluster[i,],main=paste0("Cluster ",i," top enriched IC: IC ",ic.labels[topenrich]," (",paste(genesfromICs,collapse=", "),")."))
            frame()
        }

        vps <- baseViewports()
        pushViewport(vps$inner, vps$figure, vps$plot)
        mytheme<-gridExtra::ttheme_default(core=list(fg_params=list(cex=0.8)),colhead=list(fg_params=list(cex=0.6)),rowhead=list(fg_params=list(cex=0.8)))
        disp=min(c(12,nrow(DEdata)))
        # Convert p-values to scientific notation, because otherwise format sucks for very small p-values
        # But very few significant digits, to avoid too-wide table.
        idx = which(!is.na(DEdata$pval))
        if (length(idx) > 0) {
            DEdata[idx,]$pval = formatC(DEdata[idx,]$pval, digits=1, format="g")
        }
        idx = which(!is.na(DEdata$FOLDch))
        if (length(idx) > 0) {
            DEdata[idx,]$FOLDch = formatC(DEdata[idx,]$FOLDch, digits=2, format="f")
        }
        grob<-gridExtra::tableGrob(t(DEdata[1: disp,-1]),theme=mytheme)
        grid.draw(grob)
        popViewport(3)

        topstr=paste(DEdata$GENE[1:6],collapse=", ")
        title(main=paste0("Cluster ",i,", ",length(cells.cluster)," cells :",topstr),outer=T,cex.main=1.6)
    }

    ds=min(10000,nrow(cell.matrix))

    baseline.idx=sort(sample(1:nrow(cell.matrix),size=ds))

    par(mfrow=c(3,3))
    cols=rainbow(length(levels(idents)))
    for (i in 1:ncol(cell.matrix)) {
        plot(1:length(baseline.idx),cell.matrix[baseline.idx,i],pch=20,col=cols[as.integer(idents[baseline.idx])],cex=0.3,xlab="",ylab="")
        title(main=paste0("Component ",ic.labels[i]))

    }
    title(main="IC loadings, colored by cluster ID",outer=T)

}

###  Tree building functions (from Seurat)  ####

make.dendrogram<-function(data,partition,reorder.partition=T,reorder.numeric=FALSE,min.clust.size) {
    cells.orig=names(partition)
    cat_with_time(table(partition))
    clusters.use=levels(partition)[which(table(partition)>min.clust.size)]
    if (length(clusters.use) <= 2) {
        return(list(tree=NULL, partition=partition))
    }
    cells.use=names(partition)[which(partition %in% clusters.use)]
    cells.hold=setdiff(cells.orig,cells.use)
    partition.remove=partition[cells.hold]
    partition.remove=droplevels(partition.remove)
    levels(partition.remove)<-as.character(1:length(levels(partition.remove)))
    partition=partition[cells.use]
    partition=droplevels(partition)
    data=data[cells.use,]
    tree<-tree.from.metacells(data,partition)
    if (reorder.partition) {
        old.order=sort(unique(partition))
        all.desc=getDescendants(tree,tree$Nnode+2)
        all.desc= old.order[all.desc[all.desc<=(tree $Nnode+1)] ]
        partition=factor(partition,levels=all.desc,ordered=T)
        names(partition) <-cells.use
    }
    if (reorder.numeric) {
        levels(partition)=as.character(1:length(levels(partition)))

    }
    tree.final=tree.from.metacells(data,partition)

    partition.final=factor(c(as.numeric(partition),(max(unique(as.numeric(partition)))+as.numeric(partition.remove))))
    names(partition.final)<-c(cells.use, cells.hold)
    partition.final=partition.final[cells.orig]
    return(list(tree=tree.final,partition= partition.final))

}

tree.from.metacells<-function(data,partition) {
    data=data[names(partition),]

    cat_with_time(system.time(metacells<-apply(data,2,function(c){
        tapply(as.vector(c),as.factor(partition),mean)

    })))
    cor.dist=as.dist(1-cor(t(metacells)))
    names(cor.dist) = rownames(metacells)

    tree=as.phylo(hclust(cor.dist))
    return(tree)
}

getDescendants<-function(tree,node,curr=NULL){
    if(is.null(curr)) curr<-vector()
    daughters<-tree$edge[which(tree$edge[,1]==node),2]
    curr<-c(curr,daughters)
    w<-which(daughters>=length(tree$tip))
    if(length(w)>0) for(i in 1:length(w))
                        curr<-getDescendants(tree, daughters[w[i] ], curr)
    return(curr)
}


### JACKSTRAW FUNCTIONS #####

jackStrawFast<-function(scaled.data,pcs.test=100,num.replicate=100,prop.freq=0.02,pval=1e-3,score.thresh=0.001,
        do.plot=F, random.seed=1) {

    set.seed(random.seed)
    #if(use.fast) {
    # For small data sets, this can't be bigger than the input matrix
    pcs.test = min(pcs.test, ncol(scaled.data), nrow(scaled.data))
    pc_obj<-rpca(scaled.data,k=pcs.test,center=TRUE,scale=F,retx=TRUE)
    #} else {
    #	pc_obj=prcomp(scaled.data)
    #}
    pc.genes=rownames(scaled.data)
    if (length(pc.genes)<150) prop.freq=max(prop.freq,0.05)
    # The sampling rate needs to produce at least 2 values, even for small samples
    prop.freq=max(prop.freq, 2.00001/nrow(scaled.data))
    md.x=pc_obj$x[,1:pcs.test]

    rownames(md.x)<-rownames(scaled.data)

    fake.pcVals.raw=lapply(1: num.replicate,function(x){jackRandomFast(scaled.data,pcs.test=pcs.test,prop.freq=prop.freq,seed=x)})

    fake.pcVals=simplify2array(lapply(1:pcs.test,function(x)as.numeric(unlist(lapply(1: num.replicate,function(y)fake.pcVals.raw[[y]][,x])))))
    #  fake.pcVals=sapply(1:pcs.test,function(x)as.numeric(unlist(lapply(1:num.replicate,function(y)fake.pcVals.raw[[y]][,x]))))
    #fake.pcVals=sapply(1: pcs.test,function(x)as.numeric(unlist(lapply(1:num.replicate,function(y)fake.pcVals.raw[[y]][,x]))))

    pAll=data.frame(sapply(1:pcs.test,function(x)unlist(lapply(abs(md.x[,x]),empP,abs(fake.pcVals[,x])))))

    colnames(pAll)=paste("PC",1:ncol(pAll),sep="")
    pc.score=unlist(lapply(1:pcs.test, function(x) {
        x=length(which(pAll[,x]<= score.thresh))
        y=floor(nrow(pAll)* score.thresh)
        z=nrow(pAll)
        fisher.test(matrix(c(x,z-x,y,z-y),ncol=2))$p.val
    }))
    #   pc.score=unlist(lapply(1:pcs.test, function(x) prop.test(c(length(which(pAll[,x]<=pval)),floor(nrow(pAll)*pval)),c(nrow(pAll),nrow(pAll)))$p.val))
    sig.pcs=which(pc.score < pval)
    num.pcs=length(sig.pcs)
    if (do.plot) {
        plot(1:pcs.test,log10(pc.score),type='l',xlab="PC #",ylab="PC score")
        abline(v=num.pcs,col='red')
        abline(h=log10(pval),col='blue')
        plot(1:50,log10(pc.score[1:50]),type='l',xlab="PC #",ylab="PC score")
        abline(v=num.pcs,col='red')
        abline(h=log10(pval),col='blue')
    }
    return(sig.pcs)
}


jackRandomFast <- function(scaled.data,pcs.test,prop.freq=0.02,seed=1) {
    data.mod=scaled.data
    set.seed(seed)
    # sample size without replacement must be <= population
    rand.idx=sample(1:nrow(scaled.data),size=min(max(prop.freq*nrow(scaled.data),3), nrow(scaled.data)))
    data.mod[rand.idx,]=shuffleMatRow(scaled.data[rand.idx,])
    set.seed(1)
    #if (use.fast) {
    fake.pca<-rpca(data.mod,k=pcs.test,center=T,scale=F,retx=TRUE)
    #} else {
    #	  fake.pca=prcomp(data.mod,rank=pcs.test)
    #}
    fake.x<-fake.pca$x[,1:pcs.test]

    rownames(fake.x)<-rownames(data.mod)
    #fake.rot=fake.pca$u

    return(fake.x[rand.idx,1:pcs.test])
}




### HELPER FUNCTIONS ######

weight=function(nn.1,nn.2,k) {
    int=intersect(nn.1,nn.2)
    idx.1=which(nn.1 %in% int)
    idx.2=which(nn.2 %in% int)
    sum((k+1-idx.1)*(k+1 - idx.2))
}

shuffleMatRow<- function(x) {
    x2=x
    x2 <- t(x)
    ind <- order(c(col(x2)), runif(length(x2)))
    x2 <- matrix(x2[ind], nrow=nrow(x), ncol=ncol(x), byrow=TRUE)
    return(x2)
}

empP <- function(x,nullval) {
    return(length(which(nullval>x))/length(nullval))
}

ident.plot<-function(tsrots,idents,cex.use=0.4, legend=FALSE, randomize_rainbow=TRUE, random.seed=1) {
    cols.use=rainbow(length(levels(idents)))
    if (randomize_rainbow) {
        set.seed(random.seed)
        cols.use = sample(cols.use, length(cols.use), replace=FALSE)
    }
    plot(tsrots[,1],tsrots[,2],col=cols.use[as.integer(idents)],pch=16,xlab="TSNE_1",ylab="TSNE_2",cex=cex.use)
    if (legend) {
        legend.cex = 1
        if (length(cols.use) > 25) {
            legend.cex = 0.5
        }
        legend(x="topright", legend=levels(idents), pch=16, col=cols.use, bty="n", cex=legend.cex)
    }
    k.centers=t(sapply(levels(idents),function(x) {

        cells.use=names(idents)[which(idents == x)]
        if(length(cells.use) > 1) {
            apply(tsrots[cells.use,],2,mean)
        } else {
            tsrots[cells.use,]
        }
    }))
    points(k.centers[,1],k.centers[,2],cex=1.3,col="white",pch=16); text(k.centers[,1],k.centers[,2],levels(idents),cex=1)
}

highlight.cells<-function(tsrots,cells,col='red',labels=F) {
    if(!labels) {
        labels=c("","")
    }
    plot(tsrots[,1],tsrots[,2],pch=20,cex=0.3,col=gray(0.8),xlab=labels[1],ylab=labels[2])
    points(tsrots[cells,1],tsrots[cells,2],pch=20,cex=0.3,col=col)

}

fplot<-function(tsne,factor,title,cols.use=heat.colors(10),pt.size=0.7,pch.use=20) {
    c=intersect(rownames(tsne),names(factor))
    data.use=factor[c]
    data.cut=as.numeric(as.factor(cut(as.numeric(data.use),breaks=length(cols.use))))
    data.col=rev(cols.use)[data.cut]
    plot(tsne[c,1],tsne[c,2],col=data.col,cex=pt.size,pch=pch.use,main=title)

}

muti<-function(data,verbose=T) {

    k=dim(data)[2]
    mutual_info=matrix(0,k,k)
    for (i in 1:k) {
        for (j in 1:k) {
            mutual_info[i,j] = mutinformation(discretize(data[,i]),discretize(data[,j]))
            mutual_info[j,i]=mutual_info[i,j]
            if (verbose) cat_with_time(c(i,j))

        }
    }
    return(mutual_info)
}



### t-SNE FUNCTIONS ######
#' @return Y matrix from Rtsne return value
do_tsne<-function(x, pca=FALSE, random.seed=1) {
    set.seed(random.seed)
    return(uniquify_and_transform_matrix(x, function(unique_matrix) {
        default_perplexity = 30 # This is the Rtsne default
        perplexity = default_perplexity
        # This is the perplexity test used by Rtsne wrapper, so if this condition is true, set perplexity just small
        # enough to make the condition false
        if (nrow(unique_matrix) - 1 < 3 * perplexity) {
            perplexity = as.integer((nrow(unique_matrix) - 1)/3)
            cat_with_time(sprintf("For Rtsne, default_perplexity %d is too large for %d rows, so setting perplexity=%d",
                default_perplexity, nrow(unique_matrix), perplexity))
        }
        Rtsne(unique_matrix, pca=pca, check_duplicates=FALSE, perplexity=perplexity)$Y
    }))
}

#' Apply a transformation on a matrix that requires the matrix rows to be unique, even if the rows are not unique.
#' Creates a temporary matrix with unique rows, applies the transformation, and then creates an ouput matrix
#' with the same number of rows as the input.
#` @param mat The matrix to be transformed
#' @param a transformation function that takes a matrix with unique rows and returns a matrix with the same number of rows as the input.
uniquify_and_transform_matrix<-function(mat, transformer) {
    # Stolen from R duplicated() function -- convert each row to a string
    vec_with_dupes=apply(mat, 1, function(x) paste(x, collapse = "\r"))
    # Use factor to identify the rows that are identical
    fac = factor(vec_with_dupes)
    if (length(fac) == nlevels(fac)) {
        # no duplicates
        return(transformer(mat))
    } else {
        # get numeric indices of the rows in the input matrix
        numeric_fac = as.numeric(fac)
        rm(fac)
        # create a matrix with only unique rows
        de_duped_mat = mat[unique(numeric_fac), ]
        # run the transformation
        de_duped_result = transformer(de_duped_mat)
        if (nrow(de_duped_result) != nrow(de_duped_mat)) {
            stop("Transformation changed the number of rows")
        }
        rm(de_duped_mat)
        # create a matrix with the same number of rows as the input
        result = de_duped_result[numeric_fac,]
        rownames(result) = rownames(mat)
        return(result)
    }
}

### HELPER FUNCTIONS ######
sum_dge_by_cluster=function(idents, raw_dge) {
    SumByCellClass = function(x, cell_class) tapply( x, cell_class, sum )
    return(t(apply(raw_dge, 1, SumByCellClass, cell_class=idents)))
}


#' @return list(positive_loading, negative_loading).  Each element is list num_genes long, with most positive and
#' most negative loading genes, with the first element of each list being the most (positive, negative)
get_pos_and_neg_loading_lists_for_ic=function(gene_loadings, num_genes) {
    genes = names(gene_loadings)
    num_genes = min(num_genes, length(genes))
    sort_order = order(gene_loadings)
    return(list(positive_loading=genes[rev(tail(sort_order, num_genes))],
                negative_loading=genes[sort_order[1:num_genes] ]))
}


get_subcluster_idents=function(results.curated) {
    factor(unlist(lapply(names(results.curated),function(q){
        cluster.number = as.numeric(q)
        subids = as.character(results.curated[[q]]$clustering$louvain)
        if (length(subids) == 0) {
            return(list())
        }
        ret = paste(cluster.number,subids,sep="-")
        names(ret) = names(results.curated[[q]]$clustering$louvain)
        return(ret)
    })))
}

#' @param cluster_names vector of strings of the form "x-y", or just "x", where x and y are integers.
#' @return integer vector of the major cluster number (i.e. the number before the dash, if the dash is present).
major_cluster_number=function(cluster_names) {
    sapply(strsplit(cluster_names, '-'), function(thesplit) as.integer(thesplit[1]))
}

#' @param cluster_names vector of strings of the form "x-y", or just "x", where x and y are integers.
#' @return integer vector of the minor cluster number (i.e. the number after the dash) or 0 if there is no dash.
minor_cluster_number=function(cluster_names) {
    sapply(strsplit(cluster_names, '-'), function(thesplit) {
        if (length(thesplit) == 1) {
            return(0)
        } else {
            return(as.integer(thesplit[1]))
        }
    })
}

#' @param cluster_names vector of cluster names of the form "x-y", or just "x", where x and y are integers
#' @return vector of cluster_names ordered so that, e.g. 2-1 sorts before 11-2
sort_cluster_names=function(cluster_names) {
   return(cluster_names[order(major_cluster_number(cluster_names), minor_cluster_number(cluster_names))])
}

#' Note that we've decided not to bother ordering subcluster factors this way.
#' @param cluster_assignments a character vector of cluster names, presumably the cluster assignments for cells
#' @return the same cluster_assignments, but as a factor, in which the levels of the factor are ordered according
#' to sort_cluster_names() function.
make_cluster_name_ordered_factor=function(cluster_assignments) {
    f = factor(cluster_assignments)
    return(factor(cluster_assignments, levels=sort_cluster_names(levels(f))))
}
