# MIT License
#
# Copyright 2017 Broad Institute
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

PipelineManager = R6Class("PipelineManager",
    public = list(
        pipeline.parameters = NULL,
        directories = NULL,
        stage_mgr = NULL,
        yamlpath = NULL,

        initialize = function(subdirectory_names, stages, yamlpath=NULL, defaults=list()) {
            yamlpathAndTrailingArgs = GetYamlPathAndTrailingArgs(yamlpath)
            self$yamlpath = yamlpathAndTrailingArgs$yamlpath
            pipeline.parameters = GetPipelineParameters(self$yamlpath)
            for (trailingArg in yamlpathAndTrailingArgs$other_args) {
                parsedArg = tryCatch(yaml.load(trailingArg), error=function(e) {
                    stop(sprintf("Problem parsing command-line argument '%s' as YAML: %s", trailingArg, e$message))
                })
                # Note that order of args is important -- the command-line arg overrides what is in the yaml.
                pipeline.parameters = applyDefaultSettings(parsedArg, pipeline.parameters)
            }
            # Note that order of args is important -- the yaml (plus command-line args) take precedence over the defaults.
            self$pipeline.parameters = applyDefaultSettings(pipeline.parameters, defaults)
            if (self$pipeline.parameters$workflow$stack_trace) {
                enable_full_stack_trace()
            }
            if (self$pipeline.parameters$workflow$record_history) {
                private$save_pipeline_history()
            }
            options(warn=self$pipeline.parameters$workflow$warn_level)

            working_directory = normalizePath(self$pipeline.parameters$workflow$working_directory)
            setwd(working_directory)

            self$directories = c(as.list(subdirectory_names), "stage")
            names(self$directories) = self$directories
            self$directories = lapply(self$directories, function(dirname) {return(paste0(working_directory, "/", dirname, "/"))})
            lapply(self$directories, ensure_directory_exists)

            if (!is.na(stages) && length(stages) > 0) {
                self$stage_mgr = create_stage_manager(
                    pipeline.parameters=self$pipeline.parameters,
                    stage_directory=self$directories$stage,
                    stages=stages)
            }
        },
        #' Preserve a copy of curation sheet before opening.
        #' Copy the given file to the curation history directory.  This cannot be called more than once for the
        #' same file in the same IcaCluster invocation.
        #' @param open If TRUE, return a file connection.  If FALSE, return the path to the copied file.
        #' @return connection or path
        open_curation_file_with_history=function(path, open=TRUE) {
            base = basename(path)
            to = file.path(private$get_curation_history_directory(), base)
            if (!file.copy(from=path, to=to, overwrite=FALSE)) {
                stop(sprintf("Could not copy %s to %s", path, to))
            }
            if (open) {
                return(file(to, open="r"))
            } else {
                return(to)
            }
        },
        get_history_directory=function() {
            history_directory = file.path(normalizePath(self$pipeline.parameters$workflow$working_directory), 'history')
            if (!dir.exists(history_directory)) {
                dir.create(history_directory, recursive=TRUE)
            }
            return(history_directory)
        }
    ),
    private = list(
        saved_options=options("warn"),
        curation_history_directory=NULL,
        get_curation_history_directory=function() {
            if (is.null(private$curation_history_directory)) {
                history_directory = self$get_history_directory()
                private$curation_history_directory = tempfile(pattern="curation_history.", tmpdir=history_directory)
                if (!dir.create(private$curation_history_directory)) {
                    stop(sprintf("Error creating directory %s.  See above for reason", private$curation_history_directory))
                }
            }
            return(private$curation_history_directory)
        },
        save_pipeline_history = function() {
            user = Sys.info()['user']
            cluster_library_directory=normalizePath(system.file(package='DropSeq.icaCluster'))
            cwd = getwd()
            project_directory = normalizePath(self$pipeline.parameters$workflow$working_directory)
            history_directory = self$get_history_directory()
            yaml_with_defaults_applied = as.yaml(self$pipeline.parameters)
            directory_listing = system2_capture_stdout_check_error(command='ls', args=c('-lR', project_directory))
            original_yaml_contents = system2_capture_stdout_check_error(command='cat', args=self$yamlpath)
            history_file = tempfile(pattern="history.", tmpdir=history_directory, fileext=".log")
            cat(paste("PWD:", cwd),
                paste("USER:", user),
                paste("DropSeq.icaCluster:", cluster_library_directory),
                paste("YamlPath:", normalizePath(self$yamlpath)),
                paste("CommandLine:", paste(commandArgs(), collapse = " ")),
                "YamlWithDefaultsApplied:",
                "--------------",
                yaml_with_defaults_applied,
                "--------------",
                directory_listing,
                "--------------",
                "YamlContents:",
                "--------------",
                original_yaml_contents,
                sep="\n",
                file=history_file)
        }
    )
)
