# MIT License
# 
# Copyright 2018 Broad Institute
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

SummarizeICs = function(ica) {
    data.frame(
        IC = paste( "IC", 1:ncol(ica$gene_loadings), sep="" ),  # colnames(ica$cell_rotations),
        top_genes = apply( ica$gene_loadings, 2, function(x) paste(head(names(sort(x,decreasing=T)), 9), collapse = " ")),
        l2norm = round( apply( ica$gene_loadings, 2, function(x) sqrt(sum(x^2))), 2),
        skew = round( apply( ica$cell_rotations, 2, function(x) mean(x^3) ), 2),
        kurtosis = round( apply( ica$cell_rotations, 2, function(x) mean(x^4) ), 2)
    )
}

WithinBetweenGroupVarianceRatio = function( scores1, f1 ) mean(tapply( scores1, f1, var )) / var(tapply( scores1, f1, mean ))
BetweenWithinGroupVarianceRatio = function( scores1, f1 ) 1 / WithinBetweenGroupVarianceRatio( scores1, f1 )
MatrixBetweenWithinGroupVarianceRatio = function( ica.cell.scores, f ) {
    # f is a factor containing assignments of cells to groups (cell ICs in the entity names)
    f1 = f[ match( rownames(ica.cell.scores), names(f))]
    apply( ica.cell.scores, 2, BetweenWithinGroupVarianceRatio, f1=f1 )
}

# Functions to extract useful information from cell-name prefices
TrimRHSchars = function( x, n_chars_trim ) substr( x, 1, nchar(x)-n_chars_trim )
CellNamePrefix = function(x) TrimRHSchars( x, 13 )
CellNamePrefixSummary = function(x) summary(factor( CellNamePrefix(x) ))

InitializeCellSets = function() {
      cell.sets = list()
      cell.sets$registry = data.frame( cell.set.id=character(0),
                                       n.cells = integer(0),
                                       current.name=character(0),
                                       date.created=character(0),
                                       created.by=character(0) )
      cell.sets$members = list()
      cell.sets$names = data.frame( cell.set.id=character(0),
                                    name=character(0),
                                    date=character(0),
                                    author=character(0) )
      cell.sets$comments = data.frame( cell.set.id=character(0),
                                       comment=character(0),
                                       date=character(0),
                                       author=character(0) )
      return(cell.sets)
}

RecordCellSetName = function(cell.sets, cell.set.id, name, date.now, author ) {
    cell.sets$names = rbind( cell.sets$names, data.frame(
      cell.set.id = cell.set.id,
      name = name,
      date = date.now,
      author = author ) )
    return(cell.sets)
}

RecordCellSetComment = function(cell.sets, cell.set.id, comment, date.now, author ) {
    cell.sets$comments = rbind( cell.sets$comments, data.frame(
        cell.set.id = cell.set.id,
        comment = comment,
        date = date.now,
        author = author ) )
    return(cell.sets)
}



AddCellSet = function(cell.sets, whos_in, name, comment, author ) {
  date.now = date()

  cell.set.id = ifelse( is.null(nrow(cell.sets$registry)), 1, 1+nrow(cell.sets$registry) )

  cell.sets$members[[cell.set.id]] = whos_in
  cell.sets$registry = rbind( cell.sets$registry, data.frame(
    cell.set.id =  cell.set.id,
    n.cells = length(whos_in),
    current.name = name,
    date.created = date.now,
    created.by = author ) )

  cell.sets = RecordCellSetName(cell.sets, cell.set.id, name, date.now, author )

  cell.sets = RecordCellSetComment(cell.sets, cell.set.id, comment, date.now, author )

  return(cell.sets)
}




AddCellSetsFromAssignments = function(cell.sets, assign, names_prefix, comment, author ) {
    for( i in levels(assign) ) {
        cell.sets = AddCellSet(cell.sets, names(assign[assign==i]), paste0(names_prefix,i), comment, author )
    }
    return(cell.sets)
}

ComputeBulkMetacell = function(dge) {
    # initializes the matrix withthe metacell of all cells together
    metacell = matrix( data=rowSums(dge), nrow=nrow(dge), ncol=1 )
    rownames(metacell) = rownames(dge)
    colnames(metacell) = c("bulk")
    return(metacell)
}



PrepareDropWorks = function(analysis.name, SPECIES, dge.slice.size=100, outdir=NA) {
    pipeline_mgr = PipelineManager$new(
        subdirectory_names=workflow_directories,
        stages=all_stages,
        defaults=default_settings_no_history)
    pipeline.parameters = pipeline_mgr$pipeline.parameters
    all_ics = read_agg_subica(pipeline_mgr)
    ica.analyses = c("global", sapply(names(all_ics), function(cluster_name) paste0("cluster", cluster_name)))
    names(ica.analyses) = NULL
    ica = read_round1_ica(pipeline_mgr)
    dge = readRDS(get_filtered_raw_dge_path(pipeline_mgr))
    UMIs.per.cell = colSums(dge)
    scores = cbind(ica$cell_rotations, UMIs=UMIs.per.cell)

    score.matrices = list()
    score.matrices[["ICA.global"]] = ica$cell_rotations
    score.matrix.summaries = list()
    score.matrix.summaries[["ICA.global"]] = SummarizeICs(ica)
    maps2d = list()
    maps2d[["tSNE.global"]] = ica$tsne

    for (clusterN in names(all_ics)) {
        score.matrix.name = paste0( "ICA.cluster", clusterN )
        ica.n = all_ics[[clusterN]]
        score.matrices[[ score.matrix.name ]] = ica.n$cell_rotations
        colnames( score.matrices[[ score.matrix.name ]] ) = paste( "IC", 1:ncol(ica.n$cell_rotations), sep="" )
        score.matrix.summaries[[score.matrix.name]] = SummarizeICs(ica.n)
        maps2d[[ paste0( "tSNE.cluster", clusterN)  ]] = ica.n$tsne
    }

    for (tsne_name in names(maps2d)) {
        colnames(maps2d[[tsne_name]]) = c("tSNE.x", "tSNE.y")
    }


    score.matrices[["UMIs"]] = data.frame( UMIs=UMIs.per.cell, UMIs.log10=log10(UMIs.per.cell) )

    assignments = list()

    assignments[["SeqPool"]] = as.factor( CellNamePrefix(rownames(maps2d[["tSNE.global"]])) )
    names(assignments[["SeqPool"]]) = rownames(maps2d[["tSNE.global"]])
    assignments[["Cluster (ICA)"]] = readRDS( file.path(pipeline_mgr$directories$assign, paste0(pipeline.parameters$name, cluster_assign_extension)))

    # Add to the score.matrix.summaries, an ANOVA based on the primary (first) assignment group
    #   (this is useful for identifying "replicate ICs")
    for(score.matrix.name in names(score.matrix.summaries)) {
        score.matrix.summaries[[score.matrix.name]]$var.btwn.wthn = round( MatrixBetweenWithinGroupVarianceRatio( score.matrices[[score.matrix.name]], assignments[[1]]), 2)
    }

    metacell = readRDS(file.path(pipeline_mgr$directories$metacells, paste0(pipeline.parameters$name,".metacells.RDS")))
    bulk_metacell = ComputeBulkMetacell(dge)
    metacell = cbind(bulk_metacell, metacell[rownames(bulk_metacell), ])

      # Create fast dge by downsampling the cells
    set.seed(1)
    cell.sample.10k = sample( 1:ncol(dge), min(1e04, ncol(dge)), replace=F )
    dge.fast = as.matrix( dge[,cell.sample.10k] )

    cell.sets = AddCellSetsFromAssignments(InitializeCellSets(), assignments[["Cluster (ICA)"]], "ICA cluster ",  "from ICAcluster round 1", "ICAcluster")

    DATASET_NAME = pipeline.parameters$name

    if (is.na(outdir)) {
        outdir = dropworks_dir
    }
    outdir = ensure_directory_exists(outdir)


    cat_with_time("Writing objects")
    yamlpath = pipeline_mgr$yamlpath
    for (object_name in c(
                    "ica.analyses",
                    "analysis.name",
                    "scores",
                    "score.matrix.summaries",
                    "score.matrices",
                    "metacell",
                    "maps2d",
                    "cell.sets",
                    "assignments",
                    "dge.fast",
                    "SPECIES",
                    "yamlpath")) {
        saveRDS(get(object_name), file.path(outdir, paste0(object_name, ".RDS")))
    }

    # save DGE in dense slices
    dgedir = ensure_directory_exists(file.path(outdir, paste0("dge.", dge.slice.size)))
    slice_starts = seq(to=nrow(dge), by=dge.slice.size)
    cat_with_time(sprintf("Writing %d DGE slices", length(slice_starts)))
    for (index in slice_starts) {
        last = min(nrow(dge), index + dge.slice.size - 1)
        outfile = file.path(dgedir, paste0(sprintf("%05dgene.RDS", (index %/% dge.slice.size))))
        cat_with_time(sprintf("Writing %s", outfile))
        saveRDS(as.matrix(dge[index:last, ]), outfile)
    }
    cat_with_time("Done writing")
}