# MIT License
#
# Copyright 2017 Broad Institute
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

#' Invoke matlab fitgmdist function
#' @return list of data frames (soft, BIC, loglik, lambda, sigma, mu)
matlab_gmm_cluster=function(cell_rotations, idents, tolFun=NULL) {
    tmpdir=tempdir()
    cell_rotations_file = file.path(tmpdir, "cell_rotations.txt")
    idents_file = file.path(tmpdir, "idents.txt")
    write.table(cell_rotations, file=cell_rotations_file,sep=",",col.names=F,row.names=F,quote=F)
    write.table(idents,file=idents_file,sep=",",col.names=F,row.names=F,quote=F)

    soft_file = file.path(tmpdir, "soft.txt")
    BIC_file = file.path(tmpdir, "BIC.txt")
    loglik_file = file.path(tmpdir, "loglik.txt")
    lambda_file = file.path(tmpdir, "lambda.txt")
    mu_file = file.path(tmpdir, "mu.txt")
    sigma_file = file.path(tmpdir, "sigma.txt")

    script_path = file.path(system.file(package='DropSeq.icaCluster'), 'bin','matlabgmmcluster.m')
    script_dir = dirname(script_path)
    script_name = removeSuffix(basename(script_path), ".m")


    # There doesn't appear to be a way to load a function from a script file without executing it.
    # So, cd to the directory containing the script, and execute the script as a function call.
    #
    # matlabgmmcluster(dataInFile, cluster0InFile, softOutFile, bicOutFile, logLikOutFile,
    #                   lambdaOutFile, muOutFile, sigmaOutFile, tolFun)

    funcall = sprintf("%s('%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s'",
    script_name, cell_rotations_file, idents_file, soft_file, BIC_file, loglik_file, lambda_file, mu_file, sigma_file)
    if (!is.null(tolFun)) {
        funcall=paste0(funcall, ", ", tolFun)
    }
    funcall=paste0(funcall, ")")
    cmd = sprintf("\"try, %s; catch err, disp(err.message); exit(1); end; exit(0);\"", funcall)

    # Throws an exception if matlab not found.
    matlab_plus_args = strsplit(get_matlab(), "[[:blank:]]+")[[1]]

    # get_matlab return a string with the path to matlab plus some additional args.  Pull those off and
    # prepend them to our args.
    matlab = matlab_plus_args[1]
    all_args = c(tail(matlab_plus_args, -1), cmd)

    wd = getwd()
    setwd(script_dir)
    ret = system2(matlab, all_args)
    setwd(wd)
    if (ret != 0) {
        stop(sprintf("Exit status %d from %s %s", ret, matlab, cmd))
    }

    soft=read.table(soft_file,sep=",",header=F,stringsAsFactors=F)
    BIC=read.table(BIC_file,sep=",",header=F,stringsAsFactors=F)
    loglik=read.table(loglik_file,sep=",",header=F,stringsAsFactors=F)
    sigma=read.table(sigma_file,sep=",",header=F,stringsAsFactors=F)
    lambda=read.table(lambda_file,sep=",",header=F,stringsAsFactors=F)
    mu=read.table(mu_file,sep=",",header=F,stringsAsFactors=F)

    # delete temporary files
    unlink(c(cell_rotations_file, idents_file, soft_file, BIC_file, loglik_file, lambda_file, mu_file, sigma_file))

    return(list(soft=soft, BIC=BIC, loglik=loglik, lambda=lambda, sigma=sigma, mu=mu))
}
