# MIT License
#
# Copyright 2017 Broad Institute
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

cell_cluster_outcome_reasons=list(
    min_genes='min_genes',                        # filtered by load_dge: min_genes
    singleton_cluster='singleton_cluster',        # First-round Louvain clustering has put the cell into a singleton cluster
    singleton_subcluster='singleton_subcluster',  # Subclustering has put the cell into a singleton cluster
    small_cell='small_cell',                      # Cell is below median size of cells in its first-round cluster
    not_enough_ics='not_enough_ics',              # subica + curation produced fewer than 2 real ICs
    curation='curation',                          # Either no subcluster curation sheet, or no real ICs in the curation sheet
    doublet='doublet',
    outlier='outlier',
    round1_doublet='round1_doublet',
    round1_outlier='round1_outlier'
)

# When selecting cells for computing clean markers and metacells, in addition to cells that subclustered,
# include cells that didn't subcluster for any of the following reasons.
default_clean_markers_included_reasons=c(cell_cluster_outcome_reasons$curation, cell_cluster_outcome_reasons$not_enough_ics)

#' For generating the cell_cluster_outcomes file post-hoc for projects that were subclustered before this feature
#' was implemented.  Note that apply_ic_curation stage must be re-run before doing this in order to capture doublet
#' and outlier cells.
DetermineCellClusterOutcomeClp=function(yamlpath=NULL) {
    pipeline_mgr = PipelineManager$new(
        subdirectory_names=workflow_directories,
        yamlpath=yamlpath,
        stages=NA,
        defaults=default_settings_no_history)

    saveRDS(DetermineCellClusterOutcome(pipeline_mgr),
        file=file.path(pipeline_mgr$directories$assign, paste0(pipeline_mgr$pipeline.parameters$name, cell_cluster_outcomes_extension)))
}

#' Examine many of the clustering workflow outputs in order to determine the outcome of every cell.
#' Cannot be run until subcluster idents have been written, which is at the end of subcluster stage.
#' Note that any cells that are dropped by MergeDgeSparse will not be captured by this function, but currently
#' we are running in 'seurat compatibility mode' so MergeDgeSparse does not drop any cells.
#' @return a data frame with rownames=cell barcodes and columns cluster, subcluster, reason (all factors).
DetermineCellClusterOutcome=function(pipeline_mgr) {
    pipeline.parameters = pipeline_mgr$pipeline.parameters


    cat_with_time('Computing cell cluster outcomes')
    all_cells = colnames(loadSparseDge(get_raw_dge_path(pipeline_mgr)))

    cell_cluster_outcome = data.frame(cluster=rep(NA, length(all_cells)), subcluster=NA, reason=NA, row.names=all_cells)

    cat_with_time('Finding cells that did not pass min_genes filter')
    more_than_min_genes_cells = colnames(readRDS(get_filtered_raw_dge_path(pipeline_mgr)))
    few_genes_cells = setdiff(all_cells, more_than_min_genes_cells)
    if (length(few_genes_cells) > 0) {
        cell_cluster_outcome[few_genes_cells, ]$reason = cell_cluster_outcome_reasons$min_genes
    }

    cat_with_time('Finding cells that were in singleton clusters')
    dropped_singletons = readRDS(file.path(pipeline_mgr$directories$assign, paste0(pipeline.parameters$name,".DroppedSingletons.RDS")))
    if (length(dropped_singletons) > 0) {
        cell_cluster_outcome[dropped_singletons,]$reason = cell_cluster_outcome_reasons$singleton_cluster
    }

    manual_firstround_ica_path = file.path(pipeline_mgr$directories$components, paste0(pipeline.parameters$name, firstround_manual_ica_extension))
    if (file.exists(manual_firstround_ica_path)) {
        curated.ica.obj = readRDS(manual_firstround_ica_path)
        if (length(curated.ica.obj$outliers) > 0) {
            cell_cluster_outcome[curated.ica.obj$outliers,]$reason = cell_cluster_outcome_reasons$round1_outlier
        }
        if (length(curated.ica.obj$doublets) > 0) {
            cell_cluster_outcome[curated.ica.obj$doublets,]$reason = cell_cluster_outcome_reasons$round1_doublet
        }
    }

    cat_with_time('Assigning round-1 clusters')
    round1_idents = readRDS(file.path(pipeline_mgr$directories$assign, paste0(pipeline.parameters$name,".cluster.assign.RDS")))
    cell_cluster_outcome[names(round1_idents),]$cluster = as.character(round1_idents)
    # Initially assign round-1 cluster to subcluster also
    cell_cluster_outcome[names(round1_idents),]$subcluster = as.character(round1_idents)

    cat_with_time('Finding clusters that did not have enough ICs to subcluster')
    ICs = readRDS(paste0(pipeline.parameters$name, all_ics_extension))
    not_enough_ics_clusters = setdiff(levels(round1_idents), names(ICs))
    if (length(not_enough_ics_clusters) > 0) {
        cell_cluster_outcome[cell_cluster_outcome$cluster %in% not_enough_ics_clusters, ]$reason = cell_cluster_outcome_reasons$not_enough_ics
    }

    cat_with_time('Finding cells that were not included in subclustering because they were too small')
    for (cluster_name in names(ICs)) {
        subclustered_cells = rownames(ICs[[cluster_name]]$cell_rotations)
        small_cells = (!is.na(cell_cluster_outcome$cluster) & cell_cluster_outcome$cluster == cluster_name &
            is.na(cell_cluster_outcome$reason) & !row.names(cell_cluster_outcome) %in% subclustered_cells)
        if (any(small_cells)) {
            cell_cluster_outcome[small_cells, ]$reason = cell_cluster_outcome_reasons$small_cell
        }
    }

    cat_with_time('Finding clusters that did not subcluster due to curation')
    subcluster_parameters = pipeline.parameters$subcluster
    cluster.method = subcluster_parameters$cluster.method
    cluster_method_label = get_or_else(subcluster_parameters$cluster_method_label, cluster.method)
    center.KDE = subcluster_parameters$center.KDE
    curated_results = readRDS(paste0(pipeline.parameters$name, "_", cluster_method_label, "_KDE_",center.KDE, ".curatedresults.RDS"))
    removed_by_curation_clusters = names(curated_results)[sapply(curated_results, is.null)]
    if (length(removed_by_curation_clusters) > 0) {
        curated_cells = (cell_cluster_outcome$cluster %in% removed_by_curation_clusters & is.na(cell_cluster_outcome$reason))
        if (any(curated_cells)) {
            cell_cluster_outcome[curated_cells, ]$reason = cell_cluster_outcome_reasons$curation
        }
    }


    cat_with_time('Finding cells identified as doublets when applying curation sheet')
    curation_cell_holdouts = readRDS(file.path(pipeline_mgr$directories$assign, paste0(pipeline.parameters$name, curation_cell_holdouts_extension)))
    doublets = unlist(lapply(curation_cell_holdouts, function(l) { l$doublets }))
    if (length(doublets) > 0) {
        cell_cluster_outcome[doublets, ]$reason = cell_cluster_outcome_reasons$doublet
    }

    cat_with_time('Finding cells identified as outliers when applying curation sheet')
    outliers = unlist(lapply(curation_cell_holdouts, function(l) { l$outliers }))
    if (length(outliers) > 0) {
        cell_cluster_outcome[outliers, ]$reason = cell_cluster_outcome_reasons$outlier
    }

    cat_with_time("Finding cells put into singleton subclusters")
    singleton_subcluster_cells = unlist(sapply(curated_results, function(cr) cr$clustering$droppedSingletons))
    if (length(singleton_subcluster_cells) > 0) {
        cell_cluster_outcome[singleton_subcluster_cells, ]$reason = cell_cluster_outcome_reasons$singleton_subcluster
    }

    cat_with_time('Assigning subcluster idents')
    final_idents = get_subcluster_idents(curated_results)
    if (length(final_idents) > 0) {
        cell_cluster_outcome[names(final_idents), ]$subcluster = as.character(final_idents)
    }

    # Convert to factors after completely populated
    cell_cluster_outcome$cluster = make_cluster_name_ordered_factor(as.integer(cell_cluster_outcome$cluster))
    cell_cluster_outcome$subcluster = make_cluster_name_ordered_factor(cell_cluster_outcome$subcluster)
    cell_cluster_outcome$reason = factor(cell_cluster_outcome$reason)

    return(cell_cluster_outcome)
}

