# MIT License
#
# Copyright 2017 Broad Institute
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Get locations of files that are accessed multiple times

get_dge_file_prefix = function(pipeline_mgr) {
    return(paste0(pipeline_mgr$directories$dge, pipeline_mgr$pipeline.parameters$name))
}

dge_path_helper = function(pipeline_mgr, extension) {
    return(paste0(get_dge_file_prefix(pipeline_mgr), extension))
}

get_raw_dge_path = function(pipeline_mgr) {
    return(dge_path_helper(pipeline_mgr, ".raw.dge.txt.gz"))
}

get_scaled_dge_path = function(pipeline_mgr) {
    return(dge_path_helper(pipeline_mgr, ".scaled.dge.txt.gz"))
}

get_filtered_raw_dge_path = function(pipeline_mgr) {
    return(dge_path_helper(pipeline_mgr, ".filtered.raw.dge.RDS"))
}

get_filtered_scaled_dge_path = function(pipeline_mgr) {
    return(dge_path_helper(pipeline_mgr, ".filtered.scaled.dge.RDS"))
}

#' Prefix for files in cluster<n> and components directory to distinguish between outputs in which number of
#' components was explicitly specified vs. determine using jackstraw.
make_subcluster_file_prefix=function(project.name, cluster.num, n.components, used_jackstraw) {
    file_prefix=paste0(project.name, ".cluster", cluster.num)
    if (used_jackstraw) {
        file_prefix=paste0(file_prefix, ".auto.")
    } else {
        file_prefix=paste0(file_prefix, ".n", n.components, ".")
    }
    return(file_prefix)
}

#' creates directory if it does not exist
get_cluster_directory=function(cluster.num, analysis_dir="./", ensure_exists=TRUE) {
    dir = ensure_trailing_slash(file.path(analysis_dir, paste0("cluster", cluster.num)))
    if (ensure_exists) {
        return(ensure_directory_exists(dir))
    } else {
        return(dir)
    }
}

ica_file_extension = "ICA.RDS"

get_file_from_ic_selection=function(project.name, cluster.num, n.components, used_jackstraw, dir, extension_without_dot) {
    return(paste0(ensure_trailing_slash(dir),
                    make_subcluster_file_prefix(
                            project.name=project.name,
                            cluster.num=cluster.num,
                            n.components=n.components,
                            used_jackstraw=used_jackstraw),
                    extension_without_dot))
}

#' Finds the latest subclustering for the given cluster with manual n.components selection.  If none exists, finds the subclustering
#' with automatic n.components selection.
#' @return list(used_jackstraw={TRUE or FALSE}, n.components={integer or NULL if used_jackstraw})
get_subcluster_ic_selection=function(project.name, cluster.num, analysis_dir="./") {
    # First try manually selected number of ICs
    prefix = paste0(project.name, ".cluster", cluster.num, ".n")
    suffix = paste0(".", ica_file_extension)

    components_dir = file.path(analysis_dir, "components")
    paths = list.files(path=components_dir, pattern=glob2rx(paste0(prefix, "*", ica_file_extension)),
                        full.names=TRUE)
    if (length(paths) == 0) {
        path = file.path(components_dir, paste0(make_subcluster_file_prefix(
                    project.name=project.name, cluster.num=cluster.num, n.components=NULL, used_jackstraw=TRUE), ica_file_extension))
        if (!file.exists(path)) {
            warning(paste0("No ICA file found in ", components_dir, " for cluster ", cluster.num), immediate.=TRUE)
            return(NULL)
        }
        return(list(used_jackstraw=TRUE, n.components=NULL))
    } else {
        times = sapply(paths, file.mtime)
        newest_path = paths[which.max(times)]
        newest_file = basename(newest_path)
        n_components_str = substr(newest_file, nchar(prefix) + 1, nchar(newest_file) - nchar(suffix))
        return(list(used_jackstraw=FALSE, n.components=strtoi(n_components_str)))
    }
}

#' @return a vector of filenames, with name of each element being the cluster number
get_curation_sheet_files=function(curation_sheets_dir) {
    curation.files=list.files(curation_sheets_dir, pattern="AnnotationSheet_Subcluster_.*[.]csv$")
    # Assign cluster numbers to list
    cluster.numbers = sapply(curation.files, function(curation.file) strsplit(removeSuffix(curation.file, '.csv'), split="_")[[1]][3])
    names(curation.files) = cluster.numbers
    return(curation.files)
}

#' @return Default path for a curation sheet, i.e. before it has been given a descriptive name by an investigator.
make_curation_sheet_path=function(curation_sheets_dir, cluster.num) {
    return(file.path(curation_sheets_dir, paste0("AnnotationSheet_Subcluster_", cluster.num, ".csv")))
}

replace_cluster_number_in_curation_sheet_filename=function(curation.file, cluster.num) {
    parts = strsplit(curation.file, split="_")[[1]]
    parts[3] = cluster.num
    return(paste(parts, collapse="_"))
}

#' @return Path to the components/*.ICA.RDS file with the latest n.components selection.
getLatestIcaRds=function(cluster.num, name, analysis_dir=".") {
    l = get_subcluster_ic_selection(project.name=name, cluster.num=cluster.num, analysis_dir=analysis_dir)
    if (is.null(l)) {
        return(NULL)
    }
    return(get_file_from_ic_selection(
                project.name=name,
                cluster.num=cluster.num,
                n.components=l$n.components,
                used_jackstraw=l$used_jackstraw,
                dir=file.path(analysis_dir, components_dir),
                extension_without_dot=ica_file_extension))
}

get_cluster_method_label=function(pipeline.parameters) {
    subcluster_parameters = pipeline.parameters$subcluster
    cluster.method = subcluster_parameters$cluster.method
    return(get_or_else(subcluster_parameters$cluster_method_label, cluster.method))
}

get_curated_results_path=function(pipeline.parameters) {
    return(paste0(pipeline.parameters$name, "_", get_cluster_method_label(pipeline.parameters),
        "_KDE_", pipeline.parameters$subcluster$center.KDE, ".curatedresults.RDS"))
}

make_subcluster_edge_path=function(pipeline.parameters, cluster.name, nn, resolution) {
    return(file.path(get_cluster_directory(cluster.name), louvain_subcluster_method,
            sprintf("%s.cluster%s.nn%d.resolution%s%s",
                    pipeline.parameters$name, cluster.name, nn, resolution, louvain_edges_extension)))
}

get_jar_path<-function(jarNameWithoutExtension) {
    file.path(system.file(package="DropSeq.icaCluster"),"java", paste0(jarNameWithoutExtension, ".jar"))

}