# MIT License
#
# Copyright 2017 Broad Institute
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Command-line functions for parallelizing differential expression computation



means_and_sums_extension = ".means_and_sums.RDS"

#' Shared initialization for all the CLPs
#' @return list(pipeline_mgr, pipeline.parameters, idents, scaled_dge (or NA), raw_dge (or NA) )
DiffExpInitClp=function(yamlpath=NULL, include_dge=TRUE) {
    pipeline_mgr = PipelineManager$new(
        subdirectory_names=workflow_directories,
        yamlpath=yamlpath,
        stages=NA,
        defaults=default_settings_no_history)
    pipeline.parameters = pipeline_mgr$pipeline.parameters
    idents = readRDS(file.path(pipeline_mgr$directories$assign, paste0(pipeline.parameters$name,".subcluster.assign.RDS")))
    if (include_dge) {
        filtered_scaled_dge = readRDS(get_filtered_scaled_dge_path(pipeline_mgr))
        filtered_raw_dge = readRDS(get_filtered_raw_dge_path(pipeline_mgr))
    } else {
        filtered_scaled_dge=NA
        filtered_raw_dge=NA
    }
    return(list(pipeline_mgr=pipeline_mgr, pipeline.parameters=pipeline.parameters, idents=idents,
            scaled_dge=filtered_scaled_dge, raw_dge=filtered_raw_dge))
}

#' Compute row means and row sums for a single cluster and write to disk
#' @param cluster_index: not a cluster name, but rather a number between 1 and number-of-clusters
#' @param output_file: default is <project>/diffexp/<cluster_name>.means_and_sums.RDS
DiffExpMeansAndSumsClp=function(cluster_index, yamlpath=NULL, output_file=NULL) {
    l = DiffExpInitClp(yamlpath=yamlpath)
    cluster_name = levels(l$idents)[cluster_index]
    cat_with_time(paste0("Computing means and sums for ", cluster_name))
    meanexp_digsums = compute_diffexp_mean_and_sum(norm=l$scaled_dge, dge=l$raw_dge, cells=names(l$idents)[which(l$idents == cluster_name)])
    if (is.null(output_file)) {
        output_file = file.path(l$pipeline_mgr$directories$diffexp, paste0(cluster_name, means_and_sums_extension))
    }
    saveRDS(meanexp_digsums, file=output_file)
    cat_with_time(paste0("Wrote ", output_file))
}

#' Convert a single number to a pair of cluster names
#' @return list(cluster_name1, cluster_name2) where the first name is always the "lower" name
convert_cluster_pair_index=function(cluster_pair_index, cluster_names) {
    n = length(cluster_names)
    if (cluster_pair_index > n * (n-1)/2) stop(sprintf("cluster_pair_index %d out of range for n=%d", cluster_pair_index, length(cluster_names)))
    i1 = 1
    row_length = n-1
    while (cluster_pair_index > row_length) {
        i1 = i1 + 1
        cluster_pair_index = cluster_pair_index - row_length
        row_length = row_length - 1
    }
    return(list(cluster_name1=cluster_names[i1], cluster_name2=cluster_names[cluster_pair_index + i1]))
}

#' Compute pairwise differential expression
#' @param cluster_pair_index a number 1:n*(n-1)/2 where n==number of clusters, corresponding to a pair of clusters
#' @param output_file default is <project>diffexp/<c1>_<c2>_pairwiseMarkers.RDS
DiffExpPairwiseClp=function(cluster_pair_index, yamlpath=NULL, output_file=NULL) {
    l = DiffExpInitClp(yamlpath=yamlpath)
    cluster_names = convert_cluster_pair_index(cluster_pair_index, cluster_names=levels(l$idents))
    cat_with_time(sprintf("Computing differential expression for %s and %s", cluster_names[[1]], cluster_names[[2]]))
    de = diffexp(norm=l$scaled_dge, dge=l$raw_dge, cells=names(l$idents)[which(l$idents == cluster_names[[1]])],
            cells.comp=names(l$idents)[which(l$idents == cluster_names[[2]])],
            meanexp_digsums=readRDS(file.path(l$pipeline_mgr$directories$diffexp, paste0(cluster_names[[1]], means_and_sums_extension))),
            meanexp_digsums.comp=readRDS(file.path(l$pipeline_mgr$directories$diffexp, paste0(cluster_names[[2]], means_and_sums_extension))),
            only.pos=FALSE, round.FOLDch=FALSE, # Don't round for reproducibility when inverting
            FOLDch.thresh=l$pipeline.parameters$differential_expression$FOLDch.threshold,
            pval.thresh=l$pipeline.parameters$differential_expression$pval.threshold)
    if (is.null(output_file)) {
        output_file = file.path(l$pipeline_mgr$directories$diffexp, paste0(cluster_names[[1]], "_", cluster_names[[2]], pairwise_markers_extension))
    }
    saveRDS(de, file=output_file)
    cat_with_time(paste0("Wrote ", output_file))
}

#' Load all the pairwise differential expression objects into the big list of lists.  Each will need to be inverted
#' in order to completely populate the list of lists.
#' @param output_file default <project>_pairwiseMarkers.RDS
DiffExpAggregatePairwiseClp=function(yamlpath=NULL, output_file=NULL) {
    l = DiffExpInitClp(yamlpath=yamlpath, include_dge=FALSE)
    if (is.null(output_file)) {
        output_file = paste0(l$pipeline.parameters$name, pairwise_markers_extension)
    }
    cat_with_time(paste0("Aggregating pairwise differential expression into ", output_file))
    cluster_names = levels(l$idents)
    # Initialize the list of lists with NAs
    list_of_lists = lapply(cluster_names, function(cluster_name) {
        clusters.comp = setdiff(cluster_names, cluster_name)
        l = as.list(rep(NA, length(clusters.comp)))
        names(l) = clusters.comp
        return(l)
    })
    names(list_of_lists) = cluster_names
    # for loops instead of lapply in order to access list_of_lists more easily
    for (cluster1 in names(list_of_lists)) {
        cat_with_time(paste0("Populating ", cluster1))
        for (cluster2 in names(list_of_lists[[cluster1]])) {
            index1 = which(levels(l$idents) == cluster1)
            index2 = which(levels(l$idents) == cluster2)
            list_of_lists[[cluster1]][[cluster2]] = if (index1 < index2) {
                    readRDS(file.path(l$pipeline_mgr$directories$diffexp, paste0(cluster1, "_", cluster2, pairwise_markers_extension)))
            } else if (index2 < index1) {
                inverse = list_of_lists[[cluster2]][[cluster1]]
                inverse$FOLDch = 1/inverse$FOLDch
                swap = inverse$pct.1
                inverse$pct.1 = inverse$pct.2
                inverse$pct.2 = swap
                inverse[nrow(inverse):1, ]
            } else {
                stop(paste("ERROR: should not diffexp cluster against itself: ", index1, cluster1, cluster2))
            }
        }
    }
    for (i in seq_along(list_of_lists)) for (j in seq_along(list_of_lists[[i]])) {
        list_of_lists[[i]][[j]]$FOLDch = round(list_of_lists[[i]][[j]]$FOLDch, diffexp_FOLDch_round_digits)
    }
    saveRDS(list_of_lists, file=output_file)
    cat_with_time(paste0("Wrote ", output_file))
}

#' Compute differential expression for a single cluster compared to all cells not in that cluster and write to disk
#' @param cluster_index: not a cluster name, but rather a number between 1 and number-of-clusters
#' @param output_file: default is <project>/diffexp/<cluster_name>_clusterMarkers.RDS
DiffExpClusterMarkersClp=function(cluster_index, yamlpath=NULL, output_file=NULL) {
    l = DiffExpInitClp(yamlpath=yamlpath)
    cluster_name = levels(l$idents)[cluster_index]
    cat_with_time(paste0("Computing cluster markers for ", cluster_name))
    meanexp_digsums=readRDS(file.path(l$pipeline_mgr$directories$diffexp, paste0(cluster_name, means_and_sums_extension)))
    cells=names(l$idents)[which(l$idents==cluster_name)]
    cells.comp=setdiff(names(l$idents),cells)
    de = diffexp(norm=l$scaled_dge, dge=l$raw_dge, cells=cells, cells.comp=cells.comp,
            meanexp_digsums=meanexp_digsums, only.pos=FALSE,
            FOLDch.thresh=l$pipeline.parameters$differential_expression$FOLDch.threshold,
            pval.thresh=l$pipeline.parameters$differential_expression$pval.threshold)
    if (is.null(output_file)) {
        output_file = file.path(l$pipeline_mgr$directories$diffexp, paste0(cluster_name, cluster_markers_extension))
    }
    saveRDS(de, file=output_file)
    cat_with_time(paste0("Wrote ", output_file))
}


#' Load all the cluster markers differential expression objects and write a list of all of them
#' @param output_file default <project>_clusterMarkers.RDS
DiffExpAggregateClusterMarkersClp=function(yamlpath=NULL, output_file=NULL) {
    l = DiffExpInitClp(yamlpath=yamlpath, include_dge=FALSE)
    if (is.null(output_file)) {
        output_file = paste0(l$pipeline.parameters$name, cluster_markers_extension)
    }
    cat_with_time(paste0("Aggregating cluster markers differential expression into ", output_file))
    cluster_names = levels(l$idents)
    all_de = lapply(cluster_names, function(cluster_name) {
        readRDS(file.path(l$pipeline_mgr$directories$diffexp, paste0(cluster_name, cluster_markers_extension)))
    })
    names(all_de) = cluster_names
    saveRDS(all_de, file=output_file)
    cat_with_time(paste0("Wrote ", output_file))
}

