# MIT License
#
# Copyright 2017 Broad Institute
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

#' This function assumes x, y and dge are all in the same order
#' Colors the scatterplot based on cols.use
fplot_basic<-function(x, y, dge, title,cols.use=heat.colors(10),pt.size=0.7,pch.use=20) {
    if (length(x) != length(y) || length(x) != length(dge)) {
        stop("Different lengths of vectors passed to fplot_basic")
    }
    data.cut=as.numeric(cut(as.numeric(dge),breaks=length(cols.use)))
    data.col=rev(cols.use)[data.cut]
    plot(x, y, col=data.col, cex=pt.size, pch=pch.use, main=title)
}

#' Produce a single t-SNE plot colored by gene expression
#' In addition to functionality of fplot_basic, selects only the genes that appear in factor
fplot<-function(tsne,factor,title,cols.use=heat.colors(10),pt.size=0.7,pch.use=20) {
    c=intersect(rownames(tsne),names(factor))
    #  data.use=log(factor[c]*10000+1)
    data.use = factor[c]
    fplot_basic(x=tsne[c,1],y=tsne[c,2], dge=data.use, title=title, cols.use=cols.use, pt.size=pt.size, pch.use=pch.use)
}

default_num_quanta=250

#' Adjust the values in the input tsne matrix so that there are no more than num_quanta different values
#' in each dimension.  For each dimension, a round_factor is computed by dividing the range of that dimension
#' by num_quanta.  Each value is then adjusted slightly so that it is an integral multiple of the round_factor.
#' @param tsne matrix to be adjusted.  
#' @param num_quanta Max number of different values to be allowed in each dimension in the returned matrix.
#' @return a matrix similar to the input matrix but with minor adjustments to the values.
quantize_tsne=function(tsne, num_quanta=default_num_quanta) {
    for (colnum in 1:ncol(tsne)) {
        therange = range(tsne[, colnum])
        round_factor = (therange[2] - therange[1])/num_quanta
        tsne[, colnum] = tsne[, colnum] - (tsne[, colnum] %% round_factor)
    }
    return(tsne)
}

#' @param tsne matrix with 2 columns that are x and y coordinates of t-SNE, and rownames == gene names
#' @param dge vector of log-transformed DGE, with rownames == gene names
# @return data.frame(x, y, dge) with no non-unique rows.  The order of the rows is randomized.
uniquify_tsne=function(tsne, factor) {
    c=intersect(rownames(tsne),names(factor))
    df = data.frame(x=tsne[c,1], y=tsne[c,2], dge=factor[c])
    rownames(df) = c
    before_reduction=nrow(df)
    df = unique(df)
    cat_with_time(sprintf("Reduced global tSNE from %d to %d", before_reduction, nrow(df)))

    # Because there are more zero-DGE elements than non-zero elements, on average the rounding will cause a non-zero
    # element to be both preceded and followed by zero element(s) at the same position.  
    # Because unique() returns rows in the order of the first occurrence of a tuple, a zero-DGE element
    # will precede a non-zero element at the same position most of the time.  
    # Since plotting is done in order, the non-zero elements tend to overwrite the zero elements.
    # So, randomize the order after uniquing in order to eliminate this bias.
    df = df[sample(nrow(df)), ]
    return(df)
}

# Evan decided we shouldn't use this functionality, because it tends to reduce the number of zero elements more
# than non-zero, and therefore presents a somewhat biased view of the gene expression.
fplot_reduced=function(tsne, factor, title) {
    u_tsne = uniquify_tsne(tsne, factor)
    fplot_basic(x=u_tsne$x, y=u_tsne$y, dge=u_tsne$dge, title=title)
}

default_featureplot_num_genes=10
# for now the default is PDF only, since it take so long to plot multiple formats.
# TODO: Figure out how to stitch multiple PNGs into a PDF (or something else).
default_featureplot_formats=c('pdf')

PlotIcFeaturesClp=function(yamlpath=NULL,
        num_genes=default_featureplot_num_genes,
        formats=default_featureplot_formats,
        selected_clusters=NULL,
        output_directory=NULL) {
    pipeline_mgr = PipelineManager$new(
        subdirectory_names=workflow_directories,
        yamlpath=yamlpath,
        stages=NA,
        defaults=default_settings_no_history)
    PlotIcFeatures(pipeline_mgr, num_genes=num_genes, formats=formats, selected_clusters=selected_clusters, output_directory=output_directory)
}

#' It is assumed that working directory is analysis directory
PlotIcFeatures=function(pipeline_mgr,
            num_genes=default_featureplot_num_genes,
            formats=default_featureplot_formats,
            selected_clusters=NULL,
            output_directory=NULL, ...) {
    pipeline.parameters = pipeline_mgr$pipeline.parameters

    norm.dge = log_transform_dge(readRDS(get_filtered_scaled_dge_path(pipeline_mgr)))

    clusters = levels(readRDS(paste0(pipeline_mgr$directories$assign, pipeline.parameters$name, ".cluster.assign.RDS")))
    if (!is.null(selected_clusters)) {
        clusters = clusters[selected_clusters]
    }

    tsne.global <- readRDS(paste0(pipeline_mgr$directories$tSNE, pipeline.parameters$name, tsne_xy_extension))
    #tsne.global = quantize_tsne(tsne.global)
    dge.global <- norm.dge[, rownames(tsne.global)]

    if (is.null(output_directory)) {
        output_directory = pipeline_mgr$directories$featureplots
    }
    feature_plotters = list(pdf=PdfFeaturePlotter$new(pipeline.parameters$name, output_directory),
                            jpg=JpgFeaturePlotter$new(pipeline.parameters$name, output_directory, ...),
                            png=PngFeaturePlotter$new(pipeline.parameters$name, output_directory, ...))
    feature_plotters=feature_plotters[formats]

    for (cluster_name in clusters) {
        ic_selection = get_subcluster_ic_selection(pipeline.parameters$name, cluster_name)
        if (is.null(ic_selection)) {
            cat_with_time(paste0("Cluster ", cluster_name, " did not subcluster"))
            next
        }
        tsne.sub = readRDS(get_file_from_ic_selection(
            project.name=pipeline.parameters$name,
            cluster.num=cluster_name,
            n.components=ic_selection$n.components,
            used_jackstraw=ic_selection$used_jackstraw,
            dir=get_cluster_directory(cluster_name),
            extension_without_dot=cluster_tSNExy_extension))

        dge.sub <- norm.dge[, rownames(tsne.sub)]

        gene_loadings = readRDS(getLatestIcaRds(cluster_name, pipeline.parameters$name))$gene_loadings

        for (plotter in feature_plotters) {
            # reset random number generator in case outputing multiple types of plots, so that
            # random order of point plotting is consistent across plot types.
            set.seed(1)
            for (global in c(TRUE, FALSE)) {
                tsne = if (global) {tsne.global} else {tsne.sub}
                dge = if (global) {dge.global} else {dge.sub}

                plotter$start(cluster_name, global)
                for (ic.num in 1:ncol(gene_loadings)) {
                    plotter$start_ic(cluster_name, ic.num, global)
                    pos_and_neg_loadings = get_pos_and_neg_loading_lists_for_ic(gene_loadings[,ic.num], num_genes=num_genes)
                    positive_loading = pos_and_neg_loadings$positive_loading
                    negative_loading = pos_and_neg_loadings$negative_loading
                    for (i in seq_along(positive_loading)) {
                        gene = positive_loading[[i]]
                        plotter$start_ic_gene(cluster_name, gene, ic.num, i, global)
                        fplot(tsne, dge[gene, ],
                                make_ic_plot_title(cluster_name, ic.num, i, gene, positive=TRUE, global=TRUE))
                        plotter$end_ic_gene()
                    }
                    for (i in seq_along(negative_loading)) {
                        gene = negative_loading[[i]]
                        plotter$start_ic_gene(cluster_name, gene, ic.num, -i, global)
                        fplot(tsne, dge[gene, ],
                                make_ic_plot_title(cluster_name, ic.num, i, gene, positive=FALSE, global=TRUE))
                        plotter$end_ic_gene()
                    }
                    plotter$end_ic()
                }
                plotter$end()
            }
        }
    }
    cat_with_time("Done plotting IC features")
}

make_featureplot_cluster_dir=function(featureplots_dir, cluster_name) {
    return(ensure_directory_exists(file.path(featureplots_dir, paste0("cluster", get_round_1_cluster_name(cluster_name)))))
}

make_ic_plot_title=function(cluster.num, ic.num, loading.index, gene, positive, global) {
    return(sprintf("%s t-SNE; Cluster %s; IC %d; %s %d; %s", if (global) "global" else "cluster",
            cluster.num, ic.num, if (positive) "+" else "-", loading.index, gene))
}

log_transform_dge=function(norm.dge) {
    # Keep sparse if possible
    if ('dgTMatrix' %in% class(norm.dge)) {
        # Note that this is allowed because for the transformation below, f(0) => 0
        norm.dge@x = log(10000*norm.dge@x + 1)
    } else {
        norm.dge = log(10000*norm.dge + 1)
    }
    return(norm.dge)
}

PdfFeaturePlotter = R6Class("PdfFeaturePlotter",
    public = list(
        initialize = function(project_name, featureplots_dir) {
            private$project_name=project_name
            private$featureplots_dir=featureplots_dir
        },

        # These do nothing because each IC goes into a separate file
        start = function(cluster_name, global) {},
        end = function() {},

        start_ic = function(cluster_name, ic.num, global) {
            private$start_ic_pdf(cluster_name, ic.num, if (global) {private$global_extension} else {private$local_extension} )
        },
        end_ic = function() {dev.off()},

        start_markers = function(cluster_name, global) {
            private$start_markers_pdf(cluster_name, if (global) {private$global_extension} else {private$local_extension})
        },
        end_markers = function() {dev.off()},



        # these do nothing because for PDF, the output for all genes goes into the same file.
        start_ic_gene=function(cluster_name, gene, ic.num, index, global) {},
        end_ic_gene=function() {},
        start_marker_gene=function(cluster_name, gene, global) {},
        end_marker_gene=function() {}
    ),
    private = list(
        project_name=NULL,
        featureplots_dir=NULL,
        global_extension=".global.pdf",
        local_extension=".local.pdf",
        start_ic_pdf=function(cluster_name, ic.num, extension) {
            outdir = make_featureplot_cluster_dir(private$featureplots_dir, cluster_name)
            file = file.path(outdir, paste0(private$project_name, ".cluster", cluster_name, ".ic", ic.num, extension))
            cat_with_time(paste0("Writing ", file))
            pdf(file)
        },
        start_markers_pdf=function(cluster_name, extension) {
            outdir = make_featureplot_cluster_dir(private$featureplots_dir, cluster_name)
            file = file.path(outdir, paste0(private$project_name, ".cluster", cluster_name, ".markers", extension))
            cat_with_time(paste0("Writing ", file))
            pdf(file)
        }
    )
)


JpgFeaturePlotter = R6Class("JpgFeaturePlotter",
    public = list(
        initialize = function(project_name, featureplots_dir, height=300, width=1000) {
            private$project_name=project_name
            private$featureplots_dir=featureplots_dir
            private$height=height
            private$width=width
        },

        # these do nothing because for JPG, the output for each gene goes into a separate file.
        start = function(cluster_name, global) {},
        end = function() {},
        start_ic = function(cluster_name, ic.num, global) {},
        end_ic = function() {},
        start_markers = function(cluster_name, global) {},
        end_markers = function() {},

        start_ic_gene=function(cluster_name, gene, ic.num, index, global) {
            private$open_jpeg(cluster_name, make_ic_filename( cluster_name, gene, ic.num, index, global))
        },
        end_ic_gene=function() {dev.off()},
        start_marker_gene=function(cluster_name, gene, global) {
            private$open_jpeg(cluster_name, make_marker_filename( cluster_name, gene, global))
        },
        end_marker_gene=function() {dev.off()}

    ),
    private = list(
        project_name=NULL,
        featureplots_dir=NULL,
        height=NULL,
        width=NULL,
        global_extension=".global.jpg",
        local_extension=".local.jpg",

        make_ic_filename=function(cluster_name, ic.num, index, gene, global) {
            extension = if (global) {private$global_extension} else {private$local_extension}
            return(paste0(private$project_name, ".cluster", cluster_name, ".ic", ic.num, ".", index, ".", gene, extension))
        },
        make_marker_filename=function(cluster_name, gene, global) {
            extension = if (global) {private$global_extension} else {private$local_extension}
            return(paste0(private$project_name, ".cluster", cluster_name, ".", gene, extension))
        },
        open_jpeg=function(cluster_name, filename) {
            outdir = make_featureplot_cluster_dir(private$featureplots_dir, cluster_name)
            file = file.path(outdir, filename)
            cat_with_time(paste0("Writing ", file))
            jpeg(file, height=private$height, width=private$width)}
    )
)


PngFeaturePlotter = R6Class("JpgFeaturePlotter",
    public = list(
        initialize = function(project_name, featureplots_dir, height=480, width=480) {
            private$project_name=project_name
            private$featureplots_dir=featureplots_dir
            private$height=height
            private$width=width
        },

        # these do nothing because for JPG, the output for each gene goes into a separate file.
        start = function(cluster_name, global) {},
        end = function() {},
        start_ic = function(cluster_name, ic.num, global) {},
        end_ic = function() {},
        start_markers = function(cluster_name, global) {},
        end_markers = function() {},

        start_ic_gene=function(cluster_name, gene, ic.num, index, global) {
            private$open_png(cluster_name, make_ic_filename( cluster_name, gene, ic.num, index, global))
        },
        end_ic_gene=function() {dev.off()},
        start_marker_gene=function(cluster_name, gene, global) {
            private$open_png(cluster_name, make_marker_filename( cluster_name, gene, global))
        },
        end_marker_gene=function() {dev.off()}
    ),
    private = list(
        project_name=NULL,
        featureplots_dir=NULL,
        height=NULL,
        width=NULL,
        global_extension=".global.png",
        local_extension=".local.png",

        make_ic_filename=function(cluster_name, ic.num, index, gene, global) {
            extension = if (global) {private$global_extension} else {private$local_extension}
            return(paste0(private$project_name, ".cluster", cluster_name, ".ic", ic.num, ".", index, ".", gene, extension))
        },
        make_marker_filename=function(cluster_name, gene, global) {
            extension = if (global) {private$global_extension} else {private$local_extension}
            return(paste0(private$project_name, ".cluster", cluster_name, ".", gene, extension))
        },
        open_png=function(cluster_name, filename) {
            outdir = make_featureplot_cluster_dir(private$featureplots_dir, cluster_name)
            file = file.path(outdir, filename)
            cat_with_time(paste0("Writing ", file))
            png(file, height=private$height, width=private$width)}
    )
)


#' Load gene lists from the yaml file, and produce a PDF for each gene list.
#' These are written to output_directory, with names <gene_set_name>.globaltSNE.pdf
PlotCanonicalGenes<-function(tsne, norm.dge, output_directory, canonical_genes_yaml_file, copy.yaml=TRUE) {
    output_directory = ensure_directory_exists(output_directory)

    gene_sets = yaml.load_file(canonical_genes_yaml_file)$GeneSets
    if (is.null(gene_sets)) {
        stop(paste0(canonical_genes_yaml_file, " does not contains a 'GeneSets' element"))
    }
    if (copy.yaml) {
        if (!file.copy(canonical_genes_yaml_file, output_directory, overwrite=TRUE)) {
            stop(paste0("Could not copy ", canonical_genes_yaml_file, " to ", output_directory))
        }
    }
    norm.dge <- norm.dge[, rownames(tsne)]
    norm.dge = log_transform_dge(norm.dge)

    for (gene_set_name in names(gene_sets)) {
        geneSet = gene_sets[[gene_set_name]]
        outpdf = file.path(output_directory, paste0(gene_set_name, ".globaltSNE.pdf"))
        cat_with_time(paste0("Writing ", outpdf))
        goodGenes = geneSet[sapply(geneSet, function(x) {x %in% rownames(norm.dge)})]
        badGenes = geneSet[sapply(geneSet, function(x) {!(x %in% rownames(norm.dge))})]
        if (length(badGenes) > 0) {
            cat_with_time(paste0("WARNING: Could not find genes '", paste(badGenes, collapse="', '"), "' when producing ", outpdf))
        }
        pdf(outpdf)
        for (gene in goodGenes) {
            fplot(tsne, norm.dge[gene, ], title=gene)
        }
        dev.off()
    }
    cat_with_time("Done plotting canonical genes")
}

PlotCanonicalGenesClp=function(canonical_genes_yaml_file, project_yaml, output_directory=NULL) {
    pipeline_mgr = PipelineManager$new(
        subdirectory_names=workflow_directories,
        yamlpath=project_yaml,
        stages=NA,
        defaults=default_settings_no_history)
    pipeline.parameters = pipeline_mgr$pipeline.parameters
    norm.dge = readRDS(get_filtered_scaled_dge_path(pipeline_mgr))
    tsne <- readRDS(paste0(pipeline_mgr$directories$tSNE, pipeline.parameters$name, tsne_xy_extension))
    if (is.null(output_directory)) {
        output_directory = file.path(pipeline_mgr$directories$featureplots, canonical_plots_dir)
    }
    PlotCanonicalGenes(tsne, norm.dge, output_directory, canonical_genes_yaml_file)
}

default_round1_cluster_markers_num_genes=25
default_subcluster_markers_num_genes=25

PlotClusterMarkersClp=function(round1,
                    yamlpath=NULL,
                    num_genes=NULL,
                    formats=default_featureplot_formats,
                    selected_clusters=NULL,
                    output_directory=NULL) {
    pipeline_mgr = PipelineManager$new(
        subdirectory_names=workflow_directories,
        yamlpath=yamlpath,
        stages=NA,
        defaults=default_settings_no_history)
    if (is.null(output_directory)) {
        output_directory = pipeline_mgr$directories$featureplots
    }
    cluster_markers = if (round1) {
        readRDS(paste0(pipeline_mgr$pipeline.parameters$name,firstround_label, cluster_markers_extension))
    } else {
        readRDS(paste0(pipeline_mgr$pipeline.parameters$name, cluster_markers_extension))
    }
    if (is.null(num_genes)) {
        num_genes = if (round1) {default_round1_cluster_markers_num_genes} else {default_subcluster_markers_num_genes}
    }
    if (!is.null(selected_clusters)) {
        cluster_markers = cluster_markers[selected_clusters]
        for (i in seq_along(selected_clusters)) {
            if (is.null(cluster_markers[[i]])) {
                cluster_level = if (round1) "round 1 " else "sub"
                selector_type = if (is.character(selected_clusters[[i]])) "name" else "index"
                stop(sprintf("%s is not a valid %scluster %s", selected_clusters[[i]], cluster_level, selector_type))
            }
        }
    }
    PlotClusterMarkers(pipeline_mgr=pipeline_mgr, cluster_markers=cluster_markers, num_genes=num_genes, formats=formats, output_directory=output_directory)
    cat_with_time(sprintf("Done plotting %s cluster markers", if (round1) "first-round " else "sub"))
}


PlotClusterMarkers=function(pipeline_mgr,
                    cluster_markers,
                    num_genes,
                    formats,
                    output_directory, ...) {
    pipeline.parameters = pipeline_mgr$pipeline.parameters
    norm.dge = log_transform_dge(readRDS(get_filtered_scaled_dge_path(pipeline_mgr)))
    tsne.global <- readRDS(paste0(pipeline_mgr$directories$tSNE, pipeline.parameters$name, tsne_xy_extension))
    #tsne.global = quantize_tsne(tsne.global)
    dge.global <- norm.dge[, rownames(tsne.global)]

    feature_plotters = list(pdf=PdfFeaturePlotter$new(pipeline.parameters$name, output_directory),
                            jpg=JpgFeaturePlotter$new(pipeline.parameters$name, output_directory, ...),
                            png=PngFeaturePlotter$new(pipeline.parameters$name, output_directory, ...))
    feature_plotters=feature_plotters[formats]

    for (cluster_name in names(cluster_markers)) {
        diffexp_df = cluster_markers[[cluster_name]]
        if (anyNA(diffexp_df$pval)) {
            cat_with_time(sprintf("No differential expression for cluster %s so not feature plotting this cluster",cluster_name))
        } else {
            round1_cluster_name = get_round_1_cluster_name(cluster_name)
            subcluster.tsne.path = file.path(pipeline_mgr$directories$tSNE,paste0(pipeline.parameters$name, '.cluster', round1_cluster_name,curated_tsne_extension))
            tsne.sub = if (file.exists(subcluster.tsne.path)) readRDS(subcluster.tsne.path) else NULL

            markers_to_print = get_cluster_markers(diffexp_df, num_genes)
            if (is.null(tsne.sub)) {
                cat_with_time(paste0(cluster_name, " was not subclustered so no local feature plots"))
                global_local = TRUE
            } else {
                dge.sub <- norm.dge[, rownames(tsne.sub)]
                global_local = c(TRUE, FALSE)
            }
            for (plotter in feature_plotters) {
                for (global in global_local) {
                    plotter$start(cluster_name, global)
                    plotter$start_markers(cluster_name, global)
                    tsne = if (global) {tsne.global} else {tsne.sub}
                    dge = if (global) {dge.global} else {dge.sub}
                    for (gene in rownames(markers_to_print)) {
                        plotter$start_marker_gene(cluster_name, gene, global)
                        fplot(tsne, dge[gene,],
                                make_marker_plot_title(cluster_name, gene, markers_to_print[gene,]$pval,  markers_to_print[gene,]$FOLDch, global))
                        plotter$end_marker_gene()
                    }
                    plotter$end_markers()
                    plotter$end()
                }
            }
        }
    }
}

make_marker_plot_title=function(cluster_name, gene, pval, FOLDch, global) {
    return(sprintf("%s t-SNE; Cluster %s; gene %s; p-value %.1g; FOLDch %.3g", if (global) "global" else "cluster",
                cluster_name, gene, pval, FOLDch))
}



get_cluster_markers=function(diffexp_df, num_genes) {
    pos_diffexp = diffexp_df[diffexp_df$FOLDch >1, ]
    # Annoyingly, must specify radix method if sort direction is not the same for all.
    pos_diffexp = pos_diffexp[order(pos_diffexp$pval, pos_diffexp$FOLDch, decreasing=c(FALSE, TRUE), method='radix'), ]
    neg_diffexp = diffexp_df[diffexp_df$FOLDch <= 1, ]
    neg_diffexp = neg_diffexp[order(neg_diffexp$pval, neg_diffexp$FOLDch), ]
    half_num_genes = as.integer(num_genes/2)
    if (nrow(pos_diffexp) < half_num_genes) {
        num_pos = nrow(pos_diffexp)
        num_neg = min(nrow(neg_diffexp), num_genes - num_pos)
    } else if (nrow(neg_diffexp) < half_num_genes) {
        num_neg = nrow(neg_diffexp)
        num_pos = min(nrow(pos_diffexp), num_genes - num_neg)
    } else {
        num_pos = num_genes - half_num_genes
        num_neg = half_num_genes
    }
    if (num_neg == 0) {
        return(pos_diffexp[1:num_pos,])
    } else if (num_pos == 0) {
        return(neg_diffexp[1:num_neg, ])
    } else {
        return(rbind(pos_diffexp[1:num_pos,], neg_diffexp[1:num_neg, ]))
    }
}

get_round_1_cluster_name=function(cluster_name) {
    return(strsplit(cluster_name, '-')[[1]][1])
}
