# MIT License
#
# Copyright 2017 Broad Institute
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# 2nd-generation framework for feature plotting.
# Eventually, the functions in featureplot.R should move to this framework.

TwoLevelPlotter = R6Class("TwoLevelPlotter",
    public = list(
        initialize = function(primitive_plotter) {
            private$primitive_plotter=primitive_plotter
        },

        start_plot_group = function(filename_without_extension) {
            if (private$in_group) stop("Illegal state")
            if (private$in_single_plot) stop("Illegal state")
            if (!is.null(private$current_file)) stop("Illegal state")
            if (!is.null(filename_without_extension)) {
                private$in_group = TRUE
                private$start(private$current_file)
            }
        },
        end_plot_group = function() {
            if (private$in_single_plot) stop("Illegal state")
            if (private$in_group) {
                if (is.null(private$current_file)) stop("Illegal state")
                private$end()
                private$in_group = FALSE
            }
        },

        start_single_plot = function(filename_without_extension) {
            if (private$in_single_plot) stop("Illegal state")
            if (!is.null(filename_without_extension)) {
                if (private$in_group) stop("Illegal state")
                if (!is.null(private$current_file)) stop("Illegal state")
                private$in_single_plot = TRUE
                private$start(filename_without_extension)
            }
        },
        end_single_plot = function() {
            if (private$in_single_plot) {
                if (is.null(private$current_file)) stop("Illegal state")
                private$end()
                private$in_single_plot = FALSE
            }
        }

    ),
    private = list(
        primitive_plotter=NULL,
        in_group=FALSE,
        in_single_plot=FALSE,
        current_file=NULL,
        start=function(filename_without_extension) {
            private$current_file = filename_without_extension
            cat_with_time(paste0("Writing ", private$current_file))
            ensure_directory_exists(dirname(private$current_file))
            private$primitive_plotter$start(private$current_file)
        },
        end=function() {
            cat_with_time(paste0("Done writing ", private$current_file))
            private$current_file = NULL
            private$primitive_plotter$end()
        }
    )
)

PdfPrimitivePlotter = R6Class("PdfPrimitivePlotter",
    public = list(
        initialize = function(...) {
            private$varargs = list(...)
        },
        start = function(filename_without_extension) {
            args = private$varargs
            args$file = paste0(filename_without_extension, ".pdf")
            do.call("pdf", args)
        },
        end = function() {
            dev.off()
        }
    ),
    private = list(
        varargs = NULL
    )
)

EpsPrimitivePlotter = R6Class("EpsPrimitivePlotter",
    public = list(
        initialize = function(...) {
            private$varargs = list(...)
        },
        start = function(filename_without_extension) {
            args = private$varargs
            args$file = paste0(filename_without_extension, ".eps")
            setEPS()
            do.call("postscript", args)
        },
        end = function() {
            dev.off()
        }
    ),
    private = list(
        varargs = NULL
    )
)

NakeIcPlotNameGenerator = R6Class("NakeIcPlotNameGenerator",
    public = list(
        initialize = function(output_dir, project_name, use_curated_tsne) {
            private$output_dir = output_dir
            private$project_name = project_name
            private$tsne_type = if (use_curated_tsne) "curated" else "raw"
        },
        cluster_plot = function() {return(file.path(private$output_dir, paste0(private$project_name, ".", 
                     private$tsne_type, "_tSNE")))},
        #' ic_status: one of real, artifact, doublet, outlier
        ic_plot_group = function(ic_status, cluster_num=NULL) {return(NULL)}, # Currently only supporting one plot per file
        ic_single_plot = function(ic_status, ic_num, cluster_num=NULL) {
            private$validate_ic_status(ic_status)
            if (is.null(cluster_num)) {
                outer_dir = private$output_dir
            } else {
                outer_dir = file.path(private$output_dir, paste0('cluster', cluster_num))
            }
            
            dir = file.path(outer_dir, paste0(ic_status, "_ICs"))
            return(file.path(dir, paste0(private$project_name, ".", private$ic_status_file_slugs[[ic_status]],
                "IC", ic_num, ".",  private$tsne_type, "_tSNE")))
        }

    ),
    private = list(
        output_dir = NULL,
        project_name = NULL,
        tsne_type = NULL,
        # These get plugged into filenames
        ic_status_file_slugs = list(real='', artifact='artifact_', doublet='doublet_', outlier='outlier_'),
        validate_ic_status = function(ic_status) {
            if (!is.element(ic_status, names(private$ic_status_file_slugs))) {
                stop(paste("Unrecognized IC type", ic_status))
            }
        }
    )

)

#' additional args passed through to postscript()
GenerateNakedIcPlotsClp=function(output_directory, yamlpath, use_curated_tsne=FALSE, ...) {
    # pipeline manager will change current directory, so make sure output_directory is absolute path
    output_directory = normalizePath(ensure_directory_exists(output_directory))
    pipeline_mgr = PipelineManager$new(
        subdirectory_names=workflow_directories,
        yamlpath=yamlpath,
        stages=all_stages,
        defaults=default_settings_no_history)
    pipeline.parameters = pipeline_mgr$pipeline.parameters
    filename_generator = NakeIcPlotNameGenerator$new(output_directory, pipeline.parameters$name, use_curated_tsne)
    plotter = TwoLevelPlotter$new(EpsPrimitivePlotter$new(...))

    # handle legacy case when auto_tSNE file does not exist
    round1_tsne_path = NA
    if (!use_curated_tsne) {
        round1_tsne_path = file.path(pipeline_mgr$directories$tSNE, paste0(pipeline.parameters$name, auto_tsne_xy_extension))
        if (!file.exists(round1_tsne_path)) {
            round1_tsne_path = NA
        }
    }
    if (is.na(round1_tsne_path)) {
        round1_tsne_path = file.path(pipeline_mgr$directories$tSNE, paste0(pipeline.parameters$name, tsne_xy_extension))
    }

    # plot cluster assignments
    round1_tsne = readRDS(round1_tsne_path)
    round1_assign = readRDS(file.path(pipeline_mgr$directories$assign, paste0(pipeline.parameters$name,".cluster.assign.RDS")))
    plotter$start_single_plot(filename_generator$cluster_plot())
    ident.plot(round1_tsne[names(round1_assign), ], round1_assign, legend=TRUE)
    title(main="First round, Louvain Clustering")
    plotter$end_single_plot()

    # Load first-round ICA object
    ica.obj = readRDS(file.path(pipeline_mgr$directories$components, paste0(pipeline.parameters$name, '.flipped_sorted_raw_ica.RDS')))
    cell_rotations = ica.obj$cell_rotations


    first_round_curation_sheet_path = file.path(pipeline_mgr$directories$curation_sheets, paste0(pipeline.parameters$name, first_round_curation_sheet_extension))
    if (file.exists(first_round_curation_sheet_path)) {
        curation.data = read.table(first_round_curation_sheet_path, stringsAsFactors=F, sep=",", header=T, quote="\"")
        ic_status=tolower(curation.data$status)
    } else {
        ic_status = rep('real', ncol(cell_rotations))
    }
    possible_ic_status = c('real', 'artifact', 'doublet', 'outlier')

    plot_all_ics = function(tsne, cell_rotations, cluster_num=NULL) {
        if (is.null(cluster_num)) {
            plot_title_prefix = "First-round"
        } else {
            plot_title_prefix = paste0("Cluster ", cluster_num)
        }
        for (status in possible_ic_status) {

            plotter$start_plot_group(filename_generator$ic_plot_group(status))
            for (ic_num in 1:ncol(cell_rotations)) {
                if (ic_status[ic_num] == status) {
                    plotter$start_single_plot(filename_generator$ic_single_plot(status, ic_num, cluster_num))
                    cells = intersect(rownames(cell_rotations), rownames(tsne))
                    fplot(tsne[cells,],cell_rotations[cells,ic_num],title=sprintf("%s IC %d (%s)", plot_title_prefix, ic_num, status))
                    plotter$end_single_plot()
                }
            }
            plotter$end_plot_group()
        }
    }

    plot_all_ics(round1_tsne, cell_rotations)

    # This only includes cluster with > 1 IC
    ICs = readRDS(paste0(pipeline.parameters$name, all_ics_extension))
    if (use_curated_tsne) {
        curated_subcluster_tsne = lapply(readRDS(get_curated_results_path(pipeline.parameters)), function(cr) return(cr$ic$tSNE))
    }
    curation.files=get_curation_sheet_files(pipeline_mgr$directories$curation_sheets)
    # note that cluster_num is actually a string -- not all clusters have ICs
    for (cluster_num in names(ICs)) {
        # The ICs in subclusters object have not been flipped and sorted, and therefore don't line up with ICs
        # in curation sheets, so get them from ICA.RDS, which is post-flip&sort.
        cell_rotations = readRDS(getLatestIcaRds(cluster.num=cluster_num, name=pipeline.parameters$name))$cell_rotations
        if (use_curated_tsne) {
            tsne = curated_subcluster_tsne[[cluster_num]]
            if (is.null(tsne)) {
               cat_with_time(sprintf("WARNING: No curated t-SNE for cluster %s so using raw", cluster_num))
               tsne = ICs[[cluster_num]]$tsne
            }
        } else {
            tsne = ICs[[cluster_num]]$tsne
        }
        if (is.element(cluster_num, names(curation.files))) {
            curation.data = read.table(file.path(pipeline_mgr$directories$curation_sheets, curation.files[cluster_num]), 
                          stringsAsFactors=F, sep=",", header=T, quote="\"")
            ic_status=tolower(curation.data$status)
        } else {
            ic_status = rep('real', ncol(cell_rotations))
        }
        plot_all_ics(tsne, cell_rotations, cluster_num)
    }
}
