# MIT License
#
# Copyright 2017 Broad Institute
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Handles parsing of analysis/dataset specific parameters for clustering pipeline



# library(yaml)

#' @param parameter.filepath The YAML path, or if NULL, this is obtained from the command pipeline
#' @return list(yamlpath (string), other_args (vector))
GetYamlPathAndTrailingArgs = function(parameter.filepath=NULL) {
    custom.arguments = commandArgs(trailingOnly = T)
    if (is.null(parameter.filepath)) {
        if (length(custom.arguments) == 0) {
            stop('YAML path was not specified on the command line')
        }
        parameter.filepath = custom.arguments[1]
        custom.arguments = custom.arguments[-1]
    }
    return(list(yamlpath=parameter.filepath, other_args=custom.arguments))
}

GetPipelineParameters = function(parameter.filepath, copy.yaml=TRUE) {
    parameters = yaml.load_file(parameter.filepath)

    if (!is.null(parameters$workflow$working_directory)) {
        dir.create(parameters$workflow$working_directory, showWarnings = FALSE)
    } else {
        parameters$workflow$working_directory=dirname(normalizePath(parameter.filepath))
    }
    outputDirectory = normalizePath(parameters$workflow$working_directory)
    yamlDirectory = dirname(normalizePath(parameter.filepath))
    if (copy.yaml) {
        if (outputDirectory == yamlDirectory) {
            print("Not copying yaml to output directory because it is already there")
        } else if (!file.copy(parameter.filepath, parameters$workflow$working_directory, overwrite=TRUE)) {
            stop(paste("Could not copy", parameter.filepath, "to", parameters$workflow$working_directory))
        }
    }
    if (is.null(parameters$name)) {
        parameters$name = basename(outputDirectory)
    }
    return(parameters)
}

isNamedYamlList=function(l) {
    return(is.list(l) && !is.null(names(l)))
}

applyDefaultSettings=function(parameters, default_settings) {
    for (key in names(default_settings)) {
        if (!is.element(key, names(parameters))) {
            parameters[[key]] = default_settings[[key]]
        } else if (isNamedYamlList(parameters[[key]]) && isNamedYamlList(default_settings[[key]])) {
            # recurse
            parameters[[key]] = applyDefaultSettings(parameters[[key]], default_settings[[key]])
        } else if (isNamedYamlList(parameters[[key]]) || isNamedYamlList(default_settings[[key]])) {
            stop(sprintf("Mismatch between YAML parameters and default settings for key %s -- one is a named list and the other is not", key))
        } else {
            # Don't assign, but make sure types are compatible.
            # For lists, get the first element.
            v1 = if (is.list(parameters[[key]])) parameters[[key]][[1]] else parameters[[key]]
            v2 = if (is.list(default_settings[[key]])) default_settings[[key]][[1]] else default_settings[[key]]
            if (class(v1) != class(v2) &&
                !(is.numeric(v1) && is.numeric(v2))) {
            stop(sprintf("Difference between class of %s attribute between yaml(%s) and default settings(%s)",
                key, class(v1), class(v2)))
            }
        }
    }
    return(parameters)
}
