# MIT License
#
# Copyright 2017 Broad Institute
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Class for managing a multi-stage workflow, including stopping in the middle, picking up where left off, etc.
# USAGE:
# - INITIALIZATION:
#   Provide pipeline parameters, and list of stage names in order.  
#   Stage names should be filename-friendly, i.e. not have spaces, slashes, etc.
#   stage_mgr=create_stage_manager(pipeline.parameters, "stage1", "stage2", ...)
#   pipeline.parameters may have the following attributes:
#   - first_stage: name of  stage to start with.  If not provided, start at the first stage
#   - last_stage: name of the last stage to run.  If not provided, run through to the last stage.
#   - save_intermediates: If provided and true, save state after each stage, not just the last stage executed.

# - WRAP STAGES LIKE THIS:
# if (stage_mgr$beforeStage("my_stage")) {
#   do the work for this stage
#   stage_mgr$afterStage("names", "of", "objects", "to.be.saved")
# }

#library(methods)

create_stage_manager = function(pipeline.parameters, stage_directory, stages) {
  stage_mgr = stage_manager$new(stages=stages)
  stage_mgr$initializeFromParameters(pipeline.parameters$name, pipeline.parameters$workflow, stage_directory)
  return(stage_mgr)
}

stage_manager = setRefClass("stage_manager",
              fields = list(
                    stages="character",
                    firstRequestedStage="integer",
                    lastRequestedStage="integer",
                    fileBase="character",
                    currentStage="integer",
                    beforeFirstStagePrefix="character",
                    saveIntermediates="logical",
                    cleanIntermediates="logical"),
              
              methods = list(
                    initializeFromParameters = function(name, workflow_parameters, stage_directory) {
                        firstRequestedStage <<- getStageIndex(workflow_parameters$first_stage, as.integer(1))
                        lastRequestedStage <<- getStageIndex(workflow_parameters$last_stage, length(stages))
                        fileBase <<- paste0(stage_directory, "/", name)
                        # No longer an option
                        saveIntermediates <<- TRUE
                        cleanIntermediates <<- workflow_parameters$clean_intermediates
                        currentStage <<- as.integer(0)
                        if (is.null(workflow_parameters$before_first_stage_prefix)) {
                            beforeFirstStagePrefix <<- character(0)
                        } else {
                            beforeFirstStagePrefix <<- paste0(dirname(workflow_parameters$before_first_stage_prefix), "/",
                                        stage_directory, basename(workflow_parameters$before_first_stage_prefix))
                        }
                    },
                    
                    beforeStage = function(stage, envir= parent.frame()) {
                      newStageIndex = getStageIndex(stage, NULL)
                      if (is.null(newStageIndex)) {
                        stop(paste0("Unrecognized stage: ", stage))
                      }
                      if ((newStageIndex < firstRequestedStage) || (newStageIndex > lastRequestedStage)) {
                        return(FALSE)
                      } else {
                        if (currentStage != as.integer(0) && currentStage + 1 != newStageIndex) {
                          stop(paste0("Incorrect stage order. ", stage, " should not follow ", stages[currentStage]))
                        }
                        currentStage <<- newStageIndex
                        if (currentStage == firstRequestedStage) {
                          loadStage(currentStage, envir=envir)
                        }
                        cat_with_time(paste0("Start stage ", stages[currentStage]))
                        return(TRUE)
                      }
                    },

                    getStageIndex = function(stage, defaultVal) {
                      if (is.null(stage)) {
                        if (is.null(defaultVal)) {
                          stop("stage is null and defaultVal is null")
                        }
                        return(defaultVal)
                      } else {
                        ret = which(stages==stage)
                        if (length(ret) == 0) {
                          stop(sprintf("Unknown stage: %s.  Possible stages: %s", stage, paste(stages, collapse=",")))
                        }
                        return(ret)
                      }
                    },

                    # Name is a misnomer -- it actually loads the result of the previous stage
                    loadStage = function(stage, forceLoad=FALSE, envir = parent.frame()) {
                      if (is.character(stage)) {
                        stage = getStageIndex(stage, NULL)
                      }
                      if (forceLoad || (stage > 1 && stage == firstRequestedStage)) {
                        load(getStageFile(stage - 1), envir=envir)
                      }
                    },

                    # The arguments should be strings, which are the names of the objects to be saved
                    afterStage = function(...) {
                      if (saveIntermediates || currentStage == lastRequestedStage) {
                        # save an empty file even if no objects, so next stage can read it if necessary.
                        save(list=c(...), file=getStageFile(currentStage), envir=parent.frame())
                      }
                      if (currentStage == lastRequestedStage) {
                        if (lastRequestedStage == length(stages)) {
                            if (cleanIntermediates) {
                                cat_with_time("Cleaning intermediate files")
                                # Do not clean anything before firstRequestedStage
                                for (stageIndex in firstRequestedStage:lastRequestedStage) {
                                    stageFile = getStageFile(stageIndex)
                                    if (file.exists(stageFile)) {
                                        cat_with_time(paste0("Removing stage file ", stageFile))
                                        file.remove(stageFile)
                                    }
                                }
                            }
                        } else {
                            cat_with_time(paste0("Stopping after stage ", stages[currentStage]))
                            warnings()
                            quit("no")
                        }
                      }
                      cat_with_time(paste0("End stage ", stages[currentStage]))
                      },

                      # Assumes PWD is output directory
                      getStageFile = function(stage) {
                          if (length(beforeFirstStagePrefix) == 0 || stage >= firstRequestedStage) {
                              base = fileBase
                          } else {
                              base = beforeFirstStagePrefix
                        }
                        return(paste(base, "stage", stages[stage], "Robj", sep="."))
                      }
                    ))

