# MIT License
#
# Copyright 2017 Broad Institute
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

strEndsWith<-function(theString,theExt) {
    return(substring(theString,1+nchar(theString)-nchar(theExt))==theExt)
}

strStartsWith = function(theString, thePrefix) {
    return(strtrim(theString, nchar(thePrefix)) == thePrefix)
}

removeSuffix = function(str, suffix) {
    if (!strEndsWith(str, suffix)) {
        stop(paste0("'", str, "' does not end with '", suffix, "'"))
    }
    return(strtrim(str, nchar(str) - nchar(suffix)))
}

maybeRemoveSuffix = function(str, suffix) {
    if (!strEndsWith(str, suffix)) {
        return(str)
    } else {
        return(strtrim(str, nchar(str) - nchar(suffix)))
    }
}

#' Read a data.table using data.table fread()
#' If inFile ends with .gz, it is gunzipped in a way that is robust to running out of disk space.
#' Environment variable DROPSEQ_TMPDIR should be set to a directory that has plenty of room and is writable.
fastRead<-function(inFile, ...) {
    if (!file.exists(inFile)) {
        stop(paste0(inFile, " does not exist"))
    }
    # fread writes process output to /dev/shm if it exists, which is OK so long as it is big enough.
    # If that fails, fall back to function that uses an explicit TMPDIR
    if (length(grep(".gz", inFile))==1) {
        return(tryCatch(fread(paste("gunzip","-c",inFile,sep=" "),data.table=T, ...), error=function(e) fastReadBigGz(inFile, ...)))
    } else {
        a=fread(inFile, data.table=T, ...)
        return(a)
    }
}

#' internal function
fastReadBigGz<-function(inFile, ...) {
    # Grid Engine overwrites TMPDIR, so support our own environment variable that won't get clobbered.
    tmpDir = Sys.getenv('DROPSEQ_TMPDIR')
    if (nchar(tmpDir) == 0) {
        tmpDir = Sys.getenv('TMPDIR')
    }
    if (nchar(tmpDir) == 0) {
        tmpDir = tempdir()
    }
    t=tempfile(tmpdir=tmpDir, fileext=".tsv")
    on.exit(unlink(t), add = TRUE)
    cmd=paste("gunzip -c", inFile, ">", t)
    retval = system(cmd,intern=F) != 0
    if (retval != 0) {
        stop(paste0(cmd, " failed with status ", retval))
    }
    return(fastRead(t, ...))
}

# Returns a data.frame with genes as rownames, no GENE column.
# Cells (columns) are in decreasing order by size.
read_dge_gz<-function(file) {
    dge = fastRead(file)
    setkey(dge,GENE)
    gene_names = dge$GENE
    dge[, GENE:=NULL]
    rownames(dge) = gene_names

    # order cells by expression
    colsums.full.data=dge[,lapply(.SD,sum,na.rm=TRUE),.SDcols=colnames(dge)]
    order.data=names(colsums.full.data)[order(colsums.full.data,decreasing=T)]
    full.data.cells=dge[,order.data,with=F]
    full.data.cells = as.data.frame(full.data.cells)
    rownames(full.data.cells) = gene_names
    return(full.data.cells)
}



RawDgeMultiplier = 10000

#' this function converts normalized data back to DGEs (can be useful for performing counts-based analyses)
#' rownames of input must be gene names.  If input has GENE column, it is removed.
#' @param a log-normalized DGE
#' @param returnDataFrame if TRUE, returns a data.frame with gene rownames; else returns a data.table with GENE column and gene rownames.
ConvertNormalizedToDGE = function(a, returnDataFrame=TRUE) {
    genes=row.names(a)

    aa = setDT(a)
    if ('GENE' %in% colnames(aa)) {
        aa[, GENE:=NULL]
    }
    #check to see if a column contains no 0 values.  This is false if not ok.
    idxOK=apply(a, 2, function (x) any(which(x==0)))
    if (any(idxOK==F)) warning("Some cells had no 0 values for expression.  You need at least one gene with 0 expression to correctly recover raw counts")

    aa = (exp(aa)-1)/RawDgeMultiplier

    fixColumn<-function (x) {
        txptsum<-1/(sort(unique(x))[2])
        round(x*txptsum)
    }

    for (col in colnames(aa)) {
        set(aa, j=col, value= fixColumn (aa[[col]]))
    }

    if (returnDataFrame) {
        setDF(aa)
        rownames(aa)=genes
        return(aa)
    } else {
        return(addGeneColumnToDgeDt(aa, genes))
    }
}

#' internal function
addGeneColumnToDgeDt = function(dt, genes) {
    cell_barcodes = colnames(dt)
    dt$GENE=genes
    setcolorder(dt, c('GENE', cell_barcodes))
    rownames(dt)=genes
    return(dt)
}

#' Log-normalize a DGE data frame or data.table.
#' @param a Raw(integer) DGE data fram or data.table. rownames of input must be gene names.  If input has GENE column, it is removed.
#' @param returnDataFrame If true, returns a data.frame with gene rownames; else returns a data.table with GENE column and gene rownames.
ConvertDgeToNormalized = function(a, returnDataFrame = TRUE) {
    genes=rownames(a)
    a = setDT(a)
    if ('GENE' %in% colnames(a)) {
        a[, GENE:=NULL]
    }
    a = log(RawDgeMultiplier*sweep(a,2,colSums(a),'/')+1)

    if (returnDataFrame) {
        setDF(a)
    rownames(a) = genes
    return(a)
    } else {
        setDT(a)
        return(addGeneColumnToDgeDt(a, genes))
    }
}

#' normalize a DGE data frame or data.table, so that each cell's expression sums to 1.
#' @param a Raw(integer) DGE data fram or data.table. rownames of input must be gene names.  If input has GENE column, it is removed.
#' @param returnDataFrame If true, returns a data.frame with gene rownames; else returns a data.table with GENE column and gene rownames.
ConvertDgeToNormalizedNoLog = function(a, returnDataFrame = TRUE) {
    genes=rownames(a)
    a = setDT(a)
    if ('GENE' %in% colnames(a)) {
        a[, GENE:=NULL]
    }
    a = sweep(a,2,colSums(a),'/')
    #cs=colSums(a)
    #for (i in 1:length(cs)) {
    #    a[,i]=a[,i, with=FALSE]/cs[i]
    #}

    if (returnDataFrame) {
        setDF(a)
    rownames(a) = genes
    return(a)
    } else {
        setDT(a)
    return(addGeneColumnToDgeDt(a, genes))
    }
}

#' @param normalizedDge a DGE (without GENE column) as produced by ConvertDgeToNormalizedNoLog
#' @return a DGE as produced by ConvertDgeToNormalized
ScaleAndLogDge = function(normalizedDge) {
    return(log(RawDgeMultiplier*normalizedDge + 1))
}

#' Creates a file connection with the given open mode.
#' @param file If ends with ".gz", a gzfile() is created; else a regular file() connection.
open_conn = function(file, open="") {
    if (strEndsWith(file, ".gz")) {
        return(gzfile(file, open=open))
    } else {
        return(file(file, open=open))
    }
}

#' Write a tab-separated file with row and column labels, including a label for the rownames.
#' Additional args are passed to write.table
#' @param x the data frame to be written.
#' @param file where to write the data
#' @param the column header for the rownames
write_table_with_label_for_rownames = function(x, file, rowname_label, quote=FALSE, sep="\t", col.names=TRUE, ...) {
    conn = open_conn(file=file, open="w")
    cat(paste0(rowname_label, '\t'), file=conn)
    write.table(x, conn, col.names=col.names, row.names=TRUE, quote=quote, sep=sep, ...)
    close(conn)
}

ensure_trailing_slash=function(directory) {
    if (!strEndsWith(directory, "/")) {
        directory = paste0(directory, "/")
    }
    return(directory)
}

#' Check if a directory exists, and create it if not.  Errors if cannot be created or exists but is not a directory.
#' @return the directory, always with a trailing slash.  Raises an error if cannot be created
ensure_directory_exists = function(directory, recursive=TRUE) {
    directory = ensure_trailing_slash(directory)
    if (dir.exists(directory)) {
    return(directory);
    } else if (file.exists(directory)) {
        stop(paste0("Cannot create directory because there is an existing file with the same name: ", directory))
    } else if (!dir.create(directory, recursive=recursive)) {
        stop(paste0("Error creating directory: ", directory))
    } else {
        return(directory)
    }
}

#' Convert data frame stored in RDS file to a tab-separated file
#' @param rds.file the file to be read
#' @param tsv.file the file to be written.  If ends with .gz, file is gzipped.
#' @param rowname_label The column header for the rownames
convert_rds_to_tsv = function(rds.file, tsv.file, rowname_label, ...) {
    df = readRDS(rds.file)
    write_table_with_label_for_rownames(df, file=tsv.file, rowname_label=rowname_label, ...)
}

#' Count the number of lines at the beginning of the (possible gzipped) file that start with '@'
countHeaderLines = function(path) {
    conn = file(path, "r")
    on.exit(close(conn), add = TRUE)
    headerLength = 0
    while(strStartsWith(line<-(readLines(conn, n=1)[[1]]), '@') || strStartsWith(line, '#')) {
        headerLength = headerLength + 1
    }
    return(headerLength)
}

dge_header_attribute_name='Drop-seq.dge.header'

read_dge_txt = function(path) {
    headerLength = countHeaderLines(path)
dge = fastRead(path, skip=headerLength)
if (headerLength > 0) {
    header = readLines(path, n=headerLength)
attr(dge, dge_header_attribute_name) = header
}
return(dge)
}

#' returns list(genes, cell_barcodes)
loadSparseDgeNames=function(file) {
    conn=file(file, "r")
    genes=c()
    cell_barcodes=c()
    while(TRUE) {
        line=readLines(con=conn, n=1, ok=FALSE)
        if (!strStartsWith(line, '%')) {
            break
        }
        if (strStartsWith(line, '%%GENES\t')) {
            these_genes=strsplit(line, "\t", fixed=TRUE)[[1]]
            genes = append(genes, these_genes[2:length(these_genes)])
        } else if (strStartsWith(line, '%%CELL_BARCODES\t')) {
            these_cells=strsplit(line, "\t", fixed=TRUE)[[1]]
            cell_barcodes = append(cell_barcodes, these_cells[2:length(these_cells)])
        }
    }
    close(conn)
    return(list(genes=genes, cell_barcodes=cell_barcodes))
}

#' Load a sparse DGE (either real or integer), that may have gene names and cell_barcode names in header
#' @return A sparse matrix in triplet form (dgTMatrix)
loadSparseDge = function(file, column_compressed=FALSE) {
    genes_and_cell_barcodes = loadSparseDgeNames(file)
    ret=readMM(file)
    rownames(ret)=genes_and_cell_barcodes$genes
    colnames(ret)=genes_and_cell_barcodes$cell_barcodes
    if (column_compressed) {
        return(convertTripletCompressedMatrixToColumnCompressed(ret))
    } else {
        return(ret)
    }
}

#' Convert a sparse matrix from triplet form to column-compressed form
convertTripletCompressedMatrixToColumnCompressed = function(triplet_matrix) {
    if (! ('dgTMatrix' %in% is(triplet_matrix))) {
        stop("Argument is not instance of dgTMatrix")
    }
    return(sparseMatrix(i=triplet_matrix@i, j=triplet_matrix@j, x=triplet_matrix@x, dim=dim(triplet_matrix), dimnames=dimnames(triplet_matrix), index1=FALSE))
}


cat_with_time = function(...) {
    cat(paste0(Sys.time(), ": "), ...)
cat("\n") # Force newline to be printed
}

#' Invoke list.files, and throw exception if number of files matched != 1
#' @param description of type of file being sought, for error messages
#' @param other arguments are same as for list.files
get_single_file_matching_regexp=function(description, path=".", ...) {
    files = list.files(path=path, ...)
if (length(files) > 1) {
    stop(sprintf("More than one %s file found in %s", description, path))
} else if (nchar(files) == 0) {
    stop(sprintf("No %s files found in %s", description, path))
}
return(files)
}


#' Read an rdata file and return the nth object in it without clobbering the caller's environment.
load_obj <- function(f, whichObject=1) {
    env <- new.env()
    nm <- load(f, env)[whichObject]
    env[[nm]]
}

#' Call system2 with stdout=TRUE, and stop() if non-zero exit status
#' Arguments as in system2()
#' @return stdout of command
system2_capture_stdout_check_error=function(command, args=character(), ...) {
    ret = system2(command=command, args=args, stdout=TRUE, ...)
    status = attr(ret, 'status')
    if (!is.null(status) && status != 0) {
        full_errmsg = sprintf("Exit status %d from '%s %s'", status, command, paste(args, collapse=" "))
        errmsg = attr(ret, 'errmsg')
        if (is.null(errmsg)) {
            full_errmsg = paste0(full_errmsg, " : ", errmsg)
        }
        stop(full_errmsg)
    } else {
        return(ret)
    }
}
